/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "cell_grid.h"

using namespace std;

/***************************************************************************/

bool                                        emd::uni::use_cell_info = false;

double                                      emd::uni::cell_len_x;
double                                      emd::uni::cell_len_y;
double                                      emd::uni::cell_len_z;

int                                         emd::uni::cell_num_x;
int                                         emd::uni::cell_num_y;
int                                         emd::uni::cell_num_z;

double                                      emd::uni::cell_offset_x;
double                                      emd::uni::cell_offset_y;
double                                      emd::uni::cell_offset_z;

std::vector<emd::uni::cell>                emd::uni::cells;            /* container of cells */
__gnu_cxx::hash_map<string, int, __gnu_cxx::hash<string>, emd::uni::eqstr>
                                            emd::uni::cell_key_to_idx;  /* cell key hash map */

/***************************************************************************/

int
emd::uni::init_cell_grid ()
{

  /* NOTE add interaction cutoffs and offsets here */

  double             offset_dists[]  = {ees::max_offset, re2::max_offset};
  double             cutoff_dists[]  = {ees::r_c + ees::r_s, re2::r_c + re2::r_s};

  sort( offset_dists, offset_dists + sizeof(offset_dists)/sizeof(double) );
  sort( cutoff_dists, cutoff_dists + sizeof(cutoff_dists)/sizeof(double) );

  double             max_cutoff_dist = cutoff_dists[sizeof(cutoff_dists)/sizeof(double)-1];
  double             max_offset_dist = offset_dists[sizeof(offset_dists)/sizeof(double)-1];
  double             min_len         = max_cutoff_dist + max_offset_dist;

  use_cell_info  =  false;
  cell_len_x     =  min_len;
  cell_len_y     =  min_len;
  cell_len_z     =  min_len;
  cell_offset_x  =  0;
  cell_offset_y  =  0;
  cell_offset_z  =  0;

  if (x_periodic)
  {

    cell_num_x    = (int)floor(x_width/min_len);

    if (cell_num_x==0) return EXIT_FAILURE;

    cell_len_x    = x_width/cell_num_x;
    cell_offset_x = x_lower_bound;

  }

  if (y_periodic)
  {

    cell_num_y    = (int)floor(y_width/min_len);

    if (cell_num_y==0) return EXIT_FAILURE;

    cell_len_y    = y_width/cell_num_y;
    cell_offset_y = y_lower_bound;

  }

  if (z_periodic)
  {

    cell_num_z    = (int)floor(z_width/min_len);

    if (cell_num_z==0) return EXIT_FAILURE;

    cell_len_z    = z_width/cell_num_z;
    cell_offset_z = z_lower_bound;

  }

  /* cell-gridding is possible, so use it */
  use_cell_info = true;

  return EXIT_SUCCESS;

}

/***************************************************************************/

int
emd::uni::get_cell_index (const int & n_x, const int & n_y, const int & n_z)
{

  cell_index_to_key(def_key_buff, n_x, n_y, n_z);

  cell_hash_it found_element = cell_key_to_idx.find(def_key_buff);

  if ( found_element != cell_key_to_idx.end() )
    return (*found_element).second;
  else
    return CELL_NOT_FOUND;
}

/***************************************************************************/

void
emd::uni::check_rigid_body_cell_info (const unsigned int & idx)
{

  int n_x = (int)floor( (gsl_vector_get(rigid_bodies[idx].x, 0)-cell_offset_x) / cell_len_x );
  int n_y = (int)floor( (gsl_vector_get(rigid_bodies[idx].x, 1)-cell_offset_y) / cell_len_y );
  int n_z = (int)floor( (gsl_vector_get(rigid_bodies[idx].x, 2)-cell_offset_z) / cell_len_z );

  if ( (n_x != cells[rigid_bodies[idx].cell_idx].n_x) ||
       (n_y != cells[rigid_bodies[idx].cell_idx].n_y) ||
       (n_z != cells[rigid_bodies[idx].cell_idx].n_z) )
  {
    rigid_bodies[idx].cell_state = emd::rigid_body::EMD_RIGIDBODY_CELL_STATE_UPDATE;
  }

}

/***************************************************************************/

void
emd::uni::update_rigid_body_cell_info (const unsigned int & idx)
{

  switch (rigid_bodies[idx].cell_state)
  {

    case emd::rigid_body::EMD_RIGIDBODY_CELL_STATE_NONE:
    {
      break;
    }

    case emd::rigid_body::EMD_RIGIDBODY_CELL_STATE_INIT:
    {

      add_rigid_body_to_cell(idx);

      /* set cell_idx entry of the rigid body */
      rigid_bodies[idx].cell_state = emd::rigid_body::EMD_RIGIDBODY_CELL_STATE_NONE;

      break;
    }

    case emd::rigid_body::EMD_RIGIDBODY_CELL_STATE_UPDATE:
    {

      /* remove from the previous cell */
      cells[rigid_bodies[idx].cell_idx].rigid_body_indices.remove(idx);

      add_rigid_body_to_cell(idx);

      /* set cell_idx entry of the rigid body */
      rigid_bodies[idx].cell_state = emd::rigid_body::EMD_RIGIDBODY_CELL_STATE_NONE;

      break;

    }

  }

}

/***************************************************************************/

void
emd::uni::add_rigid_body_to_cell (const unsigned int & idx)
{

  int n_x = (int)floor( (gsl_vector_get(rigid_bodies[idx].x, 0)-cell_offset_x) / cell_len_x );
  int n_y = (int)floor( (gsl_vector_get(rigid_bodies[idx].x, 1)-cell_offset_y) / cell_len_y );
  int n_z = (int)floor( (gsl_vector_get(rigid_bodies[idx].x, 2)-cell_offset_z) / cell_len_z );

  cell_index_to_key (def_key_buff, n_x, n_y, n_z);
  cell_hash_it cell_idx = cell_key_to_idx.find(def_key_buff);

  if ( cell_idx != cell_key_to_idx.end() )
  {

/*    cout << def_key_buff << "  " << n_x << "," << n_y << "," << n_z << "  "
         << cells[cell_key_to_idx[def_key_buff]].n_x << ","
         << cells[cell_key_to_idx[def_key_buff]].n_y << ","
         << cells[cell_key_to_idx[def_key_buff]].n_z << endl;*/

    /* push the rigid body idx into the appropriate cell */
    cells[(*cell_idx).second].rigid_body_indices.push_back(idx);

    /* set cell_idx entry of the rigid body */
    rigid_bodies[idx].cell_idx = (*cell_idx).second;

  }
  else
  {

    /* create a new cell */
    cell new_cell(n_x, n_y, n_z);

    /* add the cell to the cell vector */
    cells.push_back(new_cell);

    /* add the rigid body to the cell */
    cells[cells.size()-1].rigid_body_indices.push_back(idx);

    /* update hash table */
    cell_key_to_idx[def_key_buff] = cells.size() - 1;

    /* set cell_idx entry of the rigid body */
    rigid_bodies[idx].cell_idx    = cells.size() - 1;

#ifdef CELL_DEBUG
    cout << "* emd::uni::update_rigid_body_cell_info: a new cell added ("
         << n_x << ", " << n_y << ", " << n_z << ")." << endl;
#endif

  }

#ifdef CELL_DEBUG
  cout << "* emd::uni::update_rigid_body_cell_info: the cell info of rigid body ("
       << idx << ") updated (" << n_x << ", " << n_y << ", " << n_z << ")." << endl;
#endif

}

/***************************************************************************/

void
emd::uni::update_rigid_bodies_cell_info ()
{

  for (unsigned int i=0; i<rigid_bodies_count; i++)
  {
    check_rigid_body_cell_info(i);
    update_rigid_body_cell_info(i);
  }

}

/***************************************************************************/
