/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef CELL_GRID_H
#define CELL_GRID_H

#include <cstdio>
#include <vector>
#include <list>
#include <set>
#include <ext/hash_map>
#include <algorithm>
#include <string>

#include "moluniverse.h"
#include "rigid_body.h"

//#define CELL_DEBUG

namespace __gnu_cxx
{
  template<> struct hash< std::string >
  {
    size_t operator()( const std::string& x ) const
    {
      return hash< const char* >()( x.c_str() );
    }
  };
}

namespace emd
{

namespace uni
{

  /*************************************************************************/

  class cell
  {
  public:

    int                           n_x, n_y, n_z;         /* index of the cell */
    std::list<unsigned int>      rigid_body_indices;    /* indices of the contained rigid bodies */

    cell(int n_x0, int n_y0, int n_z0)
    {
      this->n_x = n_x0;
      this->n_y = n_y0;
      this->n_z = n_z0;
    }

  };

  /*************************************************************************/

  struct eqstr
  {
    bool operator()(const std::string & s1, const std::string & s2) const
    {
      return (s1==s2);
    }
  };

  /*************************************************************************/

  struct ltstr
  {
    bool operator()(const std::string & s1, const std::string & s2) const
    {
      return (s1<s2);
    }
  };

  /*************************************************************************/

  typedef  __gnu_cxx::hash_map<std::string, int, __gnu_cxx::hash<std::string>, eqstr>::iterator
    cell_hash_it;
  typedef  std::vector<cell>::iterator               cell_it;

  /*************************************************************************/

  const int                                           CELL_NOT_FOUND = -1;
  const int                                           MAX_KEY_SIZE   = 16;

  extern  bool                                        use_cell_info;

  extern  double                                      cell_len_x;
  extern  double                                      cell_len_y;
  extern  double                                      cell_len_z;

  extern  int                                         cell_num_x;
  extern  int                                         cell_num_y;
  extern  int                                         cell_num_z;

  extern  double                                      cell_offset_x;
  extern  double                                      cell_offset_y;
  extern  double                                      cell_offset_z;

  extern  std::vector<cell>                          cells;            /* container of cells */
  extern  __gnu_cxx::hash_map<std::string, int, __gnu_cxx::hash<std::string>, eqstr>
                                                      cell_key_to_idx;  /* cell key hash hash map */

  static  char                                        def_key_buff[MAX_KEY_SIZE];

  /*************************************************************************/

  /* get the key std::string of a cell vector */
  inline void
  cell_index_to_key (char * buffer, const int & n_x, const int & n_y,
                     const int & n_z)
  {
    std::sprintf (buffer, "%d,%d,%d", n_x, n_y, n_z );
  }

  /*************************************************************************/

  /* get the cell index from a cell vector */
  int
  get_cell_index (const int & n_x, const int & n_y, const int & n_z);

  /*************************************************************************/
  /* used to give a sorted list of the cell edge sizes */
  struct lt_pair_int_double
  {
    bool operator()(const std::pair<int, double> & A, const std::pair<int, double> & B)
    {
      return (A.second < B.second);
    };
  };

  /*************************************************************************/

  /* init the cell grid variables/routines */
  int init_cell_grid ();

  /******************************************************************************************
   * 1) if the state is STATE_INIT, the rigid body will be added to the appropriate cell    *
   * 2) if the state is STATE_UPDATE, the rigid_body will be added to the appropriate cell  *
   *    and it will be removed from it's current cell                                       *
   * 3) if the state is STATE_NONE, do nothing                                              *
   ******************************************************************************************/
  void update_rigid_body_cell_info (const unsigned int & rigid_body_idx);

  /* check too see if the object needs an update and update cell_state variable */
  void check_rigid_body_cell_info (const unsigned int & rigid_body_idx);

  /* loop through all rigid bodies */
  void update_rigid_bodies_cell_info ();

  /* ... */
  void add_rigid_body_to_cell (const unsigned int & rigid_body_idx);

}; /* uni */

}; /* emd */

#endif
