/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/***************************************************************************
 * todo:                                                                   *
 *                                                                         *
 ***************************************************************************/

/***************************************************************************
 * References:                                                             *
 *                                                                         *
 * [1] H. Kamberaj, R. J. Low, M. P. Neal                                  *
 *     "Time reversible and symplectic integrators for molecular dynamics  *
 *     simulations of rigid molecules", J. Chem. Phys. 122, 224114 (2005)  *
 *                                                                         *
 ***************************************************************************/

#ifndef DYNAMICS_H
#define DYNAMICS_H

#include <ctime>
#include <cmath>

#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_sf.h>

#include <quaternion/gsl_quaternion.h>

#include "consts.h"
#include "gslext.h"
#include "rigid_body.h"
#include "stringext.h"
#include "moluniverse.h"
#include "moluniverse_comp.h"
#include "paramdb.h"

#define DYNAMICS_DEBUG
// #define DETAILED_LOG

namespace emd
{

namespace uni
{

namespace dyn
{

  /*************************************************************************/

  /* ensembles */
  enum integrator_type   { NVE_MILLER_SYMPLECTIC,
                           NVT_MILLER_SYMPLECTIC_NHC,
                           NPT_MILLER_SYMPLECTIC_NHC };

  enum temp_drift_type   { TEMP_DRIFT_LOG,
                           TEMP_DRIFT_LIN };

  /*************************************************************************/

  typedef                std::vector<unsigned int>::iterator      int_it;

  /*************************************************************************/

  /* parameters and their DEFAULT values */

  extern integrator_type   integrator;

  extern std::string       integrator_desc;

  extern double            dt;
  extern long int          N_ignore_steps;
  extern long int          N_steps;

  extern int               nl_update_interval;
  extern int               stats_interval;

  extern bool              energy_conservator;

  extern double            goal_energy;
  extern double            goal_pressure;
  extern double            goal_temperature;

  extern bool              fix_cm;
  extern bool              fix_T;
  extern bool              fix_P;
  extern bool              reset_thermostat;
  extern bool              reset_barostat;
  extern bool              log_to_file;
  extern bool              log_to_stdout;
  extern bool              log_to_stats_file;
  extern bool              log_force_on_surfs;

  extern std::string       log_file_name;
  extern std::string       log_file_name_force_on_surfs;
  extern std::string       log_file_name_stats;

  /* drifts */

  extern bool              temperature_drift, concentration_drift;
  extern temp_drift_type   temp_drift;
  extern double            init_temperature, final_temperature, final_concentration;

  struct sph_radius_drift_spec
  {
  public:
    int idx;
    double init_rad, final_rad;
  };

  extern std::vector<sph_radius_drift_spec>  sph_radius_drift_vec;

  /* auxilliary variables */

  static unsigned int                 active_num;                /* number of active bodies */
  static std::vector<unsigned int>    active_list;               /* vector of active bodies */

  static double            dt_2;                                 /* integration time step squared */
  //{{{restart: changing t from static to external 
  extern double            t                       ;
  //}}}
  static double            pressure;                             /* current pressure */
  static double            temperature;                          /* current temperature */
  static double            concentration;                        /* current concentration */
  static double            new_volume, init_volume;
  static double            order_param;                          /* order parameter */

  static double            NPT_hamiltonian;
  static double            NVT_hamiltonian;

  static std::vector<double>     stat_temp_vec;
  static std::vector<double>     stat_press_vec;
  static std::vector<double>     stat_conc_vec;
  static std::vector<double>     stat_en_vec;
  static std::vector<double>     stat_p2_vec;

  static int               iter                    =  0;         /* iteration count */
  static int               prev_iter               =  0;         /* iterations before the last nl update */
  static int               stats_counter           =  0;

  /**** order parameter eval ws **********************************************/

  static gsl_vector * Q_eigen_val = gsl_vector_alloc(3);
  static gsl_matrix * Q_tensor    = gsl_matrix_alloc(3, 3);
  static gsl_matrix * Q_eigen_vec = gsl_matrix_alloc(3, 3);

  /***************************************************************************/

  /* workspace (lazy initialization) */

  static bool              init_ws_flag            =  false;     /* workspace init flag */

  static gsl_vector *      tmp_v  =  gsl_vector_alloc(3);
  static gsl_vector *      tmp_v2 =  gsl_vector_alloc(3);


  /***************************************************************************/

  void init();

  /***************************************************************************
   * time-reversible and symplectic integrator, using Miller's rigid-        *
   * body Hamiltonian [1].                                                   *
   ***************************************************************************/

  static double                      NHC_d          =  3;        /* dimension of the system */

  /* Yoshida-Suzuki constants */
  static double                      YS_3[3]        =  {  1.35120719195965763405,
                                                         -1.7024143839193152681,
                                                          1.35120719195965763405 };

  static int                         THERMO_PART_ORDER  =  3;
  static int                         BARO_PART_ORDER    =  3;

  extern double                      NHC_g_t;           /* translational degree of freedom */
  extern double                      NHC_g_r;           /* rotational degrees of freedom */
  extern double                      NHC_g;             /* degrees of freedom (trans + rot) */

  extern int                         LIOUVILLE_PART_ORDER;

  extern int                         NHC_T_length;
  extern double                      NHC_T_freq;

  extern int                         NHC_P_length;
  extern double                      NHC_P_freq;

  extern double *                    NHC_T_ksi_t;       /* NHC velocity (translational) */
  extern double *                    NHC_T_ksi_r;       /* NHC velocity (rotational) */
  extern double *                    NHC_T_G_t;         /* NHC force (translational) */
  extern double *                    NHC_T_G_r;         /* NHC force (rotational) */
  extern double *                    NHC_T_s_t;         /* NHC state variable (translational) */
  extern double *                    NHC_T_s_r;         /* NHC state variable (rotational) */
  extern double *                    NHC_T_Q_t;         /* NHC masses (translational) */
  extern double *                    NHC_T_Q_r;         /* NHC masses (rotational) */

  extern double                      NHC_P_eps;         /* NHB state (volume) */
  extern double                      NHC_P_ksi_eps;     /* NHB momentum (volume) */
  extern double                      NHC_P_G_eps;       /* NHB force (volume) */
  extern double                      NHC_P_Q_eps;       /* NHB mass (volume) */

  extern double *                    NHC_P_eta;         /* NHB chain state (chain) */
  extern double *                    NHC_P_ksi_eta;     /* NHB chain momenta (chain) */
  extern double *                    NHC_P_G_eta;       /* NHB chain forces (chain) */
  extern double *                    NHC_P_Q_eta;       /* NHB masses (chain) */

  extern double                      brendsen_kappa;

  /* 10th order Taylor series of sinh(x)/x about x_0 = 0 */
  double sinch_10 (const double & x);

  /* NOTE: R(t) must be up to date! */
  gsl_quaternion get_principal_frame_torque (const emd::rigid_body & rb);

  gsl_quaternion apply_liouville_op_step_P (const gsl_quaternion & Q,
    const gsl_quaternion & P, const gsl_quaternion & N, const double & delta_t);

  void apply_no_squish_rot_P_Q_1 (emd::rigid_body & rb, const double & delta_t);
  void apply_no_squish_rot_P_Q_2 (emd::rigid_body & rb, const double & delta_t);
  void apply_no_squish_rot_P_Q_3 (emd::rigid_body & rb, const double & delta_t);

  /* NOTE NVE */

  /* init */
  void init_NVE_miller_symplectic();

  /* stepper */
  void step_NVE_miller_symplectic();

  /* NOTE NVT */

  /* init */
  void init_NVT_miller_symplectic_NHC(const bool & reset_thermostat);

  /* evaluate thermostat forces with respect to v(t), omega(t) and l(t) */
  void eval_NHC_thermostat_forces();

  void eval_NVT_hamiltonian();

  void step_NHC_thermostat();

  void step_NHC_thermostat_new (const double & delta_t);

  /* stepper */
  void step_NVT_miller_symplectic_NHC();

  /* NOTE NPT */

  /* init */
  void init_NPT_miller_symplectic_NHC(const bool & reset_thermostat, const bool & reset_barostat);

  /* evaluate thermo/baro-stat forces with respect to v(t), omega(t) and l(t) */
  void eval_NHC_barostat_forces();

  void eval_NPT_hamiltonian();

  /* step thermo/baro-stat variables. it is assumed that the current forces G
   *  are at time "t" while (v, omega, l) are at time "t + dt/2" */
  void step_NHC_barostat_new (const double & delta_t);

  /* stepper */
  void step_NPT_miller_symplectic_NHC();

  /***************************************************************************/

  /* evaluate the temperature */
  void eval_temperature();

  /* evaluate the pressure */
  void eval_pressure();

  /* evaluate mean and std of a double vector */
  void eval_stats_vec_double(std::vector<double> & vec, double & mean, double & std);

  /***************************************************************************/

  /* do an MD step */
  void step();

  /* force the conservation of energy through velocity scaling */
  void conservate_energy(const double& en0);

  /* fix CM location */
  void zero_CM_momentum();

  /* adjust the box with the new volume */
  void adjust_volume(const double& new_vol);

  /* logarithmic temperature drift */
  void adjust_log_temperature_drift();

  /* linear temperature drift */
  void adjust_lin_temperature_drift();

  /* linear concentration drift */
  void adjust_lin_concentration_drift();

  /* linear sphere radius drift */
  void adjust_lin_sph_radius_drift();

  /* brendsen thermostat step */
  void step_brendsen_thermostat(const double & dt);

  /***************************************************************************/

  /* logfile */
  std::ofstream& log_file();

  /* stats logfile */
  std::ofstream& stats_log_file();

  /* force on surfaces log file */
  std::ofstream& force_on_surfs_log_file();

  /* destructor */
  void free_memory();

}; /* dyn */

}; /* uni */

}; /* emd */

#endif /* DYNAMICS_H */
