/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef DYNAMICS_JOB_MANAGER_H
#define DYNAMICS_JOB_MANAGER_H

#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <vector>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>

#include "quaternion/gsl_quaternion.h"

#include "gslext.h"
#include "rigid_body.h"
#include "surface.h"
#include "paramdb.h"
#include "stringext.h"
#include "moluniverse.h"
#include "dynamics.h"
#include "universesaver.h"

/***************************************************************************/

#define DYNAMICS_JOB_MANAGER_DEBUG

/***************************************************************************/

namespace emd
{

namespace dynman
{

  /***************************************************************************/

  static bool            ready_flag               =  false;
  static bool            save_universe_snapshot   =  false;
  static bool            save_povray_snapshot     =  false;
  static bool            save_nhc_snapshot        =  false;
  static bool            save_output              =  false;
  static bool            nhc_file_supplied        =  false;
  //{{{restart: declaration of save_restart
  static bool            save_restart             =  false; 
  //}}}
  
  static long int        ignore_steps             =  0;
  static long int        steps                    =  1000;
  static long int        snapshot_freq            =  100;
  static long int        restart_save_freq        =  100;

  static std::string     pov_snapshot_seq_file;
  static std::string     uni_snapshot_seq_file;
  static std::string     universe_output_file;
  static std::string     nhc_file;
  //{{{restart:restart_file declaration
  static std::string     restart_file;
  //}}}
  /*** parser ****************************************************************/

  /* parses the input file */
  int load_file(const std::string & input_file_name);

  /* parse nhc vars file */
  int load_NHC_file(const std::string & input_file_name);

  /* parsers */
  int parse_universe_input_file(const std::string & param_string);
  int parse_universe_output_file(const std::string & param_string);
  int parse_log_file(const std::string & param_string);
  int parse_pov_snapshot_seq_file(const std::string & param_string);
  int parse_uni_snapshot_seq_file(const std::string & param_string);
  int parse_fix_cm(const std::string & param_string);
  int parse_save_universe_snapshot(const std::string & param_string);
  int parse_save_povray_snapshot(const std::string & param_string);
  int parse_integrator_type(const std::string & param_string);
  int parse_time_step(const std::string & param_string);
  int parse_ignore_steps(const std::string & param_string);
  int parse_steps(const std::string & param_string);
  int parse_goal_temperature(const std::string & param_string);
  int parse_goal_pressure(const std::string & param_string);
  int parse_nl_update_interval(const std::string & param_string);
  int parse_snapshot_freq(const std::string & param_string);
  int parse_log_to_file(const std::string & param_string);
  int parse_log_to_stdout(const std::string & param_string);
  int parse_log_force_on_surfs(const std::string & param_string);
  int parse_log_file_force_on_surfs(const std::string & param_string);
  int parse_re2_cutoff(const std::string & param_string);
  int parse_re2_skin(const std::string & param_string);
  int parse_gb_cutoff(const std::string & param_string);
  int parse_gb_skin(const std::string & param_string);
  int parse_ees_cutoff(const std::string & param_string);
  int parse_ees_skin(const std::string & param_string);
  int parse_save_output(const std::string & param_string);
  int parse_stats_log_file(const std::string & param_string);
  int parse_log_to_stats_file(const std::string & param_string);
  int parse_stats_interval(const std::string & param_string);
  int parse_nhc_file(const std::string & param_string);
  int parse_save_nhc_snapshot(const std::string & param_string);

  int parse_nhc_t_length(const std::string & param_string);
  int parse_nhc_p_length(const std::string & param_string);
  int parse_nhc_t_freq(const std::string & param_string);
  int parse_nhc_p_freq(const std::string & param_string);
  int parse_nhc_t_ksi_t(const std::string & param_string);
  int parse_nhc_t_ksi_r(const std::string & param_string);
  int parse_nhc_t_s_t(const std::string & param_string);
  int parse_nhc_t_s_r(const std::string & param_string);
  int parse_nhc_p_eps(const std::string & param_string);
  int parse_nhc_p_ksi_eps(const std::string & param_string);
  int parse_nhc_p_eta(const std::string & param_string);
  int parse_nhc_p_ksi_eta(const std::string & param_string);

  int parse_log_temperature_drift(const std::string & param_string);
  int parse_lin_temperature_drift(const std::string & param_string);
  int parse_lin_concentration_drift(const std::string & param_string);
  int parse_lin_sph_radius_drift(const std::string & param_string);
//{{{restart : declaration of parse_save_restart and parse_save_restart_freq and parse_restart_file
  int parse_save_restart(const std::string & param_string);
  int parse_restart_save_freq(const std::string & param_string);
  int parse_restart_file(const std::string & param_string);
//}}}
  
  /* actions */
  typedef int (*act)(const std::string & input_string);

  /* keywords */
  static std::string KEYWORDS[] =
  {
    "universe_input_file",
    "log_file",
    "pov_snapshot_seq_file",
    "uni_snapshot_seq_file",
    "universe_output_file",
    "fix_cm",
    "save_universe_snapshot",
    "save_povray_snapshot",
    "integrator_type",
    "time_step",
    "ignore_steps",
    "steps",
    "goal_temperature",
    "goal_pressure",
    "nl_update_interval",
    "snapshot_freq",
    "log_to_file",
    "log_to_stdout",
    "log_force_on_surfs",
    "log_file_force_on_surfs",
    "re2_cutoff",
    "re2_skin",
    "gb_cutoff",
    "gb_skin",
    "ees_cutoff",
    "ees_skin",
    "save_output",
    "stats_file",
    "log_to_stats_file",
    "stats_interval",
    "nhc_file",
    "save_nhc_snapshot",
    "nhc_t_length",
    "nhc_p_length",
    "nhc_t_freq",
    "nhc_p_freq",
    "nhc_t_ksi_t",
    "nhc_t_ksi_r",
    "nhc_t_s_t",
    "nhc_t_s_r",
    "nhc_p_eps",
    "nhc_p_ksi_eps",
    "nhc_p_eta",
    "nhc_p_ksi_eta",
    "log_temperature_drift",
    "lin_temperature_drift",
    "lin_concentration_drift",
    "lin_sph_radius_drift",
    //{{{restart: adding the fields "save_restart" and "restart_save_freq" and "restart_file"
    "save_restart",
    "save_restart_freq",
    "restart_file"
    //}}}
  };

  /* actions */
  static act ACTIONS[] =
  {
    &parse_universe_input_file,
    &parse_log_file,
    &parse_pov_snapshot_seq_file,
    &parse_uni_snapshot_seq_file,
    &parse_universe_output_file,
    &parse_fix_cm,
    &parse_save_universe_snapshot,
    &parse_save_povray_snapshot,
    &parse_integrator_type,
    &parse_time_step,
    &parse_ignore_steps,
    &parse_steps,
    &parse_goal_temperature,
    &parse_goal_pressure,
    &parse_nl_update_interval,
    &parse_snapshot_freq,
    &parse_log_to_file,
    &parse_log_to_stdout,
    &parse_log_force_on_surfs,
    &parse_log_file_force_on_surfs,
    &parse_re2_cutoff,
    &parse_re2_skin,
    &parse_gb_cutoff,
    &parse_gb_skin,
    &parse_ees_cutoff,
    &parse_ees_skin,
    &parse_save_output,
    &parse_stats_log_file,
    &parse_log_to_stats_file,
    &parse_stats_interval,
    &parse_nhc_file,
    &parse_save_nhc_snapshot,
    &parse_nhc_t_length,
    &parse_nhc_p_length,
    &parse_nhc_t_freq,
    &parse_nhc_p_freq,
    &parse_nhc_t_ksi_t,
    &parse_nhc_t_ksi_r,
    &parse_nhc_t_s_t,
    &parse_nhc_t_s_r,
    &parse_nhc_p_eps,
    &parse_nhc_p_ksi_eps,
    &parse_nhc_p_eta,
    &parse_nhc_p_ksi_eta,
    &parse_log_temperature_drift,
    &parse_lin_temperature_drift,
    &parse_lin_concentration_drift,
    &parse_lin_sph_radius_drift,
    //{{{restart : parse_restart_save_freq and parse_save_restart and parse_restart_file
    &parse_save_restart,
    &parse_restart_save_freq,
    &parse_restart_file
    //}}}
  };

  /***************************************************************************/

  /* init parser actions, variables, etc */
  int init();

  /* start simulation */
  void start();

  /* stop simulation, free memory */
  void stop();

}; /* dynman */

}; /* emd */

#endif /* DYNAMICS_JOB_MANAGER_H */
