/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdlib>
#include <sstream>
#include <cmath>
#include <ctime>

#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/path.hpp>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>

#include "consts.h"
#include "stringext.h"
#include "gslext.h"
#include "uniparser.h"
#include "rigid_body.h"
#include "moluniverse.h"
#include "universesaver.h"

#define DEF_WIDTH    30

using namespace std;
using namespace boost::filesystem;

/*** definitions ***********************************************************/

int duplicate_universe(const string & input_universe, const string & output_universe,
		const int & num);

int concat_universe(const string & input_universe_1, const string & input_universe_2,
		const string & output_universe, const int & axis_idx);

int add_droplet(const string & input_universe, const string & output_universe,
		const double & r_x, const double & r_y, const double & r_z, const double & rad,
		const double & pad, const int & surf_id);

int perturb_velocities(const string & input_universe, const string & output_universe,
		const double & v_amp, const double & omega_amp);
void get_path_name(char *inp_argv[] ,int inp_argc, int &ctr ,string &path_name );

/*** implementations *******************************************************/

int main(int argc, char *argv[])
{

	/* splash */
	cout << "Ellipsoids Molecular Dynamics [Tools Module] (" << VERSION_STRING << ")" << endl;
	cout << "Kopyleft (k) 2007 Mehrtash Babadi (mehrtash@physics.sharif.edu)" << endl << endl;

	int stat = true;

	int ctr = 1;
	while (stat && (ctr < argc))
	{
		if (string(argv[ctr])=="-param")
			if (argc>ctr)
			{
				ctr++;
				emd::uni::loader::set_ext_param_file(argv[ctr]);
			}
			else
				stat = false;

		if(string(argv[ctr])=="-pov")
		{
			if (argc>ctr)
			{
				string input_universe;
				string povray_output_file;
				ctr++;
				get_path_name(argv,argc,ctr,input_universe);
				povray_output_file = input_universe;
				povray_output_file.replace(povray_output_file.length()-3,3,"pov");
				
				if(argc>ctr+1)
				{
					if (string(argv[ctr+1])=="-pov_out")
					{
						ctr+=2;
						povray_output_file=argv[ctr];
					}
				}
				cout<<povray_output_file<<endl;
				emd::uni::load_universe(input_universe.c_str()  , false);
				emd::uni::saver::save_universe_to_povray_file(povray_output_file);
				emd::uni::free_memory();
			}
			else
				stat = false;
		}

		if (string(argv[ctr])=="-d")
			if (argc > (ctr+2) )
			{

				string  input_universe   = argv[++ctr];
				string  output_universe  = argv[++ctr];
				int     num              = atoi(argv[++ctr]);

				stat = (duplicate_universe(input_universe, output_universe, num)==EXIT_SUCCESS);

			}
			else
			{
				stat = false;
			}

		if (string(argv[ctr])=="-add_sphere")
			if (argc > (ctr+7) )
			{
				string  input_universe   = argv[++ctr];
				string  output_universe  = argv[++ctr];
				double  r_x              = atof(argv[++ctr]);
				double  r_y              = atof(argv[++ctr]);
				double  r_z              = atof(argv[++ctr]);
				double  rad              = atof(argv[++ctr]);
				double  pad              = atof(argv[++ctr]);
				int     id               = atoi(argv[++ctr]);

				stat = (add_droplet(input_universe, output_universe, r_x, r_y, r_z, rad, pad, id)==EXIT_SUCCESS);

			}
			else
			{
				stat = false;
			}

		if (string(argv[ctr])=="-pert_vel")
			if (argc > (ctr+3) )
			{
				string  input_universe   = argv[++ctr];
				string  output_universe  = argv[++ctr];
				double  v_amp            = atof(argv[++ctr]);
				double  omega_amp        = atof(argv[++ctr]);

				stat = (perturb_velocities(input_universe, output_universe, v_amp, omega_amp)==EXIT_SUCCESS);

			}
			else
			{
				stat = false;
			}

		if (string(argv[ctr])=="-concat_uni")
			if (argc > (ctr+3) )
			{
				string  input_universe_1   = argv[++ctr];
				string  input_universe_2   = argv[++ctr];
				string  output_universe    = argv[++ctr];
				int     axis_idx           = atoi(argv[++ctr]);

				stat = (concat_universe(input_universe_1, input_universe_2, output_universe, axis_idx)==EXIT_SUCCESS);

			}
			else
			{
				stat = false;
			}

		ctr++;

	}

	if (argc<=1) stat = false;

	if (!stat)
	{
		cout << "usage: emd-tools [OPTIONS]" << endl << endl
		<< "     -d <input_universe> <output_universe> <num>" << endl << endl
		<< "        duplicates the @input_universe, @num times along each side of a" << endl
		<< "        larger cube." << endl << endl
		<< "     -add_sphere <input_universe> <output_universe> <r_x> <r_y> <r_z> <rad> " << endl
		<< "                 <pad> <class_id>" << endl << endl
		<< "        add a spherical surface of type @class_id to the universe at" << endl
		<< "        <@r_x, @r_y, @r_z> with a radius @rad. also, remove the intersecting" << endl
		<< "        rigid bodies with padding @pad." << endl << endl
		<< "     -pert_vel <input_universe> <output_universe> <v_amp> <omga_amp>" << endl << endl
		<< "        perturn the linear and angular velocities by a linear random variable" << endl
		<< "         of half width @v_amp and @omega_amp." << endl << endl
		<< "     -concat_uni <input_universe_1> <input_universe_2> <output_universe>" << endl
		<< "                 <axis_idx>" << endl << endl
		<< "        concatenate @input_universe_1 and @input_universe_2 along @axis_idx." << endl
		<< "     -pov <input_universe> "<<endl<<endl
		<< "        convert uni to pov file "<<endl
		<< "     -param <input_universe> "<<endl<<endl
		<< "        change param file " <<endl
		<< endl;

		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;

}

/***********************************************************************/

int
duplicate_universe(const string & input_universe, const string & output_universe,
		const int & num)
{

	int fact = num*num*num;

	/* load the universe */
	if (emd::uni::load_universe(input_universe, false)!=EXIT_SUCCESS)
	{
		emd::uni::free_memory();
		return EXIT_FAILURE;
	}

	/* create new universe */
	/* NOTE only duplicates rigid bodies and spherical surfaces */

	emd::rigid_body *  new_rigid_bodies
	= new emd::rigid_body[fact*emd::uni::rigid_bodies_count];

	emd::spherical_int_surface *  new_sph_int_surfs
	= new emd::spherical_int_surface[fact*emd::uni::sph_int_surfs_count];

	unsigned int ctr_rb = 0;
	unsigned int ctr_si = 0;

	double sub_x = emd::uni::x_lower_bound + 0.5*num*emd::uni::x_width;
	double sub_y = emd::uni::y_lower_bound + 0.5*num*emd::uni::y_width;
	double sub_z = emd::uni::z_lower_bound + 0.5*num*emd::uni::z_width;

	gsl_vector * add_vec = gsl_vector_alloc(3);

	for (int nx = 0; nx<num; nx++)
		for (int ny = 0; ny<num; ny++)
			for (int nz = 0; nz<num; nz++)
			{

				gsl_vector_set(add_vec, 0, nx * emd::uni::x_width - sub_x);
				gsl_vector_set(add_vec, 1, ny * emd::uni::y_width - sub_y);
				gsl_vector_set(add_vec, 2, nz * emd::uni::z_width - sub_z);

				for (unsigned int m=0; m<emd::uni::rigid_bodies_count; m++)
				{
					new_rigid_bodies[ctr_rb]   = emd::uni::rigid_bodies[m];
					gsl_vector_add(new_rigid_bodies[ctr_rb].x, add_vec);
					ctr_rb++;
				}

				for (unsigned int s=0; s<emd::uni::sph_int_surfs_count; s++)
				{
					new_sph_int_surfs[ctr_si]   = emd::uni::sph_int_surfs[s];
					gsl_vector_add(new_sph_int_surfs[ctr_si].r0, add_vec);
					ctr_si++;
				}

			}

	if (emd::uni::rigid_bodies_count>0) delete[] emd::uni::rigid_bodies;
	if (emd::uni::sph_int_surfs_count>0) delete[] emd::uni::sph_int_surfs;

	emd::uni::rigid_bodies_count  *= fact;
	emd::uni::rigid_bodies         = new_rigid_bodies;

	emd::uni::sph_int_surfs_count *= fact;
	emd::uni::sph_int_surfs        = new_sph_int_surfs;

	emd::uni::x_lower_bound = - 0.5*num*emd::uni::x_width;
	emd::uni::x_upper_bound = 0.5*num*emd::uni::x_width;
	emd::uni::x_center      = 0;
	emd::uni::x_width       = num*emd::uni::x_width;

	emd::uni::y_lower_bound = - 0.5*num*emd::uni::y_width;
	emd::uni::y_upper_bound = 0.5*num*emd::uni::y_width;
	emd::uni::y_center      = 0;
	emd::uni::y_width       = num*emd::uni::y_width;

	emd::uni::z_lower_bound = - 0.5*num*emd::uni::z_width;
	emd::uni::z_upper_bound = 0.5*num*emd::uni::z_width;
	emd::uni::z_center      = 0;
	emd::uni::z_width       = num*emd::uni::z_width;

	/* save to .uni file */
	emd::uni::saver::save_universe_to_file(output_universe);

	/* free the memory */
	gsl_vector_free(add_vec);
	emd::uni::free_memory();

	cout << "* emd-tools::duplicate_universe: done!" << endl << endl;

	return EXIT_SUCCESS;

}

/***********************************************************************/

int
add_droplet(const string & input_universe, const string & output_universe,
		const double & r_x, const double & r_y, const double & r_z, const double & rad,
		const double & pad, const int & surf_id)
{

	/* load the universe */
	if (emd::uni::load_universe(input_universe, false)!=EXIT_SUCCESS)
	{
		emd::uni::free_memory();
		return EXIT_FAILURE;
	}

	gsl_vector * x0 = gsl_vector_alloc(3);
	gsl_vector * r0 = gsl_vector_alloc(3);

	gsl_vector_set(r0, 0, r_x);
	gsl_vector_set(r0, 1, r_y);
	gsl_vector_set(r0, 2, r_z);

	/* indices to be included */
	vector<emd::rigid_body_ref> indices;

	for (unsigned int m=0; m<emd::uni::rigid_bodies_count; m++)
	{

		gsl_vector_memcpy(x0, emd::uni::rigid_bodies[m].x);
		gsl_vector_sub(x0, r0);

		if (gsl_blas_dnrm2(x0) > (rad+pad))
		{
			emd::rigid_body_ref ref;

			ref.idx = m;

			ref.n_x = 0;
			ref.n_y = 0;
			ref.n_z = 0;

			indices.push_back(ref);
		}

	}

	/* add the spherical surface */
	emd::spherical_int_surface *  new_sph_int_surfs
	= new emd::spherical_int_surface[emd::uni::sph_int_surfs_count+1];

	/* copy */
	for (unsigned int s=0; s<emd::uni::sph_int_surfs_count; s++)
		new_sph_int_surfs[s]   = emd::uni::sph_int_surfs[s];

	/* the new one */
	gsl_vector_set(new_sph_int_surfs[emd::uni::sph_int_surfs_count].r0, 0, r_x);
	gsl_vector_set(new_sph_int_surfs[emd::uni::sph_int_surfs_count].r0, 1, r_y);
	gsl_vector_set(new_sph_int_surfs[emd::uni::sph_int_surfs_count].r0, 2, r_z);
	new_sph_int_surfs[emd::uni::sph_int_surfs_count].class_id = surf_id;
	new_sph_int_surfs[emd::uni::sph_int_surfs_count].rad = rad;

	/* replace */
	if (emd::uni::sph_int_surfs_count>0) delete[] emd::uni::sph_int_surfs;

	emd::uni::sph_int_surfs_count++;
	emd::uni::sph_int_surfs        = new_sph_int_surfs;

	emd::uni::saver::save_universe_to_file(indices, output_universe);

	emd::uni::free_memory();

	cout << "* emd-tools::add_droplet: done!" << endl << endl;

	return EXIT_SUCCESS;

}

/***********************************************************************/

int
perturb_velocities(const string & input_universe, const string & output_universe,
		const double & v_amp, const double & omega_amp)
{

	/* load the universe */
	if (emd::uni::load_universe(input_universe, false)!=EXIT_SUCCESS)
	{
		emd::uni::free_memory();
		return EXIT_FAILURE;
	}

	/* seed the random generator */
	srand48(time(0));

	for (unsigned int m=0; m<emd::uni::rigid_bodies_count; m++)
	{

		emd::rigid_body & c_m = emd::uni::rigid_bodies[m];

		gsl_vector_set(c_m.v, 0, gsl_vector_get(c_m.v, 0) + v_amp*(2*drand48()-1));
		gsl_vector_set(c_m.v, 1, gsl_vector_get(c_m.v, 1) + v_amp*(2*drand48()-1));
		gsl_vector_set(c_m.v, 2, gsl_vector_get(c_m.v, 2) + v_amp*(2*drand48()-1));

		c_m.omega.dat[0] += omega_amp*(2*drand48()-1);
		c_m.omega.dat[1] += omega_amp*(2*drand48()-1);
		c_m.omega.dat[2] += omega_amp*(2*drand48()-1);

	}

	emd::uni::saver::save_universe_to_file(output_universe);

	emd::uni::free_memory();

	cout << "* emd-tools::perturb_velocities: done!" << endl << endl;

	return EXIT_SUCCESS;

}

/***********************************************************************/

int
concat_universe(const string & input_universe_1, const string & input_universe_2,
		const string & output_universe, const int & axis_idx)
{

	unsigned int   rigid_bodies_count_1;
	unsigned int   sph_int_surfs_count_1;

	unsigned int   rigid_bodies_count_2;
	unsigned int   sph_int_surfs_count_2;

	double         x_lower_bound_1, x_upper_bound_1;
	double         y_lower_bound_1, y_upper_bound_1;
	double         z_lower_bound_1, z_upper_bound_1;

	double         x_lower_bound_2, x_upper_bound_2;
	double         y_lower_bound_2, y_upper_bound_2;
	double         z_lower_bound_2, z_upper_bound_2;

	/* load the universe */
	if (emd::uni::load_universe(input_universe_1, false)!=EXIT_SUCCESS)
	{
		emd::uni::free_memory();
		return EXIT_FAILURE;
	}

	/* keep the data universe 1 */
	emd::rigid_body *  new_rigid_bodies_1
	= new emd::rigid_body[emd::uni::rigid_bodies_count];

	emd::spherical_int_surface *  new_sph_int_surfs_1
	= new emd::spherical_int_surface[emd::uni::sph_int_surfs_count];

	for (unsigned int m=0; m<emd::uni::rigid_bodies_count; m++)
		new_rigid_bodies_1[m]    = emd::uni::rigid_bodies[m];

	for (unsigned int s=0; s<emd::uni::sph_int_surfs_count; s++)
		new_sph_int_surfs_1[s]   = emd::uni::sph_int_surfs[s];

	rigid_bodies_count_1   = emd::uni::rigid_bodies_count;
	sph_int_surfs_count_1  = emd::uni::sph_int_surfs_count;
	x_lower_bound_1        = emd::uni::x_lower_bound;
	x_upper_bound_1        = emd::uni::x_upper_bound;
	y_lower_bound_1        = emd::uni::y_lower_bound;
	y_upper_bound_1        = emd::uni::y_upper_bound;
	z_lower_bound_1        = emd::uni::z_lower_bound;
	z_upper_bound_1        = emd::uni::z_upper_bound;

	emd::uni::free_memory();

	/* load the universe */
	if (emd::uni::load_universe(input_universe_2, false)!=EXIT_SUCCESS)
	{
		emd::uni::free_memory();
		return EXIT_FAILURE;
	}

	/* keep the data universe 2 */
	emd::rigid_body *  new_rigid_bodies_2
	= new emd::rigid_body[emd::uni::rigid_bodies_count];

	emd::spherical_int_surface *  new_sph_int_surfs_2
	= new emd::spherical_int_surface[emd::uni::sph_int_surfs_count];

	for (unsigned int m=0; m<emd::uni::rigid_bodies_count; m++)
		new_rigid_bodies_2[m]    = emd::uni::rigid_bodies[m];

	for (unsigned int s=0; s<emd::uni::sph_int_surfs_count; s++)
		new_sph_int_surfs_2[s]   = emd::uni::sph_int_surfs[s];

	rigid_bodies_count_2   = emd::uni::rigid_bodies_count;
	sph_int_surfs_count_2  = emd::uni::sph_int_surfs_count;
	x_lower_bound_2        = emd::uni::x_lower_bound;
	x_upper_bound_2        = emd::uni::x_upper_bound;
	y_lower_bound_2        = emd::uni::y_lower_bound;
	y_upper_bound_2        = emd::uni::y_upper_bound;
	z_lower_bound_2        = emd::uni::z_lower_bound;
	z_upper_bound_2        = emd::uni::z_upper_bound;

	emd::uni::free_memory();

	/* create the concatenated universe */

	emd::rigid_body *  new_rigid_bodies_3
	= new emd::rigid_body[rigid_bodies_count_1 + rigid_bodies_count_2];

	emd::spherical_int_surface *  new_sph_int_surfs_3
	= new emd::spherical_int_surface[sph_int_surfs_count_1 + sph_int_surfs_count_2];

	gsl_vector * add_vec_1 = gsl_vector_calloc(3);
	gsl_vector * add_vec_2 = gsl_vector_calloc(3);

	switch (axis_idx)
	{
	case 0:
	{

		gsl_vector_set(add_vec_1, 0, - x_upper_bound_1);
		gsl_vector_set(add_vec_2, 0, - x_lower_bound_2);

		emd::uni::x_lower_bound = -0.5*(x_upper_bound_1 + x_upper_bound_2
				- x_lower_bound_1 - x_lower_bound_2);
		emd::uni::x_upper_bound = 0.5*(x_upper_bound_1 + x_upper_bound_2
				- x_lower_bound_1 - x_lower_bound_2);
		emd::uni::x_center      = 0;
		emd::uni::x_width       = (x_upper_bound_1 + x_upper_bound_2
				- x_lower_bound_1 - x_lower_bound_2);

		emd::uni::y_lower_bound = max(y_upper_bound_1, y_upper_bound_2);
		emd::uni::y_lower_bound = min(y_lower_bound_1, y_lower_bound_2);
		emd::uni::y_center      = 0.5*(emd::uni::y_lower_bound + emd::uni::y_upper_bound);
		emd::uni::y_width       = emd::uni::y_upper_bound - emd::uni::y_lower_bound;

		emd::uni::z_lower_bound = max(z_upper_bound_1, z_upper_bound_2);
		emd::uni::z_lower_bound = min(z_lower_bound_1, z_lower_bound_2);
		emd::uni::z_center      = 0.5*(emd::uni::z_lower_bound + emd::uni::z_upper_bound);
		emd::uni::z_width       = emd::uni::z_upper_bound - emd::uni::z_lower_bound;

		break;

	}

	case 1:
	{

		gsl_vector_set(add_vec_1, 1, - y_upper_bound_1);
		gsl_vector_set(add_vec_2, 1, - y_lower_bound_2);

		emd::uni::y_lower_bound = -0.5*(y_upper_bound_1 + y_upper_bound_2
				- y_lower_bound_1 - y_lower_bound_2);
		emd::uni::y_upper_bound = 0.5*(y_upper_bound_1 + y_upper_bound_2
				- y_lower_bound_1 - y_lower_bound_2);
		emd::uni::y_center      = 0;
		emd::uni::y_width       = (y_upper_bound_1 + y_upper_bound_2
				- y_lower_bound_1 - y_lower_bound_2);

		emd::uni::x_lower_bound = max(x_upper_bound_1, x_upper_bound_2);
		emd::uni::x_lower_bound = min(x_lower_bound_1, x_lower_bound_2);
		emd::uni::x_center      = 0.5*(emd::uni::x_lower_bound + emd::uni::x_upper_bound);
		emd::uni::x_width       = emd::uni::x_upper_bound - emd::uni::x_lower_bound;

		emd::uni::z_lower_bound = max(z_upper_bound_1, z_upper_bound_2);
		emd::uni::z_lower_bound = min(z_lower_bound_1, z_lower_bound_2);
		emd::uni::z_center      = 0.5*(emd::uni::z_lower_bound + emd::uni::z_upper_bound);
		emd::uni::z_width       = emd::uni::z_upper_bound - emd::uni::z_lower_bound;

		break;

	}

	case 2:
	{

		gsl_vector_set(add_vec_1, 2, - z_upper_bound_1);
		gsl_vector_set(add_vec_2, 2, - z_lower_bound_2);

		emd::uni::z_lower_bound = -0.5*(z_upper_bound_1 + z_upper_bound_2
				- z_lower_bound_1 - z_lower_bound_2);
		emd::uni::z_upper_bound = 0.5*(z_upper_bound_1 + z_upper_bound_2
				- z_lower_bound_1 - z_lower_bound_2);
		emd::uni::z_center      = 0;
		emd::uni::z_width       = (z_upper_bound_1 + z_upper_bound_2
				- z_lower_bound_1 - z_lower_bound_2);

		emd::uni::x_lower_bound = max(x_upper_bound_1, x_upper_bound_2);
		emd::uni::x_lower_bound = min(x_lower_bound_1, x_lower_bound_2);
		emd::uni::x_center      = 0.5*(emd::uni::x_lower_bound + emd::uni::x_upper_bound);
		emd::uni::x_width       = emd::uni::x_upper_bound - emd::uni::x_lower_bound;

		emd::uni::y_lower_bound = max(y_upper_bound_1, y_upper_bound_2);
		emd::uni::y_lower_bound = min(y_lower_bound_1, y_lower_bound_2);
		emd::uni::y_center      = 0.5*(emd::uni::y_lower_bound + emd::uni::y_upper_bound);
		emd::uni::y_width       = emd::uni::y_upper_bound - emd::uni::y_lower_bound;

		break;

	}

	}

	for (unsigned int m=0; m<rigid_bodies_count_1; m++)
	{
		new_rigid_bodies_3[m]    = new_rigid_bodies_1[m];
		gsl_vector_add(new_rigid_bodies_3[m].x, add_vec_1);
	}

	for (unsigned int s=0; s<sph_int_surfs_count_1; s++)
	{
		new_sph_int_surfs_3[s]   = new_sph_int_surfs_1[s];
		gsl_vector_add(new_sph_int_surfs_3[s].r0, add_vec_1);
	}

	for (unsigned int m=0; m<rigid_bodies_count_2; m++)
	{
		new_rigid_bodies_3[m+rigid_bodies_count_1]    = new_rigid_bodies_2[m];
		gsl_vector_add(new_rigid_bodies_3[m+rigid_bodies_count_1].x, add_vec_2);
	}

	for (unsigned int s=0; s<sph_int_surfs_count_2; s++)
	{
		new_sph_int_surfs_3[s+sph_int_surfs_count_1]   = new_sph_int_surfs_2[s];
		gsl_vector_add(new_sph_int_surfs_3[s+sph_int_surfs_count_1].r0, add_vec_2);
	}

	emd::uni::rigid_bodies_count   = rigid_bodies_count_1 + rigid_bodies_count_2;
	emd::uni::sph_int_surfs_count  = sph_int_surfs_count_1 + sph_int_surfs_count_2;

	emd::uni::rigid_bodies         = new_rigid_bodies_3;
	emd::uni::sph_int_surfs        = new_sph_int_surfs_3;

	/* save to .uni file */
	emd::uni::saver::save_universe_to_file(output_universe);

	/* free the memory */
	emd::uni::free_memory();

	delete[] new_rigid_bodies_1;
	delete[] new_rigid_bodies_2;
	delete[] new_sph_int_surfs_1;
	delete[] new_sph_int_surfs_2;

	gsl_vector_free(add_vec_1);
	gsl_vector_free(add_vec_2);

	cout << "* emd-tools::concat_universe: done!" << endl << endl;

	return EXIT_SUCCESS;

}


/***************************************************************************/
void get_path_name(char *inp_argv[] ,int inp_argc, int &ctr ,string &path_name )
{
	path_name = inp_argv[ctr];


	while((ctr+1 < inp_argc) && string(inp_argv[ctr+1]).find("-")!=0)
	{
		ctr++;
		path_name = path_name + " " + string(inp_argv[ctr]);  

	}
	
}
/***************************************************************************/
