/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston,x MA  02111-1307, USA.             *
 ***************************************************************************/

#include <geom.h>

using namespace std;

/***************************************************************************/

double
emd::geom::eval_area_polygon (const std::vector<point_3d> & vertices)
{

  int n_points = vertices.size()-1;

  double       areax, areay, areaz;
  double       rx, ry, rz;
  double       dx, dy, dz;

  areax = 0;
  areay = 0;
  areaz = 0;

  for (int i=0; i<n_points; i++)
  {


    rx = vertices[i].x;
    ry = vertices[i].y;
    rz = vertices[i].z;

    dx = vertices[i+1].x - vertices[i].x;
    dy = vertices[i+1].y - vertices[i].y;
    dz = vertices[i+1].z - vertices[i].z;

    areax += (ry*dz - rz*dy);
    areay += (rz*dx - rx*dz);
    areaz += (rx*dy - ry*dx);

  }

  areax *= 0.5;
  areay *= 0.5;
  areaz *= 0.5;

  return sqrt(areax*areax + areay*areay + areaz*areaz);

};


/***************************************************************************/

int
emd::geom::find_vertex_points (const std::vector<plane_3d> & planes_0,
    const point_3d & int_point, std::vector<point_3d> & points)
{

  const double               DET_TOL             = 1.0e-6;
  const double               ORTHO_TOL           = 1.0e-6;
  const double               SAME_TOL            = 1.0e-6;

  int                        stat                = EXIT_SUCCESS;

  int                        n_planes;
  gsl_vector *               x                   = gsl_vector_alloc(3);

  /* phase 0: purging planes within the range of SAME_TOL */

  vector<plane_3d>           planes;

  for (int i=0; i<planes_0.size(); i++)
  {
    bool distinct_plane = true;

    for (int j=0; j<planes.size(); j++)
    {

      if ( (abs(planes[j].nx - planes_0[i].nx) < SAME_TOL) && 
           (abs(planes[j].ny - planes_0[i].ny) < SAME_TOL) && 
           (abs(planes[j].nz - planes_0[i].nz) < SAME_TOL) && 
           (abs(planes[j].d - planes_0[i].d) < SAME_TOL) )
        distinct_plane = false;

      if ( (abs(planes[j].nx + planes_0[i].nx) < SAME_TOL) && 
           (abs(planes[j].ny + planes_0[i].ny) < SAME_TOL) && 
           (abs(planes[j].nz + planes_0[i].nz) < SAME_TOL) && 
           (abs(planes[j].d + planes_0[i].d) < SAME_TOL) )
        distinct_plane = false;

    }

    if (distinct_plane)
      planes.push_back(planes_0[i]);

  }

  n_planes = planes.size();

  /* phase 1: finding all intersection points, exluding masked points */

  double e_x, e_y, e_z, len1, len2, dot;

  for (int i=0; i<n_planes-2; i++)
    for (int j=i+1; j<n_planes-1; j++)
      for (int k=j+1; k<n_planes; k++)
      {

        /* setup the linear system */
        double m_data[] = { planes[i].nx, planes[i].ny, planes[i].nz,
                            planes[j].nx, planes[j].ny, planes[j].nz,
                            planes[k].nx, planes[k].ny, planes[k].nz };

        double d_data[] = { planes[i].d,  planes[j].d,  planes[k].d };

        gsl_matrix_view   m = gsl_matrix_view_array(m_data, 3, 3);

        /* check the determinant */
        if ( abs(gsl_3matrix_det(&m.matrix)) < DET_TOL ) continue;

        gsl_vector_view   d = gsl_vector_view_array(d_data, 3);

        /* solve the linear system */
        int               s;
        gsl_permutation * p = gsl_permutation_alloc(3);
        gsl_linalg_LU_decomp (&m.matrix, p, &s);
        gsl_linalg_LU_solve (&m.matrix, p, &d.vector, x);

        point_3d          new_vertex;
        new_vertex.x      = gsl_vector_get(x, 0);
        new_vertex.y      = gsl_vector_get(x, 1);
        new_vertex.z      = gsl_vector_get(x, 2);

        e_x  = new_vertex.x - int_point.x;
        e_y  = new_vertex.y - int_point.y;
        e_z  = new_vertex.z - int_point.z;
        len1 = sqrt(e_x*e_x + e_y*e_y + e_z*e_z);
        e_x /= len1;
        e_y /= len1;
        e_z /= len1;

        double is_masked = false;

        for (int l=0; l<planes.size(); l++)
        {

          if (l==i || l==j || l==k) continue;

          dot = e_x*planes[l].nx + e_y*planes[l].ny + e_z*planes[l].nz;

          /* if the line and the plane are parallel, continue */
          if ( abs(dot) < ORTHO_TOL ) continue;

          /* otherwise */
          len2 =
            (planes[l].d - planes[l].nx*int_point.x - planes[l].ny*int_point.y
             - planes[l].nz*int_point.z) / dot;

          /* there is a masking plane */
          if ( (len2 > 0) && ( (len1-len2) > SAME_TOL ) )
            is_masked = true;

        }

        if (!is_masked)
        {

          /* if it is a new vertex, add a new element. otherwise, update the
           * previously added point */

          bool is_distinct = true;
          for (int l=0; l<points.size(); l++)
          {
            double dist
              = sqrt( gsl_pow_2( points[l].x - new_vertex.x ) +
                      gsl_pow_2( points[l].y - new_vertex.y ) +
                      gsl_pow_2( points[l].z - new_vertex.z ) );
            if (dist < SAME_TOL)
            {
              is_distinct = false;
              break;
            }
          }

          if (is_distinct)
            points.push_back(new_vertex);

        }

      }

  gsl_vector_free(x);

  return stat;

};

/***************************************************************************/

int
emd::geom::eval_interior_volume (const std::vector<plane_3d> & planes_0,
  const point_3d & int_point, double & volume)
{

  vector<point_3d> *       faces;
  int                      n_faces = 0;
  vector<int>              tastes;

  int stat = eval_interior_volume( planes_0, int_point, &faces, tastes, n_faces, volume );

  if (n_faces > 0)
    delete[] faces;

  return stat;

}

/***************************************************************************/

int
emd::geom::eval_interior_volume (const std::vector<plane_3d> & planes_0,
  const point_3d & int_point, std::vector<point_3d> ** out_faces,
  vector<int> & tastes, int & n_faces, double & volume)
{

  const double               DET_TOL             = 1.0e-6;
  const double               ORTHO_TOL           = 1.0e-6;
  const double               SAME_TOL            = 1.0e-6;

  int                        stat                = EXIT_SUCCESS;

  int                        n_planes;
  gsl_vector *               x                   = gsl_vector_alloc(3);
  vector<edge_3d> *          faces;
  vector<vertex_point_3d>    points;

  /* phase 0: purging planes within the range of SAME_TOL */

  vector<plane_3d>           planes;

  for (int i=0; i<planes_0.size(); i++)
  {
    bool distinct_plane = true;

    for (int j=0; j<planes.size(); j++)
    {

      if ( (abs(planes[j].nx - planes_0[i].nx) < SAME_TOL) && 
           (abs(planes[j].ny - planes_0[i].ny) < SAME_TOL) && 
           (abs(planes[j].nz - planes_0[i].nz) < SAME_TOL) && 
           (abs(planes[j].d - planes_0[i].d) < SAME_TOL) )
        distinct_plane = false;

      if ( (abs(planes[j].nx + planes_0[i].nx) < SAME_TOL) && 
           (abs(planes[j].ny + planes_0[i].ny) < SAME_TOL) && 
           (abs(planes[j].nz + planes_0[i].nz) < SAME_TOL) && 
           (abs(planes[j].d + planes_0[i].d) < SAME_TOL) )
        distinct_plane = false;

    }

    if (distinct_plane)
    {
      planes.push_back(planes_0[i]);
      tastes.push_back(planes_0[i].taste);
    }

  }

  n_planes = planes.size();
  faces    = new vector<edge_3d>[n_planes];

  /* phase 1: finding all intersection points, exluding masked points */

  double e_x, e_y, e_z, len1, len2, dot;

  for (int i=0; i<n_planes-2; i++)
    for (int j=i+1; j<n_planes-1; j++)
      for (int k=j+1; k<n_planes; k++)
      {

        /* setup the linear system */
        double m_data[] = { planes[i].nx, planes[i].ny, planes[i].nz,
                            planes[j].nx, planes[j].ny, planes[j].nz,
                            planes[k].nx, planes[k].ny, planes[k].nz };

        double d_data[] = { planes[i].d,  planes[j].d,  planes[k].d };

        gsl_matrix_view   m = gsl_matrix_view_array(m_data, 3, 3);

        /* check the determinant */
        if ( abs(gsl_3matrix_det(&m.matrix)) < DET_TOL ) continue;

        gsl_vector_view   d = gsl_vector_view_array(d_data, 3);

        /* solve the linear system */
        int               s;
        gsl_permutation * p = gsl_permutation_alloc(3);
        gsl_linalg_LU_decomp (&m.matrix, p, &s);
        gsl_linalg_LU_solve (&m.matrix, p, &d.vector, x);

        vertex_point_3d  new_vertex;
        new_vertex.x      = gsl_vector_get(x, 0);
        new_vertex.y      = gsl_vector_get(x, 1);
        new_vertex.z      = gsl_vector_get(x, 2);

        e_x  = new_vertex.x - int_point.x;
        e_y  = new_vertex.y - int_point.y;
        e_z  = new_vertex.z - int_point.z;
        len1 = sqrt(e_x*e_x + e_y*e_y + e_z*e_z);
        e_x /= len1;
        e_y /= len1;
        e_z /= len1;

        double is_masked = false;

        for (int l=0; l<planes.size(); l++)
        {

          if (l==i || l==j || l==k) continue;

          dot = e_x*planes[l].nx + e_y*planes[l].ny + e_z*planes[l].nz;

          /* if the line and the plane are parallel, continue */
          if ( abs(dot) < ORTHO_TOL ) continue;

          /* otherwise */
          len2 =
            (planes[l].d - planes[l].nx*int_point.x - planes[l].ny*int_point.y
             - planes[l].nz*int_point.z) / dot;

          /* there is a masking plane */
          if ( (len2 > 0) && ( (len1-len2) > SAME_TOL ) )
            is_masked = true;

        }

        if (!is_masked)
        {

          /* if it is a new vertex, add a new element. otherwise, update the
           * previously added point */

          bool is_distinct = true;
          for (int l=0; l<points.size(); l++)
          {
            double dist
              = sqrt( gsl_pow_2( points[l].x - new_vertex.x ) +
                      gsl_pow_2( points[l].y - new_vertex.y ) +
                      gsl_pow_2( points[l].z - new_vertex.z ) );
            if (dist < SAME_TOL)
            {
              is_distinct = false;
              points[l].planes.insert(i);
              points[l].planes.insert(j);
              points[l].planes.insert(k);
              break;
            }
          }

          if (is_distinct)
          {
            new_vertex.planes.insert(i);
            new_vertex.planes.insert(j);
            new_vertex.planes.insert(k);
            points.push_back(new_vertex);
          }

        }

      }


//    for (int i=0; i<points.size(); i++)
//    {
//      cout << points[i].x << "  " << points[i].y << "  " << points[i].z << "          ";
//      for (set<int>::iterator it=points[i].planes.begin(); it!=points[i].planes.end(); it++)
//        cout << (*it) << "  ";
//      cout << endl;
//    }

  /* phase 2 finding edges */

  if (points.size() < 4)
    stat = EXIT_FAILURE;

  if (stat == EXIT_SUCCESS)

    for (int i=0; i<points.size()-1; i++)
      for (int j=i+1; j<points.size(); j++)
      {

        /* find the intersection of the sets points[i].planes and points[j].planes */
        vector<int> shared_planes;
        for (set<int>::iterator it_i = points[i].planes.begin(); it_i != points[i].planes.end(); it_i++)
          for (set<int>::iterator it_j = points[j].planes.begin(); it_j != points[j].planes.end(); it_j++)
            if ( *it_i == *it_j ) shared_planes.push_back(*it_i);

        if (shared_planes.size() < 2) continue;

        edge_3d new_edge = {i, j};

        for (vector<int>::iterator it = shared_planes.begin(); it != shared_planes.end(); it++)
          faces[*it].push_back(new_edge);

      }

#ifdef CHECK_POLYHEDRON_CONDITION

  /* phase 2.1: (optional) checking Euler's condition */

  if (stat==EXIT_SUCCESS)
  {

    n_faces    = 0;
    int n_edges    = 0,
        n_vertices = 0;

    for (int i=0; i<n_planes; i++)
    {
      if (faces[i].size() <= 2) continue;
      n_faces++;
      n_edges += faces[i].size();
    }
    n_edges /= 2;

    n_vertices = points.size();

    if (n_faces + n_vertices - n_edges != 2)
      stat = EXIT_FAILURE;

  }

#endif


  /* phase 3: creating polygons and evaluating volume */


  if (stat==EXIT_SUCCESS)
  {

    (*out_faces) = new vector<point_3d>[n_planes];

    volume = 0;

    for (int i=0; i<n_planes; i++)
    {

      if (faces[i].size() <= 2) continue;

      vector<point_3d>  polygon_points;
      point_3d          new_point;
      int               start_point, search_point;

      /* push the first point of the first edge */
      new_point.x = points[faces[i][0].idx1].x;
      new_point.y = points[faces[i][0].idx1].y;
      new_point.z = points[faces[i][0].idx1].z;
      polygon_points.push_back(new_point);
      search_point = faces[i][0].idx2;
      start_point  = faces[i][0].idx1;
      faces[i].erase(faces[i].begin());

      while (!faces[i].empty())
      {

        bool found = false;

        for (vector<edge_3d>::iterator it = faces[i].begin(); it!=faces[i].end(); it++)
        {

          if ( (*it).idx1==search_point )
          {
            new_point.x = points[search_point].x;
            new_point.y = points[search_point].y;
            new_point.z = points[search_point].z;
            polygon_points.push_back(new_point);
            search_point = (*it).idx2;
            if ( (faces[i].size()==1) && (search_point!=start_point) ) stat = EXIT_FAILURE;
            faces[i].erase(it);
            found = true;
            break;
          }

          if ( (*it).idx2==search_point )
          {
            new_point.x = points[search_point].x;
            new_point.y = points[search_point].y;
            new_point.z = points[search_point].z;
            polygon_points.push_back(new_point);
            search_point = (*it).idx1;
            if ( (faces[i].size()==1) && (search_point!=start_point) ) stat = EXIT_FAILURE;
            faces[i].erase(it);
            found = true;
            break;
          }

        }

        if ( (stat==EXIT_FAILURE) || !found )
        {
          stat = EXIT_FAILURE;
          break;
        }

      }

      if (stat==EXIT_FAILURE) break;

      /* add the starting point to the end of the list */
      new_point.x = points[start_point].x;
      new_point.y = points[start_point].y;
      new_point.z = points[start_point].z;
      polygon_points.push_back(new_point);

      (*out_faces)[i] = polygon_points;

//       for (int q=0; q<polygon_points.size(); q++)
//         cout << polygon_points[q].x << "  " << polygon_points[q].y << "  " << polygon_points[q].z << endl;
//       cout << "-------" << endl;

      /* evaluate normal distance from the plane */
      double normal_dist = abs( planes[i].d - planes[i].nx*int_point.x
        - planes[i].ny*int_point.y - planes[i].nz*int_point.z );

      volume += ( normal_dist*eval_area_polygon(polygon_points) / 3 );

    }

  }

  gsl_vector_free(x);

  delete[] faces;

  return stat;

};
