/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef GEOM_H
#define GEOM_H

#include <iostream>
#include <cmath>
#include <vector>
#include <set>

#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_linalg.h>
#include <gslext.h>

namespace emd
{

namespace geom
{

  /*************************************************************************/

  struct point_3d
  {
    double      x, y, z;        /* coordinates */
  };

  /*************************************************************************/

  struct edge_3d
  {
    int         idx1, idx2;     /* indices of the vertex endpoints */
  };

  /*************************************************************************/

  struct vertex_point_3d
  {
    double          x, y, z;        /* coordinates */
    std::set<int>   planes;         /* indices of the intersecting planes */
  };

  /*************************************************************************/

  /* a plane is described like: nx*x + ny*y + nz*z = d */
  struct plane_3d
  {
    double      nx, ny, nz;     /* normal unit vector */
    double      d;              /* distance from origin */
    int         taste;          /* ! */
  };

  /*************************************************************************/

  /*************************************************************************
   * evaluates the surface area of a polygon. The polygon is described     *
   * by a vector of point_3d's. The points are not assumed to be coplanar. *
   * The return value is the norm of the vectorial area. The initial point *
   * MUST appear at the end again:                                         *
   *                                                                       *
   * P_0,  P_1,  P_2, ...,  P_(n-1),  P_0                                  *
   *************************************************************************/

  double
  eval_area_polygon (const std::vector<point_3d> & vertices);


  /*************************************************************************
   * evaluate the interior volume of a finite set of 3D intersecting       *
   * planes. @int_point is a given interior point.                         *
   *                                                                       *
   * The evaluated volume will be copied into @volume with the return      *
   * value of EXIT_SUCCESS in case of success. Otherwise, EXIT_FAILURE     *
   * will be returned, leaving @volume unchanged.                          *
   *************************************************************************/

#define CHECK_POLYHEDRON_CONDITION

  int
  find_vertex_points (const std::vector<plane_3d> & planes_0,
    const point_3d & int_point, std::vector<point_3d> & points);

  int
  eval_interior_volume (const std::vector<plane_3d> & planes_0,
    const point_3d & int_point, double & volume);

  int
  eval_interior_volume (const std::vector<plane_3d> & planes_0,
    const point_3d & int_point, std::vector<point_3d> ** out_faces,
    std::vector<int> & tastes, int & n_faces, double & volume);

}; /* geom */

}; /* emd */

#endif /* GEOM_H */
