/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <gslext.h>

using namespace std;

/* fast inversion of 3-matrices */
int gsl_3matrix_invert(const gsl_matrix * A, gsl_matrix * Ainv)
{

  double det = gsl_3matrix_det(A);

  gsl_matrix_set(Ainv, 0, 0,
    (gsl_matrix_get(A, 1, 1)*gsl_matrix_get(A, 2, 2)-
     gsl_matrix_get(A, 1, 2)*gsl_matrix_get(A, 2, 1))/det);

  gsl_matrix_set(Ainv, 0, 1,
    (gsl_matrix_get(A, 0, 2)*gsl_matrix_get(A, 2, 1)-
     gsl_matrix_get(A, 0, 1)*gsl_matrix_get(A, 2, 2))/det);

  gsl_matrix_set(Ainv, 0, 2,
    (gsl_matrix_get(A, 0, 1)*gsl_matrix_get(A, 1, 2)-
     gsl_matrix_get(A, 1, 1)*gsl_matrix_get(A, 0, 2))/det);

  gsl_matrix_set(Ainv, 1, 0,
    (gsl_matrix_get(A, 1, 2)*gsl_matrix_get(A, 2, 0)-
     gsl_matrix_get(A, 1, 0)*gsl_matrix_get(A, 2, 2))/det);

  gsl_matrix_set(Ainv, 1, 1,
    (gsl_matrix_get(A, 0, 0)*gsl_matrix_get(A, 2, 2)-
     gsl_matrix_get(A, 0, 2)*gsl_matrix_get(A, 2, 0))/det);

  gsl_matrix_set(Ainv, 1, 2,
    (gsl_matrix_get(A, 0, 2)*gsl_matrix_get(A, 1, 0)-
     gsl_matrix_get(A, 0, 0)*gsl_matrix_get(A, 1, 2))/det);

  gsl_matrix_set(Ainv, 2, 0,
    (gsl_matrix_get(A, 1, 0)*gsl_matrix_get(A, 2, 1)-
     gsl_matrix_get(A, 1, 1)*gsl_matrix_get(A, 2, 0))/det);

  gsl_matrix_set(Ainv, 2, 1,
    (gsl_matrix_get(A, 0, 1)*gsl_matrix_get(A, 2, 0)-
     gsl_matrix_get(A, 0, 0)*gsl_matrix_get(A, 2, 1))/det);

  gsl_matrix_set(Ainv, 2, 2,
    (gsl_matrix_get(A, 1, 1)*gsl_matrix_get(A, 0, 0)-
     gsl_matrix_get(A, 0, 1)*gsl_matrix_get(A, 1, 0))/det);

  return 0;

}

/* determinant of a 3-matrix */
double gsl_3matrix_det(const gsl_matrix * A)
{

  return -gsl_matrix_get(A, 0, 2)*gsl_matrix_get(A, 1, 1)*gsl_matrix_get(A, 2, 0) +
         gsl_matrix_get(A, 0, 1)*gsl_matrix_get(A, 1, 2)*gsl_matrix_get(A, 2, 0) +
         gsl_matrix_get(A, 0, 2)*gsl_matrix_get(A, 1, 0)*gsl_matrix_get(A, 2, 1) -
         gsl_matrix_get(A, 0, 0)*gsl_matrix_get(A, 1, 2)*gsl_matrix_get(A, 2, 1) -
         gsl_matrix_get(A, 0, 1)*gsl_matrix_get(A, 1, 0)*gsl_matrix_get(A, 2, 2) +
         gsl_matrix_get(A, 0, 0)*gsl_matrix_get(A, 1, 1)*gsl_matrix_get(A, 2, 2);

}

/* initializes a 3x3 matrix to an SO(3) matrix with respet
   to the provided Euler angles */
int gsl_matrix_euler_rot(const double & theta, const double & phi,
  const double & psi, gsl_matrix * dest)
{

  if (dest->size1!=3 || dest->size2!=3)
  {
    cout << "gsl_matrix_euler_rot: 3x3 matrix is required!" << endl;
    return 1;
  }

  gsl_matrix_set(dest, 0, 0, cos(theta)*cos(psi));
  gsl_matrix_set(dest, 0, 1, -cos(phi)*sin(psi)+sin(phi)*sin(theta)*cos(psi));
  gsl_matrix_set(dest, 0, 2, sin(phi)*sin(psi)+cos(phi)*sin(theta)*cos(psi));
  gsl_matrix_set(dest, 1, 0, cos(theta)*sin(psi));
  gsl_matrix_set(dest, 1, 1, cos(phi)*cos(psi)+sin(phi)*sin(theta)*sin(psi));
  gsl_matrix_set(dest, 1, 2, -sin(phi)*cos(psi)+cos(phi)*sin(theta)*sin(psi));
  gsl_matrix_set(dest, 2, 0, -sin(theta));
  gsl_matrix_set(dest, 2, 1, sin(phi)*cos(theta));
  gsl_matrix_set(dest, 2, 2, cos(phi)*cos(theta));

  return EXIT_SUCCESS;

}

/* SO(3) rot. generator with respect to x */
int gsl_matrix_rot_x(gsl_matrix * dest)
{

  if (dest->size1!=3 || dest->size2!=3)
  {
    cout << "gsl_matrix_euler_rot: 3x3 matrix is required!" << endl;
    return 1;
  }

  gsl_matrix_set_zero(dest);
  gsl_matrix_set(dest, 2, 1, 1);
  gsl_matrix_set(dest, 1, 2, -1);

  return 0;

}

/* SO(3) rot. generator with respect to y */
int gsl_matrix_rot_y(gsl_matrix * dest)
{

  if (dest->size1!=3 || dest->size2!=3)
  {
    cout << "gsl_matrix_euler_rot: 3x3 matrix is required!" << endl;
    return 1;
  }

  gsl_matrix_set_zero(dest);
  gsl_matrix_set(dest, 0, 2, 1);
  gsl_matrix_set(dest, 2, 0, -1);

  return 0;

}

/* SO(3) rot. generator with respect to z */
int gsl_matrix_rot_z(gsl_matrix * dest)
{

  if (dest->size1!=3 || dest->size2!=3)
  {
    cout << "gsl_matrix_euler_rot: 3x3 matrix is required!" << endl;
    return 1;
  }

  gsl_matrix_set_zero(dest);
  gsl_matrix_set(dest, 1, 0, 1);
  gsl_matrix_set(dest, 0, 1, -1);

  return 0;

}

/* rotate a vector about x axis */
int gsl_vector_rot_x(const double & theta, const gsl_vector * src, gsl_vector * dest)
{

  double cos_term = cos(theta),
         sin_term = sin(theta);

  gsl_vector_set( dest, 0, gsl_vector_get(src, 0) );
  gsl_vector_set( dest, 1, gsl_vector_get(src, 1) * cos_term - gsl_vector_get(src, 2) * sin_term );
  gsl_vector_set( dest, 2, gsl_vector_get(src, 1) * sin_term + gsl_vector_get(src, 2) * cos_term );

  return 0;

}

/* rotate a vector about x axis */
int gsl_vector_rot_y(const double & theta, const gsl_vector * src, gsl_vector * dest)
{

  double cos_term = cos(theta),
         sin_term = sin(theta);

  gsl_vector_set( dest, 1, gsl_vector_get(src, 1) );
  gsl_vector_set( dest, 2, gsl_vector_get(src, 2) * cos_term - gsl_vector_get(src, 0) * sin_term );
  gsl_vector_set( dest, 0, gsl_vector_get(src, 2) * sin_term + gsl_vector_get(src, 0) * cos_term );

  return 0;

}

/* rotate a vector about x axis */
int gsl_vector_rot_z(const double & theta, const gsl_vector * src, gsl_vector * dest)
{

  double cos_term = cos(theta),
         sin_term = sin(theta);

  gsl_vector_set( dest, 2, gsl_vector_get(src, 2) );
  gsl_vector_set( dest, 0, gsl_vector_get(src, 0) * cos_term - gsl_vector_get(src, 1) * sin_term );
  gsl_vector_set( dest, 1, gsl_vector_get(src, 0) * sin_term + gsl_vector_get(src, 1) * cos_term );

  return 0;

}

/* normalize a vector */
int gsl_vector_normalize(gsl_vector * vec)
{

  gsl_vector_scale(vec, 1/gsl_blas_dnrm2(vec));

  return 0;

}

ostream & operator<<(ostream & os, const gsl_matrix & A)
{

  for (int i=0; i<A.size1; i++)
  {
    for (int j=0; j<A.size2; j++)
      os << "(" << i << "," << j << ") = " << setw(10)
         << gsl_matrix_get(&A, i, j) << "  ";
    os << endl;
  }

  return os;

}

ostream & operator<<(ostream & os, const gsl_vector & v)
{

  for (int i=0; i<v.size; i++)
    os << "(" << i << ") = " << gsl_vector_get(&v, i) << endl;

  return os;

}

ostream & operator<<(ostream & os, const gsl_quaternion & q)
{

  os << "Re= " << GSL_QUATERNION_REAL(q);
  os << "  Im1= " << GSL_QUATERNION_IMAG1(q);
  os << "  Im2= " << GSL_QUATERNION_IMAG2(q);
  os << "  Im3= " << GSL_QUATERNION_IMAG3(q) << endl;

  return os;

}
