/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <int_ees_comp.h>

using namespace std;

/***************************************************************************/

double   emd::inter::ees::UA;
double   emd::inter::ees::UR;

gsl_vector *       emd::inter::ees::e_x       = gsl_vector_calloc(3);
gsl_vector *       emd::inter::ees::e_y       = gsl_vector_calloc(3);
gsl_vector *       emd::inter::ees::e_z       = gsl_vector_calloc(3);

gsl_matrix *       emd::inter::ees::R_x       = gsl_matrix_calloc(3, 3);
gsl_matrix *       emd::inter::ees::R_y       = gsl_matrix_calloc(3, 3);
gsl_matrix *       emd::inter::ees::R_z       = gsl_matrix_calloc(3, 3);

/***************************************************************************/

void
emd::inter::ees::init ()
{

  gsl_vector_set_basis(e_x, 0);
  gsl_vector_set_basis(e_y, 1);
  gsl_vector_set_basis(e_z, 2);

  gsl_matrix_rot_x(R_x);
  gsl_matrix_rot_y(R_y);
  gsl_matrix_rot_z(R_z);

  init_flag = true;

}

/***************************************************************************/

bool
emd::inter::ees::is_init ()
{
  return init_flag;
}

/***************************************************************************/

int
emd::inter::ees::free_memory ()
{

  gsl_vector_free(e_x);
  gsl_vector_free(e_y);
  gsl_vector_free(e_z);

  gsl_matrix_free(R_x);
  gsl_matrix_free(R_y);
  gsl_matrix_free(R_z);

}


/***************************************************************************/

double
emd::inter::ees::get_contact_distance (const gsl_matrix * A0, const gsl_vector * nhat0,
  const double & h0, const paramdb::param_cache_EES_int * param)
{

  /* evaluate nhatx, sn, en */
  gsl_blas_dgemv(CblasTrans,    1,  A0,         nhat0, 0,  dvec);
  gsl_blas_dgemv(CblasNoTrans,  1,  param->S,   dvec,  0,  ddvec);

  return ( h0 - gsl_blas_dnrm2(ddvec) );

}

/***************************************************************************/

double
emd::inter::ees::get_EES_int_energy (const gsl_matrix * A0, const gsl_vector * nhat0,
  const double & h0, const paramdb::param_cache_EES_int * param, const double & r_c,
  bool & eval_flag)
{

  /* A is set to the transposed of the provided A0 */
  gsl_matrix_transpose_memcpy(A, A0);

  /* copy nhat */
  gsl_vector_memcpy(nhat, nhat0);

  /* copy h */
  h = h0;

  /* copye trace(E) */
  trE = param->trE;

  /* evaluate nhatx, sn, en */
  gsl_blas_dgemv(CblasNoTrans, 1, A, nhat, 0, dvec);
  gsl_blas_dgemv(CblasNoTrans, 1, param->S, dvec, 0, nhatx);

  sn = gsl_blas_dnrm2(nhatx);

  /* ignore the interaction if it is outside of the cutoff distance */
  if ( h > (sn + r_c) )
  {
    eval_flag = false;
    return 0;
  }

  /* nhatx = (1/sn) |S A nhat> */
  gsl_vector_scale(nhatx, 1/sn);

  /* Enhat = E |nhatx> */
  gsl_blas_dgemv(CblasNoTrans, 1, param->E, nhatx, 0, Enhatx);

  /* en = <nhatx|E|nhatx> */
  gsl_blas_ddot(nhatx, Enhatx, &en);

  /* evaluate h and sn powers */
  h2 = h*h;       sn2 = sn*sn;
  h3 = h*h2;      sn3 = sn*sn2;
  h4 = h2*h2;     sn4 = sn2*sn2;
  h5 = h2*h3;     sn5 = sn2*sn3;
  h6 = h3*h3;     sn6 = sn3*sn3;

  lcd_term  =  h2 - sn2;
  lcd_term2 =  lcd_term*lcd_term;
  log_term  =  log( (h-sn) / (h+sn) );

  A_prf     =  param->A_fact / ( 24 * sn5 * lcd_term );
  R_prf     =  2.0 * param->R_fact / ( 4725 * gsl_pow_7 ( lcd_term ) );

  UA        =  - A_prf * ( 6*(5*en - trE)*sn*h3 - 2*(13*en - 3*trE - 2)*sn3*h
               + lcd_term*( 3*(5*en - trE)*h2 - (3*en - trE - 2)*sn2 )*log_term );

  UR        =  R_prf * ( 7*(trE + 5)*h5 + 2*(45*en - 2*trE + 35)*sn2*h3 + 3*(10*en
               - trE + 5)*sn4*h );

  eval_flag = true;

  return (UA + UR);

}

/***************************************************************************/

double
emd::inter::ees::get_dh_UA()
{

  return A_prf*2*( 2*sn*( 3*trE*h4 - 5*sn2*trE*h2 + en*(-15*h4 + 25*sn2*h2 - 8*sn4)
         + 2*sn4*(trE + 1) ) - 3*h*lcd_term2*(5*en - trE)*log_term ) / lcd_term;
}

/***************************************************************************/

double
emd::inter::ees::get_dsn_UA()
{

  return A_prf*( 3*( -5*trE*h2 + en*(25*h2 - 3*sn2) + sn2*(trE + 2) )*log_term*lcd_term2
         + 2*h*sn*( -15*trE*h4 + 2*sn2*(14*trE + 3)*h2 + en*(75*h4 - 134*sn2*h2 + 55*sn4)
         - sn4*(13*trE + 10) ) ) / (lcd_term*sn);
}
/***************************************************************************/

double
emd::inter::ees::get_den_UA()
{

  return A_prf*(26*h*sn3 - 30*h3*sn + 3*lcd_term*(sn2 - 5*h2)*log_term);
}

/***************************************************************************/

double
emd::inter::ees::get_dh_UR()
{

  return R_prf*(-3*(21*(trE + 5)*h6 + 3*sn2*(110*en - trE + 105)*h4 + sn4*(220*en
         -17*trE + 135)*h2 + sn6*(10*en - trE + 5)))/lcd_term;
}

/***************************************************************************/

double
emd::inter::ees::get_dsn_UR()
{

  return R_prf*(30*h*sn*(3*(2*en + trE + 7)*h4 + 2*sn2*(20*en - trE + 15)*h2 +
         sn4*(10*en - trE +5 )))/lcd_term;

}
/***************************************************************************/

double
emd::inter::ees::get_den_UR()
{

  return R_prf*30*h*(sn4 + 3*h2*sn2);

}

/***************************************************************************/

void
emd::inter::ees::eval_derivs ()
{

  dhUA  = get_dh_UA();
  dhUR  = get_dh_UR();
  dsnUA = get_dsn_UA();
  dsnUR = get_dsn_UR();
  denUA = get_den_UA();
  denUR = get_den_UR();

}

/***************************************************************************/
/* NOTE must be called RIGHT AFTER eval_derivs */

double
emd::inter::ees::get_EES_int_force (const gsl_vector * dir,
  const paramdb::param_cache_EES_int * param)
{

  gsl_blas_ddot(dir, nhat, &dth);

  return -dth*(dhUA + dhUR);

}

/***************************************************************************/
/* NOTE must be called RIGHT AFTER eval_derivs */

double
emd::inter::ees::get_EES_int_torque (const gsl_matrix * gen,
  const paramdb::param_cache_EES_int * param)
{

  /* dvec = - S A gen |nhat> */
  gsl_blas_dgemv(CblasNoTrans,  1, gen     , nhat , 0, dvec );
  gsl_blas_dgemv(CblasNoTrans,  1, A       , dvec , 0, ddvec);
  gsl_blas_dgemv(CblasNoTrans, -1, param->S, ddvec, 0, dvec );

  gsl_blas_ddot(dvec, nhatx, &drsn);
  gsl_blas_ddot(dvec, Enhatx, &ddbl);
  dren = 2*(ddbl - drsn*en)/sn;

  return ( -drsn*(dsnUA + dsnUR) -dren*(denUA + denUR) );

}

/***************************************************************************/
/* NOTE must be called RIGHT AFTER eval_derivs */

double
emd::inter::ees::get_EES_int_force_sph (const gsl_vector * dir, const double & r,
  const paramdb::param_cache_EES_int * param)
{

  /* dth */
  gsl_blas_ddot(dir, nhat, &dth);

  /* dtnhat = (dir - (nhat.dir)*nhat)/r */
  gsl_vector_memcpy(dnhat, dir);
  gsl_vector_memcpy(dvec, nhat);
  gsl_vector_scale(dvec, dth);
  gsl_vector_sub(dnhat, dvec);
  gsl_vector_scale(dnhat, 1/r);

  /* dtsn */
  gsl_blas_dgemv(CblasNoTrans, 1, A, dnhat, 0, dvec);
  gsl_blas_dgemv(CblasNoTrans, 1, param->S, dvec,  0, ddvec);
  gsl_blas_ddot(ddvec, nhatx, &dtsn);

  /* dten */
  gsl_blas_ddot(ddvec, Enhatx, &ddbl);
  dten = 2*(ddbl - dtsn*en)/sn;

  return ( -dth*(dhUA + dhUR) - dtsn*(dsnUA + dsnUR) - dten*(denUA + denUR) );

}

/***************************************************************************/
/* NOTE must be called RIGHT AFTER eval_derivs */

double
emd::inter::ees::get_EES_int_torque_sph (const gsl_matrix * gen, const double & r,
  const paramdb::param_cache_EES_int * param)
{

  /* dvec = - S A gen |nhat> */
  gsl_blas_dgemv(CblasNoTrans,  1, gen     , nhat , 0, dvec );
  gsl_blas_dgemv(CblasNoTrans,  1, A       , dvec , 0, ddvec);
  gsl_blas_dgemv(CblasNoTrans, -1, param->S, ddvec, 0, dvec );

  /* drsn */
  gsl_blas_ddot(dvec, nhatx, &drsn);

  /* dren */
  gsl_blas_ddot(dvec, Enhatx, &ddbl);
  dren = 2*(ddbl - drsn*en)/sn;

  return ( -drsn*(dsnUA + dsnUR) -dren*(denUA + denUR) );

}

/***************************************************************************/
