/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef INT_EES_COMP_H
#define INT_EES_COMP_H

#include <iostream>
#include <cmath>

#include <gsl/gsl_math.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>

#include "gslext.h"
#include "paramdb.h"
#include "rigid_body.h"
#include "surface.h"

namespace emd
{

namespace inter
{

namespace ees
{

  static bool               init_flag = false;

  static gsl_vector *       nhat      = gsl_vector_calloc(3);
  static gsl_vector *       dnhat     = gsl_vector_calloc(3);
  static gsl_vector *       nhatx     = gsl_vector_calloc(3);
  static gsl_vector *       Enhatx    = gsl_vector_calloc(3);

  static gsl_matrix *       A         = gsl_matrix_calloc(3, 3);
  static gsl_matrix *       dA        = gsl_matrix_calloc(3, 3);

  static double             trE;
  static double             h;
  static double             sn;
  static double             en;

  static double             dh;
  static double             dsn;
  static double             den;

  static double             log_term;
  static double             A_prf;       /* attractive prefactor */
  static double             R_prf;       /* repulsive prefactor */
  static double             lcd_term;    /* least contact distance term (h^2 - sn^2) */
  static double             lcd_term2;

  extern double             UA;
  extern double             UR;

  /* aux */
  static double             h2, h3, h4, h5, h6;        /* useful powers of h */
  static double             sn2, sn3, sn4, sn5, sn6;   /* useful powers of sn */

  /* vars (Torque) */
  static double             drh;
  static double             drsn;
  static double             dren;
  static double             drUA;
  static double             drUR;

  /* vars (Force) */
  static double             dth;
  static double             dtsn;
  static double             dten;
  static double             dtUA;
  static double             dtUR;

  static double             dhUA;
  static double             dsnUA;
  static double             denUA;
  static double             dhUR;
  static double             dsnUR;
  static double             denUR;

  /* temporary */

  static gsl_matrix *       dmat      = gsl_matrix_calloc(3, 3);

  static gsl_vector *       dvec      = gsl_vector_calloc(3);
  static gsl_vector *       ddvec     = gsl_vector_calloc(3);

  static double             ddbl, dddbl;

  /* bases and generators: that would be handy to have them around */

  extern gsl_vector *       e_x;
  extern gsl_vector *       e_y;
  extern gsl_vector *       e_z;

  extern gsl_matrix *       R_x;
  extern gsl_matrix *       R_y;
  extern gsl_matrix *       R_z;

  /***************************************************************************/

  void init();

  bool is_init();

  int free_memory();

  /**** NOTE ********************************************************************************
   *                                                                                        *
   * In all cases, the transpose of the provided rotation matrix @A will be used!           *
   * (The orientation operators of the rigid bodies transform -a vector- (in the lab frame) *
   * to the representation of a same vector in the body frame; however, the expressions     *
   * given in the paper transform -the lab frame- to the body frame, hence transforming     *
   * a vector in the body frame to the lab frame)                                           *
   *                                                                                        *
   ******************************************************************************************/

  /*** contact distance related ****************/

  /* contact distance */
  double
  get_contact_distance (const gsl_matrix * A, const gsl_vector * nhat,
    const double & h, const paramdb::param_cache_EES_int * param);

  /*** energy/force/torque related *************/

  /**** flat interface ***********************************************************/

  /* interaction energy */
  double
  get_EES_int_energy (const gsl_matrix * A, const gsl_vector * nhat,
    const double & h, const paramdb::param_cache_EES_int * param, const double & r_c,
    bool & eval_flag);

  void
  eval_derivs ();

  /* exerted force on M_2 */
  double
  get_EES_int_force (const gsl_vector * dir, const paramdb::param_cache_EES_int * param);

  /* exerted torque on M_2 with respect to its CM */
  double
  get_EES_int_torque (const gsl_matrix * gen, const paramdb::param_cache_EES_int * param);

  /**** zeroth order approximation for spherical interfaces **********************/

  /* exerted force on M_2 */
  double
  get_EES_int_force_sph (const gsl_vector * dir, const double & r,
    const paramdb::param_cache_EES_int * param);

  /* exerted torque on M_2 with respect to its CM */
  double
  get_EES_int_torque_sph (const gsl_matrix * gen, const double & r,
    const paramdb::param_cache_EES_int * param);

  /*******************************************************************************/

  /* partial derivatives */
  double get_dh_UA ();
  double get_dsn_UA ();
  double get_den_UA ();
  double get_dh_UR ();
  double get_dsn_UR ();
  double get_den_UR ();

}; /* ees */

}; /* inter */

}; /* emd */

/***********************************************/

#endif /* INT_EES_COMP_H */
