/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "int_ees_uni.h"

using namespace std;

/***************************************************************************/

vector<emd::uni::neighbor_data> *      emd::uni::ees::n_list_ees;
bool                                   emd::uni::ees::nl_init_flag = false;
double                                 emd::uni::ees::r_c          = DEF_EES_R_C;
double                                 emd::uni::ees::r_s          = DEF_EES_R_S;
double                                 emd::uni::ees::max_offset   = DEF_EES_MAX_OFFSET;

/**** neighbor list related functions **************************************/

int
emd::uni::ees::init_neighbor_list()
{

  n_list_ees  =  new vector<neighbor_data>[rigid_bodies_count];

  /* reserving contiguous memory for neighbor data */
  for (int i=0; i<rigid_bodies_count; i++)
    n_list_ees[i].reserve(DEF_EES_MAX_NEIGHBORS);

  nl_init_flag = true;

}

/***************************************************************************/

void
emd::uni::ees::update_neighbor_list()
{

  switch (nl_method)
  {

    case NL_VERLET:
    {
      update_neighbor_list_verlet();
    }

  }

}

/***************************************************************************/

void
emd::uni::ees::update_neighbor_list_verlet()
{

  double            h, r, dist, sig;
  neighbor_data     n_dat;

  /* clear the previous neighbor list */
  for (unsigned int m=0; m<rigid_bodies_count; m++)
    n_list_ees[m].clear();

  for (unsigned int m=0; m<rigid_bodies_count; m++)
  {

    /* if the body is inactive, ignore the interaction */
    if ( rigid_bodies[m].status == rigid_body::EMD_RIGIDBODY_STATUS_INACTIVE )
      continue;

    /* flat surfaces */

    /* check if the interaction is to be evaluated */
    if ( (rigid_bodies[m].int_mask & MASK_EES_INT) != MASK_EES_INT ) continue;

    for (unsigned int s=0; s<flat_int_surfs_count; s++)
    {

      /* eval. h */
      gsl_vector_memcpy(tmp_r12, rigid_bodies[m].x);
      gsl_vector_sub(tmp_r12, flat_int_surfs[s].r0);
      gsl_blas_ddot(tmp_r12, flat_int_surfs[s].n, &h);

      sig = r_c + r_s +
        emd::paramdb::ees_int_param[flat_int_surfs[s].idx][rigid_bodies[m].idx].max_thickness;

      /* preliminary check */
      if ( h < sig )
      {

        /* exhaustive check: find the least contact distance */
        dist = emd::inter::ees::get_contact_distance(
          rigid_bodies[m].R, flat_int_surfs[s].n, h,
          &emd::paramdb::ees_int_param[flat_int_surfs[s].idx][rigid_bodies[m].idx] );

        if ( dist < (r_c + r_s) )
        {

          n_dat.idx    =  s;
          n_dat.n_type =  NT_FLAT_INT_SURFACE;

          n_list_ees[m].push_back(n_dat);

        }

      }

    }

    /* spherical surfaces */

    /* check if the interaction is to be evaluated */
    if ( (rigid_bodies[m].int_mask & MASK_EES_SPHERICAL_INT) != MASK_EES_SPHERICAL_INT ) continue;

    for (unsigned int s=0; s<sph_int_surfs_count; s++)
    {

      /* eval. h */
      gsl_vector_memcpy(tmp_r12, rigid_bodies[m].x);
      gsl_vector_sub(tmp_r12, sph_int_surfs[s].r0);
      r = gsl_blas_dnrm2(tmp_r12);
      h = r - sph_int_surfs[s].rad;

      /* tmp_r12 <= nhat */
      gsl_vector_normalize(tmp_r12);

      sig = r_c + r_s +
        emd::paramdb::ees_int_param[sph_int_surfs[s].idx][rigid_bodies[m].idx].max_thickness;

      /* preliminary check */
      if ( h < sig )
      {

        /* exhaustive check: find the least contact distance */
        dist = emd::inter::ees::get_contact_distance(
          rigid_bodies[m].R, tmp_r12, h,
          &emd::paramdb::ees_int_param[sph_int_surfs[s].idx][rigid_bodies[m].idx] );

        if ( dist < (r_c + r_s) )
        {

          n_dat.idx    =  s;
          n_dat.n_type =  NT_SPHERICAL_INT_SURFACE;

          n_list_ees[m].push_back(n_dat);

        }

      }

    }

  }

}

/***************************************************************************/

void
emd::uni::ees::print_neighbor_list()
{

  cout << setprecision(3);

  cout << "* printing EES interactions neighbor list: " << endl;

  for (int i=0; i<rigid_bodies_count; i++)
  {
    cout << "* Neighbors of rigid body " << i << ": ";

    int sz = n_list_ees[i].size();

    if (sz==0)
      cout << "none" << endl;

    else
    {

      for (int j=0; j<sz-1; j++)
      {
        cout << n_list_ees[i][j].idx;
        switch (n_list_ees[i][j].n_type)
        {
          case NT_FLAT_INT_SURFACE:
            cout << " (f) ,";
            break;
          case NT_SPHERICAL_INT_SURFACE:
            cout << " (s) ,";
            break;
         }
      }
      cout << n_list_ees[i][sz-1].idx;
      switch (n_list_ees[i][sz-1].n_type)
      {
        case NT_FLAT_INT_SURFACE:
          cout << " (f)" << endl;
          break;
        case NT_SPHERICAL_INT_SURFACE:
          cout << " (s)" << endl;
          break;
       }

    }

  }

  cout << setprecision(8);

}

/***************************************************************************/

/* the number of interactions */
unsigned long
emd::uni::ees::get_int_count()
{

  unsigned long num = 0;

  for (unsigned int m=0; m<rigid_bodies_count; m++)
    num += n_list_ees[m].size();

  num /= 2;

  return num;

}

/***************************************************************************/

 /*
  * the neighbor list (n_list_ees) is exclusively used for EES
  * interactions. therefore, regardless of the @clear variable, the
  * @int_energy variable of neighbor list will be updated.
  */
double
emd::uni::ees::eval_EES_single(const int& m, const bool& dyn, int& stat)
{

  int                                     s;
  double                                  r, h, U, total_U = 0;
  emd::paramdb::param_cache_EES_int *     param;

  /* if h<0, status is EXIT_FAILURE */
  int                  curr_stat = EXIT_SUCCESS;
  bool                 eval_flag;

  /* init ees workspace */
  if ( !emd::inter::ees::is_init() ) emd::inter::ees::init();

  for (neighbor_it it = n_list_ees[m].begin(); it!=n_list_ees[m].end(); it++)
  {

    s = (*it).idx;

    switch ( (*it).n_type )
    {

      /* flat surface */
      case NT_FLAT_INT_SURFACE:

        /* fetch parameter pointer */
        param = &emd::paramdb::ees_int_param[flat_int_surfs[s].idx][rigid_bodies[m].idx];

        /* eval h */
        gsl_vector_memcpy(tmp_r12, rigid_bodies[m].x);
        gsl_vector_sub(tmp_r12, flat_int_surfs[s].r0);
        gsl_blas_ddot(tmp_r12, flat_int_surfs[s].n, &h);

        /* evaluate the interaction energy */
        U = emd::inter::ees::get_EES_int_energy(rigid_bodies[m].R, flat_int_surfs[s].n,
          h, param, r_c, eval_flag);

        if (!eval_flag) continue;

        if (dyn)
        {

          /* evaluate partial derivatives */
          emd::inter::ees::eval_derivs();

          /* force */
          gsl_vector_memcpy(tmp_force, flat_int_surfs[s].n);
          gsl_vector_scale(tmp_force, emd::inter::ees::get_EES_int_force(flat_int_surfs[s].n, param) );

          /* torque */
          gsl_vector_set(tmp_torque, 0, emd::inter::ees::get_EES_int_torque(emd::inter::ees::R_x, param));
          gsl_vector_set(tmp_torque, 1, emd::inter::ees::get_EES_int_torque(emd::inter::ees::R_y, param));
          gsl_vector_set(tmp_torque, 2, emd::inter::ees::get_EES_int_torque(emd::inter::ees::R_z, param));

          /* adding up the current force contribution */
          gsl_vector_add(rigid_bodies[m].force, tmp_force);

          /* adding up the current force contribution (surf) */
          gsl_vector_sub(flat_int_surfs[s].force, tmp_force);

          /* adding up the current torque contribution */
          gsl_vector_add(rigid_bodies[m].torque, tmp_torque);

        }

        break;

      /* spherical surface */
      case NT_SPHERICAL_INT_SURFACE:

        /* fetch parameter pointer */
        param = &emd::paramdb::ees_int_param[sph_int_surfs[s].idx][rigid_bodies[m].idx];

        /* eval h */
        gsl_vector_memcpy(tmp_r12, rigid_bodies[m].x);
        gsl_vector_sub(tmp_r12, sph_int_surfs[s].r0);
        r = gsl_blas_dnrm2(tmp_r12);
        h = r - sph_int_surfs[s].rad;

        /* tmp_r12 <= nhat */
        gsl_vector_normalize(tmp_r12);

        /* evaluate the interaction energy */
        U = emd::inter::ees::get_EES_int_energy(rigid_bodies[m].R, tmp_r12,
          h, param, r_c, eval_flag);

        if (!eval_flag) continue;

        if (dyn)
        {

          /* evaluate partial derivatives */
          emd::inter::ees::eval_derivs();

          /* force */
          gsl_vector_set(tmp_force, 0, emd::inter::ees::get_EES_int_force_sph(emd::inter::ees::e_x, r, param));
          gsl_vector_set(tmp_force, 1, emd::inter::ees::get_EES_int_force_sph(emd::inter::ees::e_y, r, param));
          gsl_vector_set(tmp_force, 2, emd::inter::ees::get_EES_int_force_sph(emd::inter::ees::e_z, r, param));

          /* torque */
          gsl_vector_set(tmp_torque, 0, emd::inter::ees::get_EES_int_torque_sph(emd::inter::ees::R_x, r, param));
          gsl_vector_set(tmp_torque, 1, emd::inter::ees::get_EES_int_torque_sph(emd::inter::ees::R_y, r, param));
          gsl_vector_set(tmp_torque, 2, emd::inter::ees::get_EES_int_torque_sph(emd::inter::ees::R_z, r, param));

          /* adding up the current force contribution (rigid body) */
          gsl_vector_add(rigid_bodies[m].force, tmp_force);

          /* adding up the current force contribution (surf) */
          gsl_vector_sub(sph_int_surfs[s].force, tmp_force);

          /* adding up the current torque contribution */
          gsl_vector_add(rigid_bodies[m].torque, tmp_torque);

        }

        break;

    }

    /* set the neighbor_data energy for m ...*/
    (*it).int_energy = U;

    /* append to the rigid_body energy */
    rigid_bodies[m].potential += U;

    total_U += U;

#ifdef MOLUNIVERSE_DEBUG_NUM
    cout << "* emd::uni::eval_EES_single_surface: Int. energy between surface "
         << s << " and rigid body " << m << ": " << U << endl;
#endif

#ifdef CHECK_OVERFLOW_EES_INT
    if ( abs(U) > OVERFLOW_ENERGY_EES_INT )
      cout << "* emd::uni::eval_EES_single: Int. energy overflow (rb:" << m << ", surf:" << s
           << ")." << endl;
#endif

  }

  return total_U;

}

/*******************************************************************************************/

double
emd::uni::ees::eval_EES(const bool& dyn, int& stat)
{

  int curr_stat    =  EXIT_SUCCESS;

  double    old_potential = potential;

  for (int m1=0; m1<rigid_bodies_count; m1++)
  {
    potential += eval_EES_single(m1, dyn, curr_stat);
    if (curr_stat==EXIT_FAILURE) stat = EXIT_FAILURE;
  }

  return (potential - old_potential);

}

/*******************************************************************************************/

int
emd::uni::ees::free_memory()
{

  /* static memory */
  /*gsl_vector_free(tmp_r12);
  gsl_vector_free(tmp_force);
  gsl_vector_free(tmp_torque);
  gsl_vector_free(tmp_v);*/

  if (nl_init_flag)
    delete[] n_list_ees;

  return EXIT_SUCCESS;
}
