/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "int_gb_comp.h"

using namespace std;
using namespace emd::inter::gb;

/***********************************************************************************/

gsl_vector *       emd::inter::gb::e_x       = gsl_vector_calloc(3);
gsl_vector *       emd::inter::gb::e_y       = gsl_vector_calloc(3);
gsl_vector *       emd::inter::gb::e_z       = gsl_vector_calloc(3);

gsl_matrix *       emd::inter::gb::R_x       = gsl_matrix_calloc(3, 3);
gsl_matrix *       emd::inter::gb::R_y       = gsl_matrix_calloc(3, 3);
gsl_matrix *       emd::inter::gb::R_z       = gsl_matrix_calloc(3, 3);

/***********************************************************************************/

void
emd::inter::gb::init ()
{

  gsl_vector_set_basis(e_x, 0);
  gsl_vector_set_basis(e_y, 1);
  gsl_vector_set_basis(e_z, 2);

  gsl_matrix_rot_x(R_x);
  gsl_matrix_rot_y(R_y);
  gsl_matrix_rot_z(R_z);

  init_flag = true;

}

/***********************************************************************************/

bool
emd::inter::gb::is_init ()
{
  return init_flag;
}

/***********************************************************************************/

int
emd::inter::gb::free_memory ()
{

  gsl_vector_free(e_x);
  gsl_vector_free(e_y);
  gsl_vector_free(e_z);

  gsl_matrix_free(R_x);
  gsl_matrix_free(R_y);
  gsl_matrix_free(R_z);

}

/***********************************************************************************/

double
emd::inter::gb::get_GB_int_energy(const gsl_matrix * A1_0, const gsl_matrix * A2_0,
  const gsl_vector * r12_0, const paramdb::param_cache_GB_int * param, const double& r_c0,
  bool& eval_flag, int& stat)
{

  /* cutoff distance */
  r_c = r_c0;

  /* rnorm */
  rnorm = gsl_blas_dnrm2(r12_0);

  /* ignore the interaction if it is outside the cutoff distance */
  if ( rnorm > r_c )
  {
    eval_flag = false;
    return 0;
  }

  /* A_i are set to the transposed of the provided A_i */
  gsl_matrix_transpose_memcpy(A1, A1_0);
  gsl_matrix_transpose_memcpy(A2, A2_0);

  /* rhat */
  gsl_vector_memcpy(rhat, r12_0);
  gsl_vector_scale(rhat, 1/rnorm);

  /* Gamma1 */
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1, param->S1p2, A1, 0, dmat);
  gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1, A1, dmat, 0, Gamma1);

  /* Gamma2 */
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1, param->S2p2, A2, 0, dmat);
  gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1, A2, dmat, 0, Gamma2);

  /* G and Ginv*rhat */
  int    sig;
  gsl_matrix_memcpy(dmat, Gamma1);
  gsl_matrix_add(dmat, Gamma2);
  gsl_linalg_LU_decomp(dmat, perm, &sig);
  gsl_linalg_LU_solve(dmat, perm, rhat, s);

  /* anisotropic distance function and contact distance */
  gsl_blas_ddot(s, rhat, &ddbl);
  sigma12 = 1/sqrt(2.0*ddbl);
  sigma12p3 = gsl_pow_3(sigma12);
  h = rnorm - sigma12;

  if (h<0)
    stat = EXIT_FAILURE;
  else
    stat = EXIT_SUCCESS;

  double ddbl, dddbl;

  /* G */
  gsl_matrix_memcpy(G, Gamma1);
  gsl_matrix_add(G, Gamma2);

  /* det G */
  dG = gsl_3matrix_det(G);

  /* Psi1 and Psi2 */
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1, param->E1, A1, 0, dmat);
  gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1, A1, dmat, 0, Psi1);

  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1, param->E2, A2, 0, dmat);
  gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1, A2, dmat, 0, Psi2);

  /* B and Binv*rhat */
  gsl_matrix_memcpy(dmat, Psi1);
  gsl_matrix_add(dmat, Psi2);
  gsl_linalg_LU_decomp(dmat, perm, &sig);
  gsl_linalg_LU_solve(dmat, perm, rhat, w);

  /* orientation dependent vars */
  eta = sqrt(2.0 * param->dS1 * param->dS2 / dG);
  gsl_blas_ddot(rhat, w, &ddbl);
  chi = 2*ddbl;

  /* integer power of sig/(h + sig) */
  gam   = param->sig / (h + param->sig);
  gam2  = gam*gam;
  gam3  = gam*gam2;
  gam6  = gam3*gam3;
  gam12 = gam6*gam6;

  /* anisotropic prefactor */
  aniso_term = 4.0 * param->eps * gsl_pow_int(eta, param->nu) * gsl_pow_int(chi, param->mu);

  /* LJ-like term */
  lj_term = gam12 - gam6;

  eval_flag = true;

#ifdef EMD_GB_SHIFT

  gam_c   = param->sig / (r_c - sigma12 + param->sig);
  gam2_c  = gam_c*gam_c;
  gam3_c  = gam_c*gam2_c;
  gam6_c  = gam3_c*gam3_c;
  gam12_c = gam6_c*gam6_c;

  lj_term_c = gam12_c - gam6_c;

  return aniso_term*(lj_term - lj_term_c);

#else

  return aniso_term*lj_term;

#endif

};

/***********************************************************************************/

/* exerted force on M_2 */
double
emd::inter::gb::get_GB_int_force (const gsl_vector * dir, const paramdb::param_cache_GB_int * param)
{

  /* gamma */
  gsl_blas_ddot(dir, rhat, &gamma);

  /* u and dtsigma12 */
  gsl_vector_memcpy(dvec, rhat);
  gsl_vector_memcpy(u, dir);
  gsl_vector_scale(dvec, gamma);
  gsl_vector_sub(u, dvec);
  gsl_vector_scale(u, 1.0/rnorm);
  gsl_blas_ddot(u, s, &ddbl);
  dtsigma12 = 2.0*sigma12p3*ddbl;

  /* dteta = 0 */

  /* dtchi */
  gsl_blas_ddot(u, w, &ddbl);
  dtchi = 4.0*ddbl;

  double dtU =
    aniso_term*( lj_term*param->mu*dtchi/chi - gam*(12*gam12 - 6*gam6)*(gamma + dtsigma12)/param->sig );

#ifdef EMD_GB_SHIFT

  double dtU_c =
    aniso_term*( lj_term_c*param->mu*dtchi/chi - gam_c*(12*gam12_c - 6*gam6_c)*dtsigma12/param->sig );

  return - dtU - dtU_c;

#else

  return - dtU;

#endif

};

/* exerted torque on M_2 */
double
emd::inter::gb::get_GB_int_torque (const gsl_matrix * gen, const paramdb::param_cache_GB_int * param)
{

  /* drh */
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, -1, Gamma2, gen, 0, dmat);
  gsl_blas_dgemv(CblasNoTrans, 1, dmat, s, 0, dvec);
  gsl_blas_ddot(s, dvec, &ddbl);

  drh = -2.0*sigma12p3*ddbl;

  /* drG */
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 2, gen, Gamma2, 0, dmat);

  /* drdG */
  drdG = 0;
  for (int i=0; i<3; i++)
  {
    gsl_matrix_get_row(dvec, dmat, i);
    gsl_matrix_memcpy(ddmat, G);
    gsl_matrix_set_row(ddmat, i, dvec);
    ddbl = gsl_3matrix_det(ddmat);
    drdG += ddbl;
  }

  /* dreta */
  dreta = -0.5*drdG*eta/dG;

  /* drchi */
  gsl_vector_memcpy(dvec, w);
  gsl_blas_dgemv(CblasNoTrans, -1, gen, dvec, 0, ddvec);
  gsl_blas_dgemv(CblasNoTrans, 1, Psi2, ddvec, 0, dvec);
  gsl_blas_ddot(w, dvec, &(drchi));
  drchi *= -4.0;

  double drU =
    aniso_term*( lj_term*(param->nu*dreta/eta + param->mu*drchi/chi) - gam*(12*gam12 - 6*gam6)*drh/param->sig);

#ifdef EMD_GB_SHIFT

  double drU_c =
    aniso_term*( lj_term_c*(param->nu*dreta/eta + param->mu*drchi/chi) - gam_c*(12*gam12_c - 6*gam6_c)*drh/param->sig);

  return - drU - drU_c;

#else

  return - drU;

#endif

};
