/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "int_re2_comp.h"

using namespace std;
using namespace emd::inter::re2;

/***********************************************************************************/

gsl_vector *       emd::inter::re2::e_x       = gsl_vector_calloc(3);
gsl_vector *       emd::inter::re2::e_y       = gsl_vector_calloc(3);
gsl_vector *       emd::inter::re2::e_z       = gsl_vector_calloc(3);

gsl_matrix *       emd::inter::re2::R_x       = gsl_matrix_calloc(3, 3);
gsl_matrix *       emd::inter::re2::R_y       = gsl_matrix_calloc(3, 3);
gsl_matrix *       emd::inter::re2::R_z       = gsl_matrix_calloc(3, 3);

/***********************************************************************************/

void
emd::inter::re2::init ()
{

  gsl_vector_set_basis(e_x, 0);
  gsl_vector_set_basis(e_y, 1);
  gsl_vector_set_basis(e_z, 2);

  gsl_matrix_rot_x(R_x);
  gsl_matrix_rot_y(R_y);
  gsl_matrix_rot_z(R_z);

  init_flag = true;

}

/***********************************************************************************/

bool
emd::inter::re2::is_init ()
{
  return init_flag;
}

/***********************************************************************************/

int
emd::inter::re2::free_memory ()
{

  gsl_vector_free(e_x);
  gsl_vector_free(e_y);
  gsl_vector_free(e_z);

  gsl_matrix_free(R_x);
  gsl_matrix_free(R_y);
  gsl_matrix_free(R_z);

}

/***********************************************************************************/

/* contact distance */
double
emd::inter::re2::get_contact_distance(const gsl_matrix * A1_0, const gsl_matrix * A2_0,
  const gsl_vector * r12_0, const paramdb::param_cache_RE2_int * param)
{

  double ddbl;
  int    sig;

  /* A_i are set to the transposed of the provided A_i */
  gsl_matrix_transpose_memcpy(A1, A1_0);
  gsl_matrix_transpose_memcpy(A2, A2_0);

  /* rnorm and rhat */
  rnorm = gsl_blas_dnrm2(r12_0);
  gsl_vector_memcpy(rhat, r12_0);
  gsl_vector_scale(rhat, 1/rnorm);

  /* Gamma1 */
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1, param->S1p2, A1, 0, dmat);
  gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1, A1, dmat, 0, Gamma1);

  /* Gamma2 */
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1, param->S2p2, A2, 0, dmat);
  gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1, A2, dmat, 0, Gamma2);

  /* G and Ginv*rhat */
  gsl_matrix_memcpy(dmat, Gamma1);
  gsl_matrix_add(dmat, Gamma2);
  gsl_linalg_LU_decomp(dmat, perm, &sig);
  gsl_linalg_LU_solve(dmat, perm, rhat, s);

  /* anisotropic distance function and contact distance */
  gsl_blas_ddot(s, rhat, &ddbl);
  sigma12 = 1/sqrt(0.5*ddbl);
  sigma12p3 = gsl_pow_3(sigma12);
  h = rnorm - sigma12;

  return h;

}

/***********************************************************************************/

/* the lower bound of the contact distance */
double
emd::inter::re2::get_contact_distance_lb(const gsl_matrix * A1_0, const gsl_matrix * A2_0,
  const gsl_vector * r12_0, const paramdb::param_cache_RE2_int * param)
{

  double ddbl;
  int    sig;

  /* rnorm and rhat */
  rnorm = gsl_blas_dnrm2(r12_0);
  gsl_vector_memcpy(rhat, r12_0);
  gsl_vector_scale(rhat, 1/rnorm);

  double r1, r2;

  gsl_blas_dgemv(CblasTrans, 1, A1_0, rhat, 0, dvec);
  gsl_blas_dgemv(CblasNoTrans, 1, param->S1pm2, dvec, 0, ddvec);
  gsl_blas_ddot(ddvec, dvec, &r1);

  gsl_blas_dgemv(CblasTrans, 1, A2_0, rhat, 0, dvec);
  gsl_blas_dgemv(CblasNoTrans, 1, param->S1pm2, dvec, 0, ddvec);
  gsl_blas_ddot(ddvec, dvec, &r1);

  return rnorm - 1/sqrt(r1) - 1/sqrt(r2);

}

/***********************************************************************************/

/* translational variation on contact distance */
double
emd::inter::re2::get_dt_contact_distance(const gsl_vector * dir,
  const paramdb::param_cache_RE2_int * param)
{

  /* gamma */
  gsl_blas_ddot(dir, rhat, &gamma);

  /* u */
  gsl_vector_memcpy(dvec, rhat);
  gsl_vector_memcpy(u, dir);
  gsl_vector_scale(dvec, gamma);
  gsl_vector_sub(u, dvec);
  gsl_vector_scale(u, 1.0/rnorm);

  gsl_blas_ddot(u, s, &ddbl);
  dth = gamma + 0.5*sigma12p3*ddbl;

  return dth;

}

/***********************************************************************************/

/* rotational variation on contact distance */
double
emd::inter::re2::get_dr_contact_distance(const gsl_matrix * gen,
  const paramdb::param_cache_RE2_int * param)
{

  /* Lambda */
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1, A2, gen, 0, Lambda);
  gsl_matrix_scale(Lambda, -1);

  /* Delta, drh */
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1, param->S2p2, Lambda, 0, dmat);
  gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1, A2, dmat, 0, Delta);
  gsl_blas_dgemv(CblasNoTrans, 1, Delta, s, 0, ddvec);
  gsl_blas_ddot(s, ddvec, &ddbl);

  drh = -0.5*sigma12p3*ddbl;

  return drh;

}

/***********************************************************************************/

double
emd::inter::re2::get_RE2_int_energy(const gsl_matrix * A1_0, const gsl_matrix * A2_0,
  const gsl_vector * r12_0, const paramdb::param_cache_RE2_int * param, const double& r_c0,
  bool& eval_flag, int& stat)
{

  /* cutoff distance */
  r_c = r_c0;

  /*
     get the least contact distance
     the following vars will be set, upon a call to get_dt_contact_distance(...):
     { A1, A2, rnorm, rhat, Gamma1, Gamma2, s, sigma12, sigma12p3, h }
  */
  get_contact_distance(A1_0, A2_0, r12_0, param);

  /* ignore the interaction if it is outside the cutoff distance */
  if ( h > r_c )
  {
    eval_flag = false;
    return 0;
  }

#ifdef EMD_RE2_SHIFT

  h_c = r_c;

#endif

  double ddbl, dddbl;
  int    sig;

  /* projected diameters */
  gsl_blas_dgemv(CblasNoTrans, 1, A1, rhat, 0, dvec);
  gsl_blas_dgemv(CblasNoTrans, 1, param->S1pm2, dvec, 0, v1);
  gsl_blas_ddot(dvec, v1, &ddbl);
  sigma1 = 1/sqrt(ddbl);
  sigma1p3 = gsl_pow_3(sigma1);

  gsl_blas_dgemv(CblasNoTrans, 1, A2, rhat, 0, dvec);
  gsl_blas_dgemv(CblasNoTrans, 1, param->S2pm2, dvec, 0, v2);
  gsl_blas_ddot(dvec, v2, &ddbl);
  sigma2 = 1/sqrt(ddbl);
  sigma2p3 = gsl_pow_3(sigma2);

  /* H */
  gsl_matrix_memcpy(dmat, Gamma2);
  gsl_matrix_scale(dmat, 1/sigma2);
  gsl_matrix_memcpy(H, Gamma1);
  gsl_matrix_scale(H, 1/sigma1);
  gsl_matrix_add(H, dmat);

  /* det H */
  dH = gsl_3matrix_det(H);

  lambda = param->dS1/(sigma1*sigma1) + param->dS2/(sigma2*sigma2);
  nu = sqrt(dH/(sigma1+sigma2));

  /* Psi1 and Psi2 */
  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1, param->E1, A1, 0, dmat);
  gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1, A1, dmat, 0, ddmat);

  gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1, param->E2, A2, 0, dmat);
  gsl_blas_dgemm(CblasTrans, CblasNoTrans, 1, A2, dmat, 0, dddmat);

  /* B and Binv*rhat */
  gsl_matrix_memcpy(dmat, ddmat);
  gsl_matrix_add(dmat, dddmat);
  gsl_linalg_LU_decomp(dmat, perm, &sig);
  gsl_linalg_LU_solve(dmat, perm, rhat, w);

  /* orientation dependent vars */
  eta = lambda/nu;
  gsl_blas_ddot(rhat, w, &ddbl);
  chi = 2*ddbl;

  if (h<0)
    stat = EXIT_FAILURE;
  else
    stat = EXIT_SUCCESS;

  /* attractive part */
  double tmp;

  ddbl = -param->hamaker*param->dS1*param->dS2*(1.0+3.0*param->sigma_c*eta*chi/h);
  tmp = h/2.0;
  dddbl = 36.0*
    (gsl_matrix_get(param->S1, 0, 0) + tmp)*(gsl_matrix_get(param->S1, 1, 1) + tmp)*
    (gsl_matrix_get(param->S1, 2, 2) + tmp)*(gsl_matrix_get(param->S2, 0, 0) + tmp)*
    (gsl_matrix_get(param->S2, 1, 1) + tmp)*(gsl_matrix_get(param->S2, 2, 2) + tmp);
  UA = ddbl/dddbl;

  ddbl = param->hamaker*param->dS1*param->dS2*
    (1+45.0*param->sigma_c*eta*chi/(56.0*h))*gsl_pow_6(param->sigma_c/h);
  tmp = h/EMD_RE2_SIXTYROOT;
  dddbl = 2025.0*
    (gsl_matrix_get(param->S1, 0, 0) + tmp)*(gsl_matrix_get(param->S1, 1, 1) + tmp)*
    (gsl_matrix_get(param->S1, 2, 2) + tmp)*(gsl_matrix_get(param->S2, 0, 0) + tmp)*
    (gsl_matrix_get(param->S2, 1, 1) + tmp)*(gsl_matrix_get(param->S2, 2, 2) + tmp);
  UR = ddbl/dddbl;

  eval_flag = true;

#ifdef EMD_RE2_SHIFT

  ddbl = -param->hamaker*param->dS1*param->dS2*(1.0+3.0*param->sigma_c*eta*chi/h_c);
  tmp = h_c/2.0;
  dddbl = 36.0*
    (gsl_matrix_get(param->S1, 0, 0) + tmp)*(gsl_matrix_get(param->S1, 1, 1) + tmp)*
    (gsl_matrix_get(param->S1, 2, 2) + tmp)*(gsl_matrix_get(param->S2, 0, 0) + tmp)*
    (gsl_matrix_get(param->S2, 1, 1) + tmp)*(gsl_matrix_get(param->S2, 2, 2) + tmp);
  UA_c = ddbl/dddbl;

  ddbl = param->hamaker*param->dS1*param->dS2*
    (1+45.0*param->sigma_c*eta*chi/(56.0*h_c))*gsl_pow_6(param->sigma_c/h_c);
  tmp = h_c/EMD_RE2_SIXTYROOT;
  dddbl = 2025.0*
    (gsl_matrix_get(param->S1, 0, 0) + tmp)*(gsl_matrix_get(param->S1, 1, 1) + tmp)*
    (gsl_matrix_get(param->S1, 2, 2) + tmp)*(gsl_matrix_get(param->S2, 0, 0) + tmp)*
    (gsl_matrix_get(param->S2, 1, 1) + tmp)*(gsl_matrix_get(param->S2, 2, 2) + tmp);
  UR_c = ddbl/dddbl;

  return UA + UR - UA_c - UR_c;

#else

  return UA + UR;

#endif

};

/***********************************************************************************/

/* exerted force on M_2 */
double
emd::inter::re2::get_RE2_int_force (const gsl_vector * dir, const paramdb::param_cache_RE2_int * param)
{

  /*
     get the rotation variation of least contact distance
     the following vars will be set, upon a call to get_dr_contact_distance(...):
     { gamma, u, dth }
  */
  get_dt_contact_distance(dir, param);

  /* dtsigma1 and dtsigma2 */
  gsl_blas_dgemv(CblasNoTrans, 1, A1, u, 0, dvec);
  gsl_blas_ddot(dvec, v1, &ddbl);
  dtsigma1 = -sigma1p3*ddbl;

  gsl_blas_dgemv(CblasNoTrans, 1, A2, u, 0, dvec);
  gsl_blas_ddot(dvec, v2, &ddbl);
  dtsigma2 = -sigma2p3*ddbl;

  /* dtH */
  gsl_matrix_memcpy(dmat, Gamma2);
  gsl_matrix_scale(dmat, -dtsigma2/gsl_pow_2(sigma2));
  gsl_matrix_memcpy(ddmat, Gamma1);
  gsl_matrix_scale(ddmat, -dtsigma1/gsl_pow_2(sigma1));
  gsl_matrix_add(dmat, ddmat);

  /* dtdH */
  dtdH = 0.0;
  for (int i=0; i<3; i++)
  {
    gsl_matrix_get_row(dvec, dmat, i);
    gsl_matrix_memcpy(ddmat, H);
    gsl_matrix_set_row(ddmat, i, dvec);
    ddbl = gsl_3matrix_det(ddmat);
    dtdH += ddbl;
  }

  /* dteta and dtchi */
  dteta = eta*((dtsigma1+dtsigma2)/(2.0*sigma1+2.0*sigma2) 
    - 0.5*dtdH/dH - (2.0/lambda)*(param->dS1*dtsigma1/sigma1p3
    + param->dS2*dtsigma2/sigma2p3));
  gsl_blas_ddot(u, w, &ddbl);
  dtchi = 4.0*ddbl;

  dtUA = UA*(3.0*param->sigma_c*(eta*dtchi+dteta*chi)/(h+3.0*param->sigma_c*eta*chi) -
    dth*(1.0/h - 1.0/(h+3.0*param->sigma_c*eta*chi) +
    1.0/(h+2.0*gsl_matrix_get(param->S1, 0, 0)) + 1.0/(h+2.0*gsl_matrix_get(param->S1, 1, 1)) +
    1.0/(h+2.0*gsl_matrix_get(param->S1, 2, 2)) + 1.0/(h+2.0*gsl_matrix_get(param->S2, 0, 0)) +
    1.0/(h+2.0*gsl_matrix_get(param->S2, 1, 1)) + 1.0/(h+2.0*gsl_matrix_get(param->S2, 2, 2))));

  dtUR = UR*((45.0/56.0)*param->sigma_c*(eta*dtchi+dteta*chi)/(h+(45.0/56.0)*param->sigma_c*eta*chi) -
    dth*(7.0/h - 1.0/(h+(45.0/56.0)*param->sigma_c*eta*chi) + 
    1.0/(h+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S1, 0, 0)) +
    1.0/(h+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S1, 1, 1)) +
    1.0/(h+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S1, 2, 2)) +
    1.0/(h+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S2, 0, 0)) +
    1.0/(h+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S2, 1, 1)) +
    1.0/(h+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S2, 2, 2))));

#ifdef EMD_RE2_SHIFT

  ddbl = rnorm/r_c;
  dth_c = (dth - gamma)*ddbl + gamma;

  dtUA_c = UA_c*(3.0*param->sigma_c*ddbl*(eta*dtchi+dteta*chi)/(h_c + 3.0*param->sigma_c*eta*chi) -
    dth_c*(1.0/h_c - 1.0/(h_c + 3.0*param->sigma_c*eta*chi) +
    1.0/(h_c+2.0*gsl_matrix_get(param->S1, 0, 0)) + 1.0/(h_c+2.0*gsl_matrix_get(param->S1, 1, 1)) +
    1.0/(h_c+2.0*gsl_matrix_get(param->S1, 2, 2)) + 1.0/(h_c+2.0*gsl_matrix_get(param->S2, 0, 0)) +
    1.0/(h_c+2.0*gsl_matrix_get(param->S2, 1, 1)) + 1.0/(h_c+2.0*gsl_matrix_get(param->S2, 2, 2))));

  dtUR_c = UR_c*((45.0/56.0)*param->sigma_c*ddbl*(eta*dtchi+dteta*chi)/(h_c+(45.0/56.0)*param->sigma_c*eta*chi) -
    dth_c*(7.0/h_c - 1.0/(h_c+(45.0/56.0)*param->sigma_c*eta*chi) + 
    1.0/(h_c+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S1, 0, 0)) +
    1.0/(h_c+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S1, 1, 1)) +
    1.0/(h_c+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S1, 2, 2)) +
    1.0/(h_c+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S2, 0, 0)) +
    1.0/(h_c+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S2, 1, 1)) +
    1.0/(h_c+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S2, 2, 2))));

    return - dtUA - dtUR + dtUA_c + dtUR_c;

#else

  return - dtUA - dtUR;

#endif

};

/* exerted torque on M_2 */
double
emd::inter::re2::get_RE2_int_torque (const gsl_matrix * gen, const paramdb::param_cache_RE2_int * param)
{

  /*
     get the rotation variation of least contact distance
     the following vars will be set, upon a call to get_contact_distance(...):
     { Lambda, Delta, drh }
  */
  get_dr_contact_distance(gen, param);

  /* dtsigma2 */
  gsl_blas_dgemv(CblasNoTrans, 1, Lambda, rhat, 0, p);
  gsl_blas_ddot(p, v2, &ddbl);
  drsigma2 = -sigma2p3*ddbl;

  /* drH */
  gsl_matrix_transpose_memcpy(dmat, Delta);     /* Delta is set by get_dr_contact_distance(...) */
  gsl_matrix_add(dmat, Delta);
  gsl_matrix_scale(dmat, 1.0/sigma2);
  gsl_matrix_memcpy(ddmat, Gamma2);
  gsl_matrix_scale(ddmat, -drsigma2/gsl_pow_2(sigma2));
  gsl_matrix_add(dmat, ddmat);                                   /* <- drH */

  /* drdH */
  drdH = 0;
  for (int i=0; i<3; i++)
  {
    gsl_matrix_get_row(dvec, dmat, i);
    gsl_matrix_memcpy(ddmat, H);
    gsl_matrix_set_row(ddmat, i, dvec);
    ddbl = gsl_3matrix_det(ddmat);
    drdH += ddbl;
  }

  dreta = eta*(0.5*drsigma2/(sigma1+sigma2) - 0.5*drdH/dH -
    2.0*param->dS2*drsigma2/(lambda*sigma2p3));

  gsl_vector_memcpy(dvec, w);
  gsl_blas_dgemv(CblasNoTrans, 1, Lambda, dvec, 0, dddvec);
  gsl_blas_dgemv(CblasNoTrans, 1, param->E2, dddvec, 0, dvec);
  gsl_blas_dgemv(CblasTrans, 1, A2, dvec, 0, dddvec);
  gsl_blas_ddot(w, dddvec, &(drchi));
  drchi *= -4.0;

  drUA = UA*(3.0*param->sigma_c*(eta*drchi+dreta*chi)/(h+3.0*param->sigma_c*eta*chi) -
    drh*(1.0/h - 1.0/(h+3.0*param->sigma_c*eta*chi) +
    1.0/(h+2.0*gsl_matrix_get(param->S1, 0, 0)) + 1.0/(h+2.0*gsl_matrix_get(param->S1, 1, 1)) +
    1.0/(h+2.0*gsl_matrix_get(param->S1, 2, 2)) + 1.0/(h+2.0*gsl_matrix_get(param->S2, 0, 0)) +
    1.0/(h+2.0*gsl_matrix_get(param->S2, 1, 1)) + 1.0/(h+2.0*gsl_matrix_get(param->S2, 2, 2))));

  drUR = UR*((45.0/56.0)*param->sigma_c*(eta*drchi+dreta*chi)/(h+(45.0/56.0)*param->sigma_c*eta*chi) -
    drh*(7.0/h - 1/(h+(45.0/56.0)*param->sigma_c*eta*chi) + 
    1.0/(h+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S1, 0, 0)) +
    1.0/(h+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S1, 1, 1)) +
    1.0/(h+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S1, 2, 2)) +
    1.0/(h+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S2, 0, 0)) +
    1.0/(h+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S2, 1, 1)) +
    1.0/(h+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S2, 2, 2))));

#ifdef EMD_RE2_SHIFT

  drUA_c = UA_c*(3.0*param->sigma_c*(eta*drchi+dreta*chi)/(h_c+3.0*param->sigma_c*eta*chi) -
    drh*(1.0/h_c - 1.0/(h_c+3.0*param->sigma_c*eta*chi) +
    1.0/(h_c+2.0*gsl_matrix_get(param->S1, 0, 0)) + 1.0/(h_c+2.0*gsl_matrix_get(param->S1, 1, 1)) +
    1.0/(h_c+2.0*gsl_matrix_get(param->S1, 2, 2)) + 1.0/(h_c+2.0*gsl_matrix_get(param->S2, 0, 0)) +
    1.0/(h_c+2.0*gsl_matrix_get(param->S2, 1, 1)) + 1.0/(h_c+2.0*gsl_matrix_get(param->S2, 2, 2))));

  drUR_c = UR_c*((45.0/56.0)*param->sigma_c*(eta*drchi+dreta*chi)/(h_c+(45.0/56.0)*param->sigma_c*eta*chi) -
    drh*(7.0/h_c - 1/(h_c+(45.0/56.0)*param->sigma_c*eta*chi) + 
    1.0/(h_c+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S1, 0, 0)) +
    1.0/(h_c+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S1, 1, 1)) +
    1.0/(h_c+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S1, 2, 2)) +
    1.0/(h_c+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S2, 0, 0)) +
    1.0/(h_c+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S2, 1, 1)) +
    1.0/(h_c+EMD_RE2_SIXTYROOT*gsl_matrix_get(param->S2, 2, 2))));

  return - drUA - drUR + drUA_c + drUR_c;

#else

  return - drUA - drUR;

#endif

};
