/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/***************************************************************************
 *   RE^2 energy and first derivatives                                     *
 *   - A container for workspace                                           *
 *   - static functions for the energy and its derivates                   *
 ***************************************************************************/

#ifndef INT_RE2_COMP_H
#define INT_RE2_COMP_H

#include <iostream>

#include <gsl/gsl_math.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>

#include "gslext.h"
#include "paramdb.h"
#include "rigid_body.h"

#define RE2INTWORKSPACE_DEBUG

#define EMD_RE2_SIXTYROOT 3.914867641
#define EMD_RE2_SHIFT

namespace emd
{

namespace inter
{

namespace re2
{

  static bool               init_flag = false;

  static gsl_vector *       rhat      = gsl_vector_calloc(3);
  static gsl_vector *       s         = gsl_vector_calloc(3);
  static gsl_vector *       v1        = gsl_vector_calloc(3);
  static gsl_vector *       v2        = gsl_vector_calloc(3);
  static gsl_vector *       w         = gsl_vector_calloc(3);

  static gsl_matrix *       A1        = gsl_matrix_calloc(3, 3);
  static gsl_matrix *       A2        = gsl_matrix_calloc(3, 3);
  static gsl_matrix *       Gamma1    = gsl_matrix_calloc(3, 3);
  static gsl_matrix *       Gamma2    = gsl_matrix_calloc(3, 3);
  static gsl_matrix *       Delta     = gsl_matrix_calloc(3, 3);
  static gsl_matrix *       H         = gsl_matrix_calloc(3, 3);

  static double             rnorm;
  static double             r_c;

  static double             sigma1;
  static double             sigma2;
  static double             sigma12;
  static double             sigma1p3;
  static double             sigma2p3;
  static double             sigma12p3;

  static double             dH;
  static double             lambda;
  static double             nu;
  static double             h;
  static double             eta;
  static double             chi;
  static double             UA;
  static double             UR;

  static double             h_c;                                  /* for shifting */
  static double             UA_c;                                 /* for shifting */
  static double             UR_c;                                 /* for shifting */

  /* vars (Torque) */

  static double             drsigma2;
  static double             drdH;
  static double             dreta;
  static double             drchi;
  static double             drh;
  static double             drUA;
  static double             drUR;

  static double             drUA_c;                               /* for shifting */
  static double             drUR_c;                               /* for shifting */

  static gsl_vector *       torque    = gsl_vector_calloc(3);

  /* vars (Force) */

  static double             gamma, ddbl, dddbl;
  static double             dtsigma1;
  static double             dtsigma2;
  static double             dtdH;
  static double             dteta;
  static double             dtchi;
  static double             dth;
  static double             dtUA;
  static double             dtUR;

  static double             dth_c;                                /* for shifting */
  static double             dtUA_c;                               /* for shifting */
  static double             dtUR_c;                               /* for shifting */

  static gsl_vector *       force     = gsl_vector_calloc(3);

  /* temporary */

  static gsl_matrix *       dmat      = gsl_matrix_calloc(3, 3);
  static gsl_matrix *       ddmat     = gsl_matrix_calloc(3, 3);
  static gsl_matrix *       dddmat    = gsl_matrix_calloc(3, 3);
  static gsl_matrix *       Lambda    = gsl_matrix_calloc(3, 3);

  static gsl_vector *       dvec      = gsl_vector_calloc(3);
  static gsl_vector *       ddvec     = gsl_vector_calloc(3);
  static gsl_vector *       dddvec    = gsl_vector_calloc(3);
  static gsl_vector *       u         = gsl_vector_calloc(3);
  static gsl_vector *       p         = gsl_vector_calloc(3);

  static gsl_permutation *  perm      = gsl_permutation_alloc(3);

  /* bases and generators: that would be handy to have them around */

  extern gsl_vector *       e_x;
  extern gsl_vector *       e_y;
  extern gsl_vector *       e_z;

  extern gsl_matrix *       R_x;
  extern gsl_matrix *       R_y;
  extern gsl_matrix *       R_z;

  void init();

  bool is_init();

  int free_memory();

  /**** NOTE ********************************************************************************
   *                                                                                        *
   * In all cases, the transpose of the provided rotation matrices will be used!            *
   * (The orientation operators of the rigid bodies transform -a vector- (in the lab frame) *
   *  to the representation of a same vector in the body frame; however, the expressions    *
   *  given in the paper transform -the lab frame- to the body frame, hence transforming    *
   *  a vector in the body frame to the lab frame)                                          *
   *                                                                                        *
   ******************************************************************************************/

  /*** contact distance related *************************************************************/

  /* contact distance */
  double
  get_contact_distance(const gsl_matrix * A1, const gsl_matrix * A2,
    const gsl_vector * r12, const paramdb::param_cache_RE2_int * param);

  /* the lower bounds of the contact distance */
  double
  get_contact_distance_lb(const gsl_matrix * A1, const gsl_matrix * A2,
    const gsl_vector * r12, const paramdb::param_cache_RE2_int * param);

  /* translational variation on contact distance */
  double
  get_dt_contact_distance(const gsl_vector * dir, const paramdb::param_cache_RE2_int * param);

  /* rotational variation on contact distance */
  double
  get_dr_contact_distance(const gsl_matrix * gen, const paramdb::param_cache_RE2_int * param);

  /*** energy/force/torque related **********************************************************/

  /* interaction energy */
  double
  get_RE2_int_energy (const gsl_matrix * A1, const gsl_matrix * A2,
    const gsl_vector * r12, const paramdb::param_cache_RE2_int * param, const double & r_c,
    bool & eval_flag, int & stat);

  /* exerted force on M_2 */
  double
  get_RE2_int_force (const gsl_vector * dir, const paramdb::param_cache_RE2_int * param);

  /* exerted torque on M_2 with respect to its CM */
  double
  get_RE2_int_torque (const gsl_matrix * gen, const paramdb::param_cache_RE2_int * param);

}; /* re2 */

}; /* inter */

}; /* emd */

/***************************************************************************/

#endif /* INT_RE2_COMP_H */
