/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/***************************************************************************
 * todo:
 *
 * - faster force evaluation (one call)
 * - faster distance evaluation (distance from origin)
 *
 ***************************************************************************/

#include "int_re2_uni.h"

using namespace std;

/***************************************************************************/

vector<emd::uni::neighbor_data> *      emd::uni::re2::n_list_re2;
bool                                   emd::uni::re2::nl_init_flag = false;
double                                 emd::uni::re2::r_c          = DEF_RE2_R_C;
double                                 emd::uni::re2::r_s          = DEF_RE2_R_S;
double                                 emd::uni::re2::max_offset   = DEF_RE2_MAX_OFFSET;

/**** neighbor list related functions **************************************/

int
emd::uni::re2::init_neighbor_list()
{

  n_list_re2  =  new vector<neighbor_data>[rigid_bodies_count];

  /* reserving contiguous memory for neighbor data */
  for (unsigned int i=0; i<rigid_bodies_count; i++)
  {

    n_list_re2[i].reserve(DEF_RE2_MAX_NEIGHBORS);

  }

  nl_init_flag = true;

}

/***************************************************************************/

void
emd::uni::re2::update_neighbor_list()
{

  /* update the neighbor list */
  switch (nl_method)
  {

    case NL_VERLET:
    {
      if (use_cell_info)
        update_neighbor_list_verlet_using_cell_info();
      else
        update_neighbor_list_verlet();
    }

  }

}

/***************************************************************************/

void
emd::uni::re2::update_neighbor_list_verlet()
{

  double            rnorm;
  neighbor_data     n_dat;
  int               n_x;
  int               n_y;
  int               n_z;
  double            d_x, d_y, d_z, sig;

  /* clear the previous neighbor list */
  for (unsigned int m1=0; m1<rigid_bodies_count; m1++)
    n_list_re2[m1].clear();

  /* find the neighbors exclusively */
  for (unsigned int m1=0; m1<rigid_bodies_count-1; m1++)
  {

    /* check if the interaction is to be evaluated */
    if ( (rigid_bodies[m1].int_mask & MASK_RE2_INT) != MASK_RE2_INT ) continue;

    for (unsigned int m2=m1+1; m2<rigid_bodies_count; m2++)
    {

      /* check if the interaction is to be evaluated */
      if ( (rigid_bodies[m2].int_mask & MASK_RE2_INT) != MASK_RE2_INT ) continue;

      /* if both are inactive, ignore the interaction */
      if ( (rigid_bodies[m1].status==rigid_body::EMD_RIGIDBODY_STATUS_INACTIVE) &&
           (rigid_bodies[m2].status==rigid_body::EMD_RIGIDBODY_STATUS_INACTIVE) )
        continue;

      /* loop through the allowed images */
      for (img_it it = images.begin(); it!=images.end(); it++)
      {

        int n_x = (*it).img_x;
        int n_y = (*it).img_y;
        int n_z = (*it).img_z;

        d_x = gsl_vector_get(rigid_bodies[m2].x, 0) - gsl_vector_get(rigid_bodies[m1].x, 0) + n_x*x_width;
        d_y = gsl_vector_get(rigid_bodies[m2].x, 1) - gsl_vector_get(rigid_bodies[m1].x, 1) + n_y*y_width;
        d_z = gsl_vector_get(rigid_bodies[m2].x, 2) - gsl_vector_get(rigid_bodies[m1].x, 2) + n_z*z_width;

        gsl_vector_set(tmp_r12, 0, d_x);
        gsl_vector_set(tmp_r12, 1, d_y);
        gsl_vector_set(tmp_r12, 2, d_z);

        sig = r_c + r_s +
          emd::paramdb::re2_int_param[rigid_bodies[m1].idx][rigid_bodies[m2].idx].max_thickness;

        /* preliminary check */
        if ( (abs(d_x) < sig) && (abs(d_y) < sig) && (abs(d_z) < sig) )
        {

          /* exhaustive check: find the least contact distance */
          rnorm = emd::inter::re2::get_contact_distance(
            rigid_bodies[m1].R, rigid_bodies[m2].R, tmp_r12,
            &emd::paramdb::re2_int_param[rigid_bodies[m1].idx][rigid_bodies[m2].idx] );


          if ( rnorm < (r_c + r_s) )
          {

            n_dat.idx    =  m2;
            n_dat.img_x  =  n_x;
            n_dat.img_y  =  n_y;
            n_dat.img_z  =  n_z;

            n_list_re2[m1].push_back(n_dat);

            n_dat.idx    =  m1;
            n_dat.img_x  =  -n_x;
            n_dat.img_y  =  -n_y;
            n_dat.img_z  =  -n_z;

            n_list_re2[m2].push_back(n_dat);

          }

        }

      }

    }

  }

}

/***************************************************************************/

void
emd::uni::re2::update_neighbor_list_verlet_using_cell_info()
{

  double            rnorm;
  neighbor_data     n_dat;
  int               n_x, n_y, n_z;                /* cell vector */
  int               img_n_x, img_n_y, img_n_z;    /* periodic image idx vector */
  unsigned int      neighbor_cell_idx;
  int               cell_idx;
  unsigned int      m1, m2;
  double            d_x, d_y, d_z, sig;

  /* clear the previous neighbor list */
  for (unsigned int m1=0; m1<rigid_bodies_count; m1++)
    n_list_re2[m1].clear();

  cell_idx = 0;

  for (cell_it c_base = cells.begin(); c_base != cells.end(); c_base++)
  {

    n_x = (*c_base).n_x;
    n_y = (*c_base).n_y;
    n_z = (*c_base).n_z;

    /* loop through the neighboring cells */
    for (int nn_x = (n_x-1); nn_x <= (n_x+1); nn_x++)
      for (int nn_y = (n_y-1); nn_y <= (n_y+1); nn_y++)
        for (int nn_z = (n_z-1); nn_z <= (n_z+1); nn_z++)
        {
          /* zero out periodic image idx vector */
          img_n_x = 0;
          img_n_y = 0;
          img_n_z = 0;

          /* positive images */
          if ( x_periodic && (nn_x>=cell_num_x) )
            img_n_x = 1;
          if ( y_periodic && (nn_y>=cell_num_y) )
            img_n_y = 1;
          if ( z_periodic && (nn_z>=cell_num_z) )
            img_n_z = 1;

          /* negative images */
          if ( x_periodic && (nn_x<0) )
            img_n_x = -1;
          if ( y_periodic && (nn_y<0) )
            img_n_y = -1;
          if ( z_periodic && (nn_z<0) )
            img_n_z = -1;

          if ( (nn_x==n_x) && (nn_y==n_y) && (nn_z==n_z) ) /* same cell */
            neighbor_cell_idx = cell_idx;
          else
          {
            /* search for the appropriate cell */
            neighbor_cell_idx = get_cell_index( nn_x - img_n_x*cell_num_x,
                                                nn_y - img_n_y*cell_num_y,
                                                nn_z - img_n_z*cell_num_z );
          }

          if (neighbor_cell_idx == CELL_NOT_FOUND)
            continue;

          cell & neigh_cell = cells[neighbor_cell_idx];

          /* loop through the particles of the base cell */
          for ( list<unsigned int>::iterator it_1 = (*c_base).rigid_body_indices.begin();
                it_1 != (*c_base).rigid_body_indices.end();
                it_1++)
          {

            m1 = *it_1;

            /* check if the interaction is to be evaluated */
            if ( (rigid_bodies[m1].int_mask & MASK_RE2_INT) != MASK_RE2_INT ) continue;

            /* loop through the found neighbor cell */
            for ( list<unsigned int>::iterator it_2 = neigh_cell.rigid_body_indices.begin();
                  it_2 != neigh_cell.rigid_body_indices.end();
                  it_2++)
            {

              m2 = *it_2;

              /* only once */
              if (m2<=m1) continue;

               /* check if the interaction is to be evaluated */
              if ( (rigid_bodies[m2].int_mask & MASK_RE2_INT) != MASK_RE2_INT ) continue;

              /* if both are inactive, ignore the interaction */
              if ( (rigid_bodies[m1].status==rigid_body::EMD_RIGIDBODY_STATUS_INACTIVE) &&
                   (rigid_bodies[m2].status==rigid_body::EMD_RIGIDBODY_STATUS_INACTIVE) )
                continue;

              d_x = gsl_vector_get(rigid_bodies[m2].x, 0) - gsl_vector_get(rigid_bodies[m1].x, 0) +
                img_n_x*x_width;
              d_y = gsl_vector_get(rigid_bodies[m2].x, 1) - gsl_vector_get(rigid_bodies[m1].x, 1) +
                img_n_y*y_width;
              d_z = gsl_vector_get(rigid_bodies[m2].x, 2) - gsl_vector_get(rigid_bodies[m1].x, 2) +
                img_n_z*z_width;

              gsl_vector_set(tmp_r12, 0, d_x);
              gsl_vector_set(tmp_r12, 1, d_y);
              gsl_vector_set(tmp_r12, 2, d_z);

              sig = r_c + r_s +
                emd::paramdb::re2_int_param[rigid_bodies[m1].idx][rigid_bodies[m2].idx].max_thickness;

              /* preliminary check */
              if ( (abs(d_x) < sig) && (abs(d_y) < sig) && (abs(d_z) < sig) )
              {

                /* exhaustive check: find the least contact distance */
                rnorm = emd::inter::re2::get_contact_distance(
                  rigid_bodies[m1].R, rigid_bodies[m2].R, tmp_r12,
                  &emd::paramdb::re2_int_param[rigid_bodies[m1].idx][rigid_bodies[m2].idx] );

                if ( rnorm < (r_c + r_s) )
                {

                  n_dat.idx    =  m2;
                  n_dat.img_x  =  img_n_x;
                  n_dat.img_y  =  img_n_y;
                  n_dat.img_z  =  img_n_z;

                  n_list_re2[m1].push_back(n_dat);

                  n_dat.idx    =  m1;
                  n_dat.img_x  =  -img_n_x;
                  n_dat.img_y  =  -img_n_y;
                  n_dat.img_z  =  -img_n_z;

                  n_list_re2[m2].push_back(n_dat);

                }

              } /* preliminary check */

            } /* found neighbor cell inside loop */

          } /* base cell inside loop */

    } /* neighbor cell search loop */

    cell_idx++;

  } /* main cell loop */

}

/***************************************************************************/

void
emd::uni::re2::print_neighbor_list()
{

  cout << setprecision(3);

  cout << "* printing RE-squared interactions neighbor list: " << endl;

  for (int i=0; i<rigid_bodies_count; i++)
  {
    cout << "* Neighbors cout rigid body " << i << ": ";

    int sz = n_list_re2[i].size();
    if (sz==0)
      cout << "none" << endl;
    else
    {
      for (int j=0; j<sz-1; j++)
        cout << n_list_re2[i][j].idx << "(" << n_list_re2[i][j].img_x << n_list_re2[i][j].img_y
             << n_list_re2[i][j].img_z << "), ";
      cout << n_list_re2[i][sz-1].idx << "(" << n_list_re2[i][sz-1].img_x << n_list_re2[i][sz-1].img_y
           << n_list_re2[i][sz-1].img_z << ")" << endl;
    }
  }

  cout << setprecision(8);

}

/***************************************************************************/

/* the number of interactions */
unsigned long
emd::uni::re2::get_int_count()
{

  unsigned long num = 0;

  for (unsigned int m=0; m<rigid_bodies_count; m++)
    num += n_list_re2[m].size();

  num /= 2;

  return num;

}

/***************************************************************************/

 /*
  * the neighbor list (n_list_re2) is exclusively used for RE-squared
  * interactions. therefore, regardless of the @clear variable, the
  * @int_energy variable of neighbor list will be updated.
  */
double
emd::uni::re2::eval_RE2_single(const int& m1, const bool& exclusive, const bool& dyn, int& stat)
{

  int                                     m2, m3;
  double                                  U,total_U=0;
  emd::paramdb::param_cache_RE2_int *     param;

  /* if h<0, status is EXIT_FAILURE */
  int                  curr_stat = EXIT_SUCCESS;

  /* indicates if the energy was evaluated or not */
  bool                 eval_flag;

  /* init re2 workspace */
  if ( !emd::inter::re2::is_init() ) emd::inter::re2::init();

  for (neighbor_it it = n_list_re2[m1].begin(); it!=n_list_re2[m1].end(); it++)
  {

    m2 = (*it).idx;

    /* locating the respective image of the neighboring BODY */
    gsl_vector_set(img_vec, 0, (*it).img_x*x_width);
    gsl_vector_set(img_vec, 1, (*it).img_y*y_width);
    gsl_vector_set(img_vec, 2, (*it).img_z*z_width);

    /* evaluate each interaction only once */
    if ( exclusive && (m2>m1) ) continue;

    /* fetch parameter pointer */
    param = &emd::paramdb::re2_int_param[rigid_bodies[m1].idx][rigid_bodies[m2].idx];

    gsl_vector_memcpy (tmp_r12, rigid_bodies[m2].x);
    gsl_vector_sub    (tmp_r12, rigid_bodies[m1].x);
    gsl_vector_add    (tmp_r12, img_vec);          /* adding image guiding vector */

    /* evaluate the interaction energy */
    U = emd::inter::re2::get_RE2_int_energy(rigid_bodies[m1].R, rigid_bodies[m2].R,
      tmp_r12, param, r_c, eval_flag, curr_stat);
    if (curr_stat==EXIT_FAILURE) stat = EXIT_FAILURE;

    /* if the interaction is outside the cutoff distance, no need to evaluate force/torque
     * as well */
    if (!eval_flag) continue;

    /* set the neighbor_data energy for m1 ...*/
    (*it).int_energy = U;

    /* ... and m2 */
    for (neighbor_it it2 = n_list_re2[m2].begin(); it2!=n_list_re2[m2].end(); it2++)
    {
      m3 = (*it2).idx;
      if (m3==m1)
      {
        (*it2).int_energy = U;
      }
    }

    /* append to the rigid_body energy */
    rigid_bodies[m1].potential += U;
    rigid_bodies[m2].potential += U;
    total_U += U;

#ifdef BODYUNIVERSE_DEBUG_NUM
    cout << "* emd::uni::eval_RE2_single: Int. energy between " << m1 << " and "
         << m2 << ": " << U << endl;
#endif

#ifdef CHECK_OVERFLOW_RE2_INT
    if ( abs(U) > OVERFLOW_ENERGY_RE2_INT )
      cout << "* emd::uni::eval_RE2_single: Int. energy overflow (" << m1 << ", " << m2
           << ")." << endl;
#endif

    /* the variable space is available in the static memory of re2_ws class */
    if (dyn)
    {

      /* force */
      gsl_vector_set(tmp_force, 0, emd::inter::re2::get_RE2_int_force(emd::inter::re2::e_x, param));
      gsl_vector_set(tmp_force, 1, emd::inter::re2::get_RE2_int_force(emd::inter::re2::e_y, param));
      gsl_vector_set(tmp_force, 2, emd::inter::re2::get_RE2_int_force(emd::inter::re2::e_z, param));

      /* torque */
      gsl_vector_set(tmp_torque, 0, emd::inter::re2::get_RE2_int_torque(emd::inter::re2::R_x, param));
      gsl_vector_set(tmp_torque, 1, emd::inter::re2::get_RE2_int_torque(emd::inter::re2::R_y, param));
      gsl_vector_set(tmp_torque, 2, emd::inter::re2::get_RE2_int_torque(emd::inter::re2::R_z, param));

      /* adding up the current force contribution */
      gsl_vector_add(rigid_bodies[m2].force, tmp_force);
      gsl_vector_sub(rigid_bodies[m1].force, tmp_force);

      /* adding up the current torque contribution */
      gsl_vector_add(rigid_bodies[m2].torque, tmp_torque);
      gsl_vector_sub(rigid_bodies[m1].torque, tmp_torque);

      /* adding up the CM-force contribution to the torque */
      /* N(2->1) = -N(1->2) - r(1->2) X F(1->2) */
      gsl_vector_set(tmp_v, 0, gsl_vector_get(tmp_r12, 1)*gsl_vector_get(tmp_force, 2) -
                              gsl_vector_get(tmp_r12, 2)*gsl_vector_get(tmp_force, 1));
      gsl_vector_set(tmp_v, 1, gsl_vector_get(tmp_r12, 2)*gsl_vector_get(tmp_force, 0) -
                              gsl_vector_get(tmp_r12, 0)*gsl_vector_get(tmp_force, 2));
      gsl_vector_set(tmp_v, 2, gsl_vector_get(tmp_r12, 0)*gsl_vector_get(tmp_force, 1) -
                              gsl_vector_get(tmp_r12, 1)*gsl_vector_get(tmp_force, 0));

      gsl_vector_sub(rigid_bodies[m1].torque, tmp_v);

      /* virial components */
      vir_x += (gsl_vector_get(tmp_r12, 0) * gsl_vector_get(tmp_force, 0));
      vir_y += (gsl_vector_get(tmp_r12, 1) * gsl_vector_get(tmp_force, 1));
      vir_z += (gsl_vector_get(tmp_r12, 2) * gsl_vector_get(tmp_force, 2));

    }

  }

  return total_U;

}

/*******************************************************************************************/

double
emd::uni::re2::eval_RE2(const bool& dyn, int& stat)
{

  int        curr_stat     = EXIT_SUCCESS;

  /* init re2 workspace */
  if ( !emd::inter::re2::is_init() ) emd::inter::re2::init();

  double     old_potential = potential;

  for (int m1=0; m1<rigid_bodies_count; m1++)
  {
    /* do it exclusively (hence, don't clear previous evaluations) */
    potential += eval_RE2_single(m1, true, dyn, curr_stat);
    if (curr_stat==EXIT_FAILURE) stat = EXIT_FAILURE;
  }

  return (potential - old_potential);

}

/*******************************************************************************************/

int
emd::uni::re2::free_memory()
{

  /* static memory */
/*  gsl_vector_free(tmp_r12);
  gsl_vector_free(tmp_force);
  gsl_vector_free(tmp_torque);
  gsl_vector_free(tmp_v);
  gsl_vector_free(img_vec); */

  if (nl_init_flag)
  {
    delete[] n_list_re2;
  }

  return EXIT_SUCCESS;

}
