/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "moluniverse.h"

using namespace std;

/***** memory allocation for extern variables *****************************************/

/* interaction bitmasks */
const int         emd::uni::MASK_RE2_INT            = 2;
const int         emd::uni::MASK_EES_INT            = 4;
const int         emd::uni::MASK_EES_SPHERICAL_INT  = 8;
const int         emd::uni::MASK_GB_INT             = 16;
const int         emd::uni::MASK_DIPOLE_INT         = 32;

unsigned int      emd::uni::rigid_bodies_count      = 0;   /* number of bodys in the system */
unsigned int      emd::uni::flat_int_surfs_count    = 0;   /* flat surfaces */
unsigned int      emd::uni::sph_int_surfs_count     = 0;   /* spherical surfaces */

bool              emd::uni::nl_init_flag            = false;

bool              emd::uni::inter_RE2_present       = false;
bool              emd::uni::inter_GB_present        = false;
bool              emd::uni::inter_EES_present       = false;
bool              emd::uni::inter_dipole_present    = false;

/*******************************************************************************************/

emd::rigid_body *             emd::uni::rigid_bodies;      /* the central array of rigid bodys */
emd::flat_int_surface *       emd::uni::flat_int_surfs;    /* the central array of flat interacting surfaces */
emd::spherical_int_surface *  emd::uni::sph_int_surfs;     /* the central array of spherical interacting surfaces */

/*******************************************************************************************/

double        emd::uni::kinetic_r;         /* kinetic energy (rot) */
double        emd::uni::kinetic_t;         /* kinetic energy (trans) */
double        emd::uni::kinetic;           /* kinetic energy */
double        emd::uni::potential;         /* potential energy */
double        emd::uni::volume;            /* current volume of the system */
double        emd::uni::vir_x;             /* Virial in x direction */
double        emd::uni::vir_y;             /* Virial in y direction */
double        emd::uni::vir_z;             /* Virial in z direction */
bool          emd::uni::box_closed_flag;

/*******************************************************************************************/

emd::uni::neighbor_list_method   emd::uni::nl_method;      /* neighbor list method */
vector<emd::uni::image_index>    emd::uni::images;         /* vector of images */

/*******************************************************************************************/

bool          emd::uni::x_periodic;        /* periodic in x direction */
bool          emd::uni::y_periodic;        /* periodic in y direction */
bool          emd::uni::z_periodic;        /* periodic in z direction */

double        emd::uni::x_upper_bound;     /* x upper bound (periodic geom.) */
double        emd::uni::x_lower_bound;     /* x lower bound (periodic geom.) */
double        emd::uni::x_width;           /* x_upper_bound - x_lower_bound */
double        emd::uni::x_center;          /* x co-ordinate of the center of the box */

double        emd::uni::y_upper_bound;     /* y upper bound (periodic geom.) */
double        emd::uni::y_lower_bound;     /* y lower bound (periodic geom.) */
double        emd::uni::y_width;           /* y_upper_bound - y_lower_bound */
double        emd::uni::y_center;          /* x co-ordinate of the center of the box */

double        emd::uni::z_upper_bound;     /* z upper bound (periodic geom.) */
double        emd::uni::z_lower_bound;     /* z lower bound (periodic geom.) */
double        emd::uni::z_width;           /* z_upper_bound - z_lower_bound */
double        emd::uni::z_center;          /* x co-ordinate of the center of the box */

/* an interior point of the simulation box */
gsl_vector *  emd::uni::interior_point     = gsl_vector_calloc(3);

/* CM vector */
gsl_vector *  emd::uni::r_cm               = gsl_vector_calloc(3);

/* global constant electrical field (set to zero initially) */
gsl_vector *  emd::uni::E_field            = gsl_vector_calloc(3);

/*******************************************************************************************/

int
emd::uni::init_universe_pre()
{

  rigid_bodies_count =  0;
  kinetic            =  0;
  potential          =  0;

  x_periodic         =  false;
  x_lower_bound      =  -20;
  x_upper_bound      =  20;
  x_width            =  40;

  y_periodic         =  false;
  y_lower_bound      =  -20;
  y_upper_bound      =  20;
  y_width            =  40;

  z_periodic         =  false;
  z_lower_bound      =  -20;
  z_upper_bound      =  20;
  z_width            =  40;

  nl_init_flag       =  false;

  inter_RE2_present    = false;
  inter_GB_present     = false;
  inter_EES_present    = false;
  inter_dipole_present = false;

  /* NOTE initialize interaction workspaces here */
  emd::inter::re2::init();
  emd::inter::gb::init();
  emd::inter::ees::init();
  emd::inter::dipole::init();

#ifdef MOLUNIVERSE_DEBUG
  cout << "* emd::uni: initialized. " << endl;
#endif

}

/*******************************************************************************************/

int
emd::uni::init_universe_post()
{

  /* fold to box */
  emd::uni::fold_to_box();

  /* examine the present interaction */
  for (int m=0; m<rigid_bodies_count; m++)
  {

    if ( (rigid_bodies[m].int_mask & MASK_RE2_INT) == MASK_RE2_INT )
      inter_RE2_present = true;

    if ( (rigid_bodies[m].int_mask & MASK_GB_INT) == MASK_GB_INT )
      inter_GB_present = true;

    if ( (rigid_bodies[m].int_mask & MASK_EES_INT) == MASK_EES_INT )
      inter_EES_present = true;

    if ( (rigid_bodies[m].int_mask & MASK_EES_SPHERICAL_INT) == MASK_EES_SPHERICAL_INT )
      inter_EES_present = true;

    if ( (rigid_bodies[m].int_mask & MASK_DIPOLE_INT) == MASK_DIPOLE_INT )
      inter_dipole_present = true;

  }

  return EXIT_SUCCESS;

}

/*******************************************************************************************/

/* memory allocation for rigid bodies */
int
emd::uni::malloc_rigid_body(const int& num)
{

  if (rigid_bodies_count>0)
  {
    cerr << "* emd::uni::malloc_rigid_body: memory already allocated!" << endl;
    return EXIT_FAILURE;
  }

  if (num<=0)
  {
    cerr << "* emd::uni::malloc_rigid_body: positive values please!" << endl;
    return EXIT_FAILURE;
  }

  rigid_bodies_count = num;
  rigid_bodies       = new rigid_body[num];

#ifdef MOLUNIVERSE_DEBUG
  cout << "* emd::uni::malloc_rigid_body: memory allocated for " << num << " rigid bodies."
       << endl;
#endif

  return EXIT_SUCCESS;
}

/*******************************************************************************************/

/* memory allocation for flat int surfaces */
int
emd::uni::malloc_flat_int_surf(const int& num)
{

  if (flat_int_surfs_count>0)
  {
    cerr << "* emd::uni::malloc_flat_int_surf: memory already allocated!" << endl;
    return EXIT_FAILURE;
  }

  if (num<=0)
  {
    cerr << "* emd::uni::malloc_flat_int_surf: positive values please!" << endl;
    return EXIT_FAILURE;
  }

  flat_int_surfs_count = num;
  flat_int_surfs       = new flat_int_surface[num];

#ifdef MOLUNIVERSE_DEBUG
  cout << "* emd::uni::malloc_flat_int_surf: memory allocated for " << num << " flat int. surfaces."
       << endl;
#endif

  return EXIT_SUCCESS;
}

/*******************************************************************************************/

/* memory allocation for spherical int surfaces */
int
emd::uni::malloc_spherical_int_surf(const int& num)
{

  if (sph_int_surfs_count>0)
  {
    cerr << "* emd::uni::malloc_spherical_int_surf: memory already allocated!" << endl;
    return EXIT_FAILURE;
  }

  if (num<=0)
  {
    cerr << "* emd::uni::malloc_spherical_int_surf: positive values please!" << endl;
    return EXIT_FAILURE;
  }

  sph_int_surfs_count = num;
  sph_int_surfs       = new spherical_int_surface[num];

#ifdef MOLUNIVERSE_DEBUG
  cout << "* emd::uni::malloc_spherical_int_surf: memory allocated for " << num << " spherical int. surfaces."
       << endl;
#endif

  return EXIT_SUCCESS;
}

/*******************************************************************************************/

int
emd::uni::init_neighbor_list(const neighbor_list_method& nl_method0)
{

  nl_method  =  nl_method0;

  /* create the image index */
  images.clear();
  image_index  img_idx;

  for (int k_x=-1; k_x<=1; k_x++)
    for (int k_y=-1; k_y<=1; k_y++)
      for (int k_z=-1; k_z<=1; k_z++)
      {

        if (!x_periodic && k_x!=0) continue;
        if (!y_periodic && k_y!=0) continue;
        if (!z_periodic && k_z!=0) continue;

        img_idx.img_x = k_x;
        img_idx.img_y = k_y;
        img_idx.img_z = k_z;

        images.push_back(img_idx);

      }

  /* NOTE init the neighbor list routines of all interactions */

  /* re2 */
  emd::uni::re2::init_neighbor_list();

  /* gb */
  emd::uni::gb::init_neighbor_list();

  /* ees */
  emd::uni::ees::init_neighbor_list();

  return EXIT_SUCCESS;

}


/*******************************************************************************************/

void
emd::uni::fold_to_box()
{

  for (int i=0; i<rigid_bodies_count; i++)
    fold_to_box(i);

}

/*******************************************************************************************/

void
emd::uni::fold_to_box(const int& idx)
{

  double   coord;
  int      idx2;
  int      per;        /* periodicity */

  /* x direction */
  if (x_periodic)
  {

    coord  =  gsl_vector_get(rigid_bodies[idx].x, 0);
    per    =  (int)( floor ( (coord - x_lower_bound) / x_width ) );

    if (per!=0)
      gsl_vector_set(rigid_bodies[idx].x, 0, coord - per*x_width);

  }

  /* y direction */
  if (y_periodic)
  {

    coord  =  gsl_vector_get(rigid_bodies[idx].x, 1);
    per    =  (int)( floor( (coord - y_lower_bound ) / y_width ) );

    if (per!=0)
      gsl_vector_set(rigid_bodies[idx].x, 1, coord - per*y_width);

  }

  /* z direction */
  if (z_periodic)
  {

    coord  =  gsl_vector_get(rigid_bodies[idx].x, 2);
    per    =  (int)( floor(  (coord - z_lower_bound) / z_width ) );

    if (per!=0)
      gsl_vector_set(rigid_bodies[idx].x, 2, coord - per*z_width);

  }

}

/*******************************************************************************************/

/* evaluate the volume of the simulation box
 *
 * NOTE:
 *
 *  the spheres must NOT intersect with the planes!
 *
 */
int
emd::uni::eval_volume()
{

  /* planes */
  vector<emd::geom::plane_3d>  planes;
  emd::geom::plane_3d          new_plane;
  int                          stat  =  EXIT_SUCCESS;

  if (x_periodic)
  {
    new_plane.nx = 1;  new_plane.ny = 0;  new_plane.nz = 0;
    new_plane.d  = x_lower_bound;
    planes.push_back(new_plane);
    new_plane.d  = x_upper_bound;
    planes.push_back(new_plane);
  }

  if (y_periodic)
  {
    new_plane.nx = 0;  new_plane.ny = 1;  new_plane.nz = 0;
    new_plane.d  = y_lower_bound;
    planes.push_back(new_plane);
    new_plane.d  = y_upper_bound;
    planes.push_back(new_plane);
  }

  if (z_periodic)
  {
    new_plane.nx = 0;  new_plane.ny = 0;  new_plane.nz = 1;
    new_plane.d  = z_lower_bound;
    planes.push_back(new_plane);
    new_plane.d  = z_upper_bound;
    planes.push_back(new_plane);
  }

  /* flat int surfaces */
  for (int i=0; i<flat_int_surfs_count; i++)
  {
    new_plane.nx = -gsl_vector_get( flat_int_surfs[i].n, 0 );
    new_plane.ny = -gsl_vector_get( flat_int_surfs[i].n, 1 );
    new_plane.nz = -gsl_vector_get( flat_int_surfs[i].n, 2 );
    gsl_blas_ddot( flat_int_surfs[i].n, flat_int_surfs[i].r0, &(new_plane.d) );
    new_plane.d *= -1;
    planes.push_back(new_plane);
  }

  geom::point_3d interior_point_3d = { gsl_vector_get(interior_point, 0),
                                       gsl_vector_get(interior_point, 1),
                                       gsl_vector_get(interior_point, 2) };

  /* evaluate the volume bounded by planes and periodic bounds */
  stat = emd::geom::eval_interior_volume( planes, interior_point_3d, volume );

  /* substract spheres */
  for (int i=0; i<sph_int_surfs_count; i++)
    volume -= ( 4 * M_PI * gsl_pow_3 (sph_int_surfs[i].rad) / 3 );

  return stat;

}

/*******************************************************************************************/

int
emd::uni::eval_CM_vector()
{

  gsl_vector *   tmp_v = gsl_vector_alloc(3);
  double         mass  =  0.0;

  int m;
  for (long int m=0; m<rigid_bodies_count; m++)
  {

    gsl_vector_memcpy(tmp_v, rigid_bodies[m].x);
    gsl_vector_scale(tmp_v, rigid_bodies[m].param->mass);

    gsl_vector_add(r_cm, tmp_v);
    mass += rigid_bodies[m].param->mass;
  }

  /* evaluate CM vector */
  gsl_vector_scale(r_cm, 1.0/mass);

  gsl_vector_free(tmp_v);

}

/*******************************************************************************************/

/* setups the universe from a given file */
int
emd::uni::load_universe(const string & input_file_name, const bool & clear)
{

  init_universe_pre();

  int stat = loader::load_file(input_file_name, clear);
  if (stat==EXIT_FAILURE) return EXIT_FAILURE;

  return EXIT_SUCCESS;

}

/*******************************************************************************************/

/* loads the parameter set */
int
emd::uni::load_param_file(const std::string & param_file, const bool & clear)
{
  return emd::paramdb::load_param_file(param_file, clear);
}

/*******************************************************************************************/

void
emd::uni::clear_evals()
{

  potential = 0;
  kinetic   = 0;
  vir_x     = 0;
  vir_y     = 0;
  vir_z     = 0;

  /* clear out rigid bodys forces/torques */
  for (int i=0; i<rigid_bodies_count; i++)
  {

    rigid_bodies[i].potential = 0;

    if (rigid_bodies[i].is_ext_alloc())
    {
      gsl_vector_set_zero(rigid_bodies[i].force);
      gsl_vector_set_zero(rigid_bodies[i].torque);
    }

  }

  /* clear out the forces on surfaces */
  for (int s=0; s<flat_int_surfs_count; s++)
  {
    gsl_vector_set_zero(flat_int_surfs[s].force);
  }

  for (int s=0; s<sph_int_surfs_count; s++)
  {
    gsl_vector_set_zero(sph_int_surfs[s].force);
  }

}

/*******************************************************************************************/

/* NOTE intermolecular interactions here */

int
emd::uni::eval_all_inter(const bool& dyn)
{

  int curr_stat;
  int stat         =  EXIT_SUCCESS;

  /* RE^2 ineteractions */
  if (inter_RE2_present) re2::eval_RE2(dyn, curr_stat);

  /* GB ineteractions */
  if (inter_GB_present) gb::eval_GB(dyn, curr_stat);

  if (curr_stat==EXIT_FAILURE) stat = EXIT_FAILURE;

  return stat;

}

/*******************************************************************************************/

/* NOTE external interactions here */

int
emd::uni::eval_all_extern(const bool& dyn)
{

  int curr_stat;
  int stat         =  EXIT_SUCCESS;

  /* dipole ineteractions */
  if (inter_dipole_present) dipole::eval_dipole(dyn);

  if (curr_stat==EXIT_FAILURE) stat = EXIT_FAILURE;

  return stat;

}

/*******************************************************************************************/

/* NOTE surface interactions here */

int
emd::uni::eval_all_surf(const bool& dyn)
{

  int curr_stat;
  int stat         =  EXIT_SUCCESS;

  /* EES ineteractions */
  if (inter_EES_present) ees::eval_EES(dyn, curr_stat);

  if (curr_stat==EXIT_FAILURE) stat = EXIT_FAILURE;

  return stat;

}

/*******************************************************************************************/

/* NOTE update neighbor lists */

void
emd::uni::update_neighbor_lists()
{

  /* RE^2 ineteractions */
  if (inter_RE2_present) emd::uni::re2::update_neighbor_list();

  /* GB ineteractions */
  if (inter_GB_present) emd::uni::gb::update_neighbor_list();

  /* EES ineteractions */
  if (inter_EES_present) emd::uni::ees::update_neighbor_list();

}

/*******************************************************************************************/

double
emd::uni::eval_kinetic()
{

  kinetic_r = 0;
  kinetic_t = 0;
  kinetic   = 0;

  double tmp;

  for (int m=0; m<rigid_bodies_count; m++)
    if (rigid_bodies[m].status==rigid_body::EMD_RIGIDBODY_STATUS_ACTIVE &&
        rigid_bodies[m].is_ext_alloc())
    {

      /* CM kinetic energy */
      kinetic_t +=
        0.5*rigid_bodies[m].param->mass*( gsl_pow_2( gsl_vector_get(rigid_bodies[m].v, 0) ) +
                                          gsl_pow_2( gsl_vector_get(rigid_bodies[m].v, 1) ) +
                                          gsl_pow_2( gsl_vector_get(rigid_bodies[m].v, 2) ) );

      /* rotational kinetic energy */
      kinetic_r +=
        0.5 * ( rigid_bodies[m].omega.dat[1]*rigid_bodies[m].l.dat[1] +
                rigid_bodies[m].omega.dat[2]*rigid_bodies[m].l.dat[2] +
                rigid_bodies[m].omega.dat[3]*rigid_bodies[m].l.dat[3] );

    }

  kinetic = kinetic_t + kinetic_r;

  return kinetic;

}

/*******************************************************************************************/

void
emd::uni::print_inter_details()
{
  cout << "* emd::uni::print_inter_details:" << endl;

  for (int i=0; i<rigid_bodies_count; i++)
  {

    cout << "* rigid_bodies[" << i << "] : " << endl;
    cout << "- potential: " << rigid_bodies[i].potential << endl;

    if (rigid_bodies[i].is_ext_alloc())
    {

      cout << "- force: " << "("
           << gsl_vector_get(rigid_bodies[i].force, 0) << ", "
           << gsl_vector_get(rigid_bodies[i].force, 1) << ", "
           << gsl_vector_get(rigid_bodies[i].force, 2) << ")" << endl;

      cout << "- torque: " << "("
           << gsl_vector_get(rigid_bodies[i].torque, 0) << ", "
           << gsl_vector_get(rigid_bodies[i].torque, 1) << ", "
           << gsl_vector_get(rigid_bodies[i].torque, 2) << ")" << endl;
    }

  }

}

/*******************************************************************************************/

int
emd::uni::free_memory()
{

  emd::paramdb::free_memory();
  emd::uni::re2::free_memory();
  emd::uni::gb::free_memory();
  emd::uni::ees::free_memory();

  if (rigid_bodies_count>0) delete[] rigid_bodies;
  if (flat_int_surfs_count>0) delete[] flat_int_surfs;
  if (sph_int_surfs_count>0) delete[] sph_int_surfs;

  rigid_bodies_count = 0;
  flat_int_surfs_count = 0;
  sph_int_surfs_count = 0;

#ifdef MOLUNIVERSE_DEBUG
  cout << "* emd::uni: destructed." << endl;
#endif

}
