/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef MOLUNIVERSE_H
#define MOLUNIVERSE_H

#include <iostream>
#include <iomanip>
#include <vector>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>

#include <quaternion/gsl_quaternion.h>

#include "consts.h"
#include "gslext.h"
#include "rigid_body.h"
#include "surface.h"
#include "paramdb.h"
#include "stringext.h"
#include "neighbor_data.h"
#include "image_index.h"
#include "universeloader.h"
#include "cell_grid.h"
#include "geom.h"

/***** interactions *********************/

#include "int_re2_comp.h"
#include "int_gb_comp.h"
#include "int_ees_comp.h"
#include "int_dipole_comp.h"

#include "int_re2_uni.h"
#include "int_gb_uni.h"
#include "int_ees_uni.h"
#include "int_dipole_uni.h"

/***** computational routines ***********/

#include "moluniverse_comp.h"

/****************************************/

//#define MOLUNIVERSE_DEBUG

/****************************************/

namespace emd
{

namespace uni
{

  /************************************************************************************/

  extern const int                MASK_RE2_INT;
  extern const int                MASK_GB_INT;
  extern const int                MASK_EES_INT;
  extern const int                MASK_EES_SPHERICAL_INT;
  extern const int                MASK_DIPOLE_INT;

  /************************************************************************************/

  extern unsigned int             rigid_bodies_count;        /* number of bodies in the system */
  extern unsigned int             flat_int_surfs_count;      /* flat surfaces */
  extern unsigned int             sph_int_surfs_count;       /* spherical surfaces */

  extern bool                     nl_init_flag;

  /************************************************************************************/

  enum neighbor_list_method
  {
    NL_VERLET
  };

  /************************************************************************************/

  typedef  std::vector<neighbor_data>::iterator   neighbor_it;
  typedef  std::vector<image_index>::iterator     img_it;

  /************************************************************************************/

  extern rigid_body *               rigid_bodies;      /* the central array of rigid bodies */
  extern flat_int_surface *         flat_int_surfs;    /* the central array of flat int. surfaces */
  extern spherical_int_surface *    sph_int_surfs;     /* the central array of sph. int. surfaces */

  extern double                     kinetic_r;         /* kinetic energy (rot) */
  extern double                     kinetic_t;         /* kinetic energy (trans) */
  extern double                     kinetic;           /* kinetic energy */
  extern double                     potential;         /* potential energy */
  extern double                     volume;            /* effective volume */
  extern bool                       box_closed_flag;

  extern double                     vir_x;             /* Virial (x component) */
  extern double                     vir_y;             /* Virial (y component) */
  extern double                     vir_z;             /* Virial (z component) */

  /************************************************************************************/

  extern neighbor_list_method       nl_method;         /* neighbor list method */

  extern std::vector<image_index>  images;

  /************************************************************************************/

  extern bool                       inter_RE2_present;
  extern bool                       inter_GB_present;
  extern bool                       inter_EES_present;
  extern bool                       inter_dipole_present;

  /************************************************************************************/

  extern bool                       x_periodic;        /* periodic in x direction */
  extern bool                       y_periodic;        /* periodic in y direction */
  extern bool                       z_periodic;        /* periodic in z direction */

  extern double                     x_upper_bound;     /* x upper bound (periodic geom.) */
  extern double                     x_lower_bound;     /* x lower bound (periodic geom.) */
  extern double                     x_width;           /* x_upper_bound - x_lower_bound */
  extern double                     x_center;          /* x co-ordinate of the center of the box */

  extern double                     y_upper_bound;     /* y upper bound (periodic geom.) */
  extern double                     y_lower_bound;     /* y lower bound (periodic geom.) */
  extern double                     y_width;           /* y_upper_bound - y_lower_bound */
  extern double                     y_center;          /* x co-ordinate of the center of the box */

  extern double                     z_upper_bound;     /* z upper bound (periodic geom.) */
  extern double                     z_lower_bound;     /* z lower bound (periodic geom.) */
  extern double                     z_width;           /* z_upper_bound - z_lower_bound */
  extern double                     z_center;          /* x co-ordinate of the center of the box */

  extern gsl_vector *               interior_point;    /* an interior point of the simulation box */
  extern gsl_vector *               r_cm;              /* CM vector */

  /************************************************************************************/

  extern gsl_vector *               E_field;           /* global constant electrical field */

  /************************************************************************************/

  /* preliminary initialization */
  int init_universe_pre();

  /* final initialization (once the universe was constructed) */
  int init_universe_post();

  /* allocate memory for rigid bodies */
  int malloc_rigid_body(const int& num);

  /* allocate memory for flat spherical surfaces */
  int malloc_flat_int_surf(const int& num);

  /* allocate memory for spherical spherical surfaces */
  int malloc_spherical_int_surf(const int& num);

  /* loads a parameter file */
  int load_param_file(const std::string & param_file, const bool & clear = true);

  /* loads a universe */
  int load_universe(const std::string & input_file_name, const bool & clear = true);

  /* evaluate the accessible volume */
  int eval_volume();

  /* evaluate CM vector */
  int eval_CM_vector();

  /************************************************************************************/

  /* init neighbor list algorithm */
  int init_neighbor_list(const neighbor_list_method& type);

  void fold_to_box();
  void fold_to_box(const int& idx);

  /************************************************************************************/

  /******* energy/force/torque evaluation *********************************************/

  /* clear the previous evaluations and zero out all force/torque vectors */
  void clear_evals();

  /* evaluate all of the intermolecular interactions */
  int eval_all_inter(const bool& dyn);

  /* evaluate all of the external interactions */
  int eval_all_extern(const bool& dyn);

  /* evaluate all of the surface interactions */
  int eval_all_surf(const bool& dyn);

  /******* energy/force/nl related utility functions **********************************/

  /* update all neighbor lists */
  void update_neighbor_lists();

  /************************************************************************************
   * this function must be called after a call to single in order to                  *
   *  sync the neighbor data enegies.                                                 *
   ************************************************************************************/
  void sync_neighbor_data_energies(const int& idx);

  void print_inter_details();

  /******* dynamical evaluations*******************************************************/

  /* evaluate the kinetic energy */
  double eval_kinetic();

  /************************************************************************************/

  /* nukes the universe! */
  int free_memory();

};

};

#endif /* MOLUNIVERSE_H */
