/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "moluniverse_comp.h"

using namespace std;

/***************************************************************************/

int emd::uni::eval_Q_tensor( std::vector<unsigned int> & indices, const int & axis,
    gsl_matrix * Q_tensor )
  {

    gsl_vector * orig_axis                = gsl_vector_alloc(3);
    gsl_vector * rot_axis                 = gsl_vector_alloc(3);
    gsl_matrix * c_Q_tensor               = gsl_matrix_alloc(3, 3);

    switch (axis)
      {

    case 0:
      {
        gsl_vector_set(orig_axis, 0, 1);
        gsl_vector_set(orig_axis, 1, 0);
        gsl_vector_set(orig_axis, 2, 0);

        break;
      }

    case 1:
      {
        gsl_vector_set(orig_axis, 0, 0);
        gsl_vector_set(orig_axis, 1, 1);
        gsl_vector_set(orig_axis, 2, 0);

        break;
      }

    case 2:
      {
        gsl_vector_set(orig_axis, 0, 0);
        gsl_vector_set(orig_axis, 1, 0);
        gsl_vector_set(orig_axis, 2, 1);

        break;
      }

      }

    rigid_body * rb;

    gsl_matrix_set_zero(Q_tensor);
    gsl_matrix_set_zero(c_Q_tensor);

    double q00, q11, q22, q01, q12, q20;

    /* evaluating the tensor elements */
    for ( vector<unsigned int>::iterator it = indices.begin();
    it != indices.end(); it++)
      {

        rb = &(rigid_bodies[(*it)]);

        /* find the rotated axis */
        gsl_blas_dgemv( CblasNoTrans, 1, rb->R, orig_axis, 0, rot_axis);

        q00 = gsl_pow_2( gsl_vector_get(rot_axis, 0) );
        q11 = gsl_pow_2( gsl_vector_get(rot_axis, 1) );
        q22 = gsl_pow_2( gsl_vector_get(rot_axis, 2) );
        q01 = gsl_vector_get(rot_axis, 0) * gsl_vector_get(rot_axis, 1);
        q12 = gsl_vector_get(rot_axis, 1) * gsl_vector_get(rot_axis, 2);
        q20 = gsl_vector_get(rot_axis, 2) * gsl_vector_get(rot_axis, 0);

        /* diagonal elements */
        gsl_matrix_set( c_Q_tensor, 0, 0, 1.5 * q00 - 0.5 );
        gsl_matrix_set( c_Q_tensor, 1, 1, 1.5 * q11 - 0.5 );
        gsl_matrix_set( c_Q_tensor, 2, 2, 1.5 * q22 - 0.5 );

        /* off diagonal elements */
        gsl_matrix_set( c_Q_tensor, 0, 1, 1.5 * q01 );
        gsl_matrix_set( c_Q_tensor, 1, 2, 1.5 * q12 );
        gsl_matrix_set( c_Q_tensor, 2, 0, 1.5 * q20 );

        gsl_matrix_add( Q_tensor, c_Q_tensor );

      }

    /* set the lower triangle */
    gsl_matrix_set( Q_tensor, 1, 0, gsl_matrix_get( Q_tensor, 0, 1) );
    gsl_matrix_set( Q_tensor, 2, 1, gsl_matrix_get( Q_tensor, 1, 2) );
    gsl_matrix_set( Q_tensor, 0, 2, gsl_matrix_get( Q_tensor, 2, 0) );

    gsl_matrix_scale( Q_tensor, 1.0 / indices.size() );

    /* free the memory */
    gsl_vector_free(orig_axis);
    gsl_vector_free(rot_axis);
    gsl_matrix_free(c_Q_tensor);

    return EXIT_SUCCESS;

  }

/***************************************************************************/

int emd::uni::eval_Q_tensor( std::vector<unsigned int> & indices, const int & axis,
    gsl_matrix * Q_tensor, gsl_vector * Q_eigen_val, gsl_matrix * Q_eigen_vec )
  {

    gsl_matrix * c_Q_tensor               = gsl_matrix_alloc(3, 3);
    gsl_eigen_symmv_workspace * eigen_WS  = gsl_eigen_symmv_alloc(3);

    /* eval the Q_tensor */
    eval_Q_tensor(indices, axis, Q_tensor);

    /* make a temporary copy */
    gsl_matrix_memcpy(c_Q_tensor, Q_tensor);

    /* diagonalize */
    gsl_eigen_symmv(c_Q_tensor, Q_eigen_val, Q_eigen_vec, eigen_WS);

    /* free the memory */
    gsl_matrix_free(c_Q_tensor);
    gsl_eigen_symmv_free(eigen_WS);

    return EXIT_SUCCESS;

  }

/***************************************************************************/

int
emd::uni::eval_radial_dist( const int Class_id, const double & max_rad, const int & divs, gsl_vector * dir,
    gsl_vector * para_dist, gsl_vector * perp_dist,
    gsl_vector * scalar_dist)
  {

    if (eval_volume()==EXIT_FAILURE)
      return EXIT_FAILURE;

    gsl_vector * r12 = gsl_vector_alloc (3);

    double r_c = sqrt(2.0) * max_rad;

    int max_img_x = (int)ceil(r_c / x_width);
    int max_img_y = (int)ceil(r_c / y_width);
    int max_img_z = (int)ceil(r_c / z_width);

    gsl_vector_set_zero(para_dist);
    gsl_vector_set_zero(perp_dist);
    gsl_vector_set_zero(scalar_dist);

    /* cutoff length */

    for (int m1=0; m1 < rigid_bodies_count; m1++)
      {

        for (int m2=0; m2 < rigid_bodies_count; m2++)
          {

            for (int nx=-max_img_x; nx <= max_img_x ; nx++ )
              {
                for (int ny=-max_img_y; ny <= max_img_y ; ny++ )
                  {
                    for ( int nz=-max_img_z; nz <= max_img_z ; nz++ )
                      {

                        if ( (m1==m2) && (nx==0) && (ny==0) && (nz==0) ) continue;
                        //changed

                        if(emd::uni::rigid_bodies[m1].class_id != Class_id || emd::uni::rigid_bodies[m2].class_id != Class_id) continue;

                        gsl_vector_memcpy(r12, rigid_bodies[m1].x);
                        gsl_vector_sub(r12, rigid_bodies[m2].x);
                        gsl_vector_set(r12, 0, gsl_vector_get(r12, 0) - nx*x_width);
                        gsl_vector_set(r12, 1, gsl_vector_get(r12, 1) - ny*y_width);
                        gsl_vector_set(r12, 2, gsl_vector_get(r12, 2) - nz*z_width);

                        if ( (abs(gsl_vector_get(r12, 0))>r_c) || (abs(gsl_vector_get(r12, 1))>r_c)
                            || (abs(gsl_vector_get(r12, 2))>r_c) ) continue;

                        double s_scalar_dist = gsl_blas_dnrm2(r12);
                        double s_para_dist;
                        gsl_blas_ddot(r12, dir, &s_para_dist);
                        s_para_dist = abs(s_para_dist);
                        double s_perp_dist = sqrt( gsl_pow_2(s_scalar_dist) - gsl_pow_2(s_para_dist) );

                        int idx_scalar = (int)floor( (s_scalar_dist/max_rad) * divs);
                        int idx_para   = (int)floor( (s_para_dist/max_rad) * divs);
                        int idx_perp   = (int)floor( (s_perp_dist/max_rad) * divs);

                        if (idx_scalar < divs)
                          {
                            gsl_vector_set( scalar_dist, idx_scalar,
                                gsl_vector_get( scalar_dist, idx_scalar ) + 1);
                          }

                        if ( (idx_para < divs) && (idx_perp < divs) )
                          {
                            gsl_vector_set( para_dist, idx_para,
                                gsl_vector_get( para_dist, idx_para ) + 1);
                            gsl_vector_set( perp_dist, idx_perp,
                                gsl_vector_get( perp_dist, idx_perp) + 1);
                          }

                      };
                  };
              }; /* end of images loop */

          };
      }; /* end of rigid_bodies loop */

    double dr = max_rad / divs;

    /* normalization of the RDF's to the ideal gas */
    for(int idx = 0; idx < divs; idx++) 
      {

        double r_lower = idx * dr;
        double r_upper = r_lower + dr;

        /* scalar */
        double scl_rdf_ideal =
          4 * M_PI * rigid_bodies_count * rigid_bodies_count * ( gsl_pow_3(r_upper) - gsl_pow_3(r_lower) ) /
          ( 3 * volume );

        gsl_vector_set( scalar_dist, idx, gsl_vector_get( scalar_dist, idx ) / scl_rdf_ideal);

        /* parallel */
        double para_rdf_ideal =
          2 * M_PI * max_rad * max_rad * dr * rigid_bodies_count * rigid_bodies_count / volume;

        gsl_vector_set( para_dist, idx, gsl_vector_get( para_dist, idx ) / para_rdf_ideal);

        /* normal */
        double perp_rdf_ideal =
          2 * max_rad * M_PI * ( gsl_pow_2(r_upper) - gsl_pow_2(r_lower) ) *
          rigid_bodies_count * rigid_bodies_count / volume;

        gsl_vector_set( perp_dist, idx, gsl_vector_get( perp_dist, idx ) / perp_rdf_ideal);

      }

    gsl_vector_free(r12);

    return EXIT_SUCCESS;

  }

/***************************************************************************/

int
emd::uni::detect_planes( const int Class_id, const double & min_wl, double & k_x, double & k_y, double & k_z,
    double & phi, gsl_vector * director, const double & penalty_coeff)
  {

    /* normal to the planes */
    gsl_vector * n = gsl_vector_alloc(3);

    /* max wavelength */
    double max_wl = 0;

    if (x_width>max_wl) max_wl = x_width;
    if (y_width>max_wl) max_wl = y_width;
    if (z_width>max_wl) max_wl = z_width;

    /* max and min wavevector norm */
    double max_k_norm2 = gsl_pow_2(2*M_PI/min_wl);
    double min_k_norm2 = gsl_pow_2(2*M_PI/max_wl);
    double max_k_norm  = sqrt(max_k_norm2);
    double min_k_norm  = sqrt(min_k_norm2);

    double k_fact_x = 2*M_PI/x_width;
    double k_fact_y = 2*M_PI/y_width;
    double k_fact_z = 2*M_PI/z_width;

    int max_n_x = (int)ceil(max_k_norm/k_fact_x);
    int max_n_y = (int)ceil(max_k_norm/k_fact_y);
    int max_n_z = (int)ceil(max_k_norm/k_fact_z);

    double c_k_x, c_k_y, c_k_z, c_k_norm2;
    double max_power2, curr_power2, sin2_dir_k;
    gsl_complex curr_comp, max_comp;

    max_power2 = 0;

    /* exhaustive search in the reciprocal lattice vectors */
    for (int n_x=-max_n_x; n_x<=max_n_x; n_x++)
      for (int n_y=-max_n_y; n_y<=max_n_y; n_y++)
        for (int n_z=-max_n_z; n_z<=max_n_z; n_z++)
          {

            c_k_x      = n_x*k_fact_x;
            c_k_y      = n_y*k_fact_y;
            c_k_z      = n_z*k_fact_z;
            c_k_norm2  = c_k_x*c_k_x + c_k_y*c_k_y + c_k_z*c_k_z;

            if ( (c_k_norm2>max_k_norm2) || (c_k_norm2<min_k_norm2) ) continue;

            get_fourier_density_comps(Class_id, c_k_x, c_k_y, c_k_z, curr_comp);

            /* evaluate and add the penalty function */
            gsl_vector_set(n, 0, c_k_x);
            gsl_vector_set(n, 1, c_k_y);
            gsl_vector_set(n, 2, c_k_z);
            gsl_vector_scale(n, 1/sqrt(c_k_norm2));

            sin2_dir_k =
              gsl_pow_2 ( gsl_vector_get(director, 1)*gsl_vector_get(n, 2)
                  - gsl_vector_get(director, 2)*gsl_vector_get(n, 1) ) + 
                  gsl_pow_2 ( gsl_vector_get(director, 2)*gsl_vector_get(n, 0)
                      - gsl_vector_get(director, 0)*gsl_vector_get(n, 2) ) + 
                      gsl_pow_2 ( gsl_vector_get(director, 0)*gsl_vector_get(n, 1)
                          - gsl_vector_get(director, 1)*gsl_vector_get(n, 0) );

            curr_power2 = gsl_complex_abs2(curr_comp)*(1 - penalty_coeff*sin2_dir_k);

            if (curr_power2>max_power2)
              {
                max_comp   = curr_comp;
                max_power2 = curr_power2;
                k_x        = c_k_x;
                k_y        = c_k_y;
                k_z        = c_k_z;
              }

          }

    phi = gsl_complex_arg(max_comp);
    phi = phi -
    2*M_PI * ceil( ( phi - (x_center*k_x + y_center*k_y + z_center*k_z) ) / (2*M_PI) );

    gsl_vector_free(n);

  }

/***************************************************************************/
void emd::uni::discrete_fourier_transform(gsl_vector * input,const int k,gsl_complex & comp){
  int vector_size = input->size;
  if(k>=vector_size | k<0){
    cout<<"DFT input is out of range ..."<<endl;
    exit(0);
  }

  double real = 0;
  double imag = 0;
  double phi;

  for (unsigned int n=0; n < vector_size; n++){
    phi = -2*M_PI*k*n/vector_size;

    real += gsl_vector_get(input,n)*cos(phi);
    imag += gsl_vector_get(input,n)*sin(phi);
  }

  GSL_SET_COMPLEX(&comp, real, imag);

}
/***************************************************************************/
double emd::uni::eval_structure_factor(const int Class_id_1,const int Class_id_2,
    const double & k_x, const double & k_y, const double & k_z){

      double phi;
      double sum=0;
      gsl_vector * r12 = gsl_vector_alloc(3);

      for(unsigned int m1=0;m1<uni::rigid_bodies_count; m1++){
        for(unsigned int m2=0;m2<uni::rigid_bodies_count; m2++){
          if(emd::uni::rigid_bodies[m1].class_id!=Class_id_1 
              || emd::uni::rigid_bodies[m2].class_id!=Class_id_2) continue;

          gsl_vector_memcpy(r12,rigid_bodies[m1].x);
          gsl_vector_sub(r12, rigid_bodies[m2].x);
          phi =
            k_x*gsl_vector_get(r12, 0) + 
            k_y*gsl_vector_get(r12, 1) +
            k_z*gsl_vector_get(r12, 2);

          sum += cos(phi);   
        }
      }
      return sum/rigid_bodies_count;
}
/***************************************************************************/
void
emd::uni::get_fourier_density_comps(const int Class_id,const double & k_x, const double & k_y,
    const double & k_z, gsl_complex & comp)
  {

    double real = 0;
    double imag = 0;
    double phi;

    for (unsigned int m=0; m<uni::rigid_bodies_count; m++)
      {
        //changed

        if(emd::uni::rigid_bodies[m].class_id != Class_id) continue;

        phi =
          k_x*gsl_vector_get(uni::rigid_bodies[m].x, 0) + 
          k_y*gsl_vector_get(uni::rigid_bodies[m].x, 1) +
          k_z*gsl_vector_get(uni::rigid_bodies[m].x, 2);

        real += cos(phi);
        imag += sin(phi);

      }

    GSL_SET_COMPLEX(&comp, real, imag);

  }

/***************************************************************************/

int
emd::uni::eval_in_plane_radial_dist( const int Class_id,const int & order, const double & max_rad, const int & divs,
    const double & k_x, const double & k_y, const double & k_z, const double & phi,
    gsl_vector * in_plane_dist)
  {

    if (eval_volume()==EXIT_FAILURE)
      return EXIT_FAILURE;

    gsl_vector * r12 = gsl_vector_alloc(3);
    gsl_vector * k   = gsl_vector_alloc(3);
    gsl_vector * n   = gsl_vector_alloc(3);

    /* set wave vector */
    gsl_vector_set(k, 0, k_x);
    gsl_vector_set(k, 1, k_y);
    gsl_vector_set(k, 2, k_z);

    /* plane unit normal vector */
    gsl_vector_memcpy(n, k);
    gsl_vector_scale(n, 1.0/gsl_blas_dnrm2(n));

    /* wave length and max image */
    double wl     = 2*M_PI / gsl_blas_dnrm2(k);
    int max_img_x = (int)ceil(sqrt( gsl_pow_2(order*wl) + gsl_pow_2(max_rad) ) / x_width);
    int max_img_y = (int)ceil(sqrt( gsl_pow_2(order*wl) + gsl_pow_2(max_rad) ) / y_width);
    int max_img_z = (int)ceil(sqrt( gsl_pow_2(order*wl) + gsl_pow_2(max_rad) ) / z_width);
    double phase;

    /* zero out the dist vector */
    gsl_vector_set_zero(in_plane_dist);

    for (int  m1=0 ; m1 < rigid_bodies_count ; m1++)
      {
        //changed

        if(emd::uni::rigid_bodies[m1].class_id != Class_id) continue;

        /* find the plane index of m1 */
        gsl_blas_ddot(k, rigid_bodies[m1].x, &phase);
        phase -= phi;
        int pl_idx_1 = (int)floor(0.5 + phase/(2*M_PI));

        for (int  m2=0; m2 < rigid_bodies_count ; m2++)
          {
            for (int nx=-max_img_x; nx <= max_img_x ; nx++ )
              {
                for (int ny=-max_img_y; ny <= max_img_y ; ny++ )
                  {
                    for ( int nz=-max_img_z; nz <= max_img_z ; nz++ )
                      {

                        if ( (m1==m2) && (nx==0) && (ny==0) && (nz==0) ) continue;
                        //changed

                        if(emd::uni::rigid_bodies[m2].class_id != Class_id) continue;

                        /* location of the 2nd body */
                        gsl_vector_set(r12, 0, gsl_vector_get(rigid_bodies[m2].x, 0) + nx*x_width);
                        gsl_vector_set(r12, 1, gsl_vector_get(rigid_bodies[m2].x, 1) + ny*y_width);
                        gsl_vector_set(r12, 2, gsl_vector_get(rigid_bodies[m2].x, 2) + nz*z_width);

                        /* find the plane index of m2 */
                        gsl_blas_ddot(k, r12, &phase);
                        phase -= phi;
                        int pl_idx_2 = (int)floor(0.5 + phase/(2*M_PI));

                        /* separation vector */
                        gsl_vector_sub(r12, rigid_bodies[m1].x);

                        if ( abs(pl_idx_2 - pl_idx_1) != order ) continue;

                        double s_scalar_dist = gsl_blas_dnrm2(r12);
                        double s_para_dist;
                        gsl_blas_ddot(r12, n, &s_para_dist);
                        double s_perp_dist = sqrt( gsl_pow_2(s_scalar_dist) - gsl_pow_2(s_para_dist) );

                        int idx_perp   = (int)floor( (s_perp_dist/max_rad) * divs);

                        if (idx_perp < divs)
                          {
                            gsl_vector_set( in_plane_dist, idx_perp,
                                gsl_vector_get( in_plane_dist, idx_perp) + 1);
                          }

                      };
                  };
              }; /* end of images loop */

          };
      }; /* end of rigid_bodies loop */

    double dr = max_rad / divs;

    /* normalization of the RDF to the ideal gas */
    for(int idx = 0; idx < divs; idx++) 
      {

        double r_lower = idx * dr;
        double r_upper = r_lower + dr;

        /* normal */
        double perp_rdf_ideal =
          M_PI * ( gsl_pow_2(r_upper) - gsl_pow_2(r_lower) ) * wl *
          rigid_bodies_count * rigid_bodies_count / volume;

        if (order!=0) perp_rdf_ideal *= 2;

        gsl_vector_set( in_plane_dist, idx, gsl_vector_get( in_plane_dist, idx ) / perp_rdf_ideal);

      }

    gsl_vector_free(r12);
    gsl_vector_free(k);
    gsl_vector_free(n);

    return EXIT_SUCCESS;

  }

/***************************************************************************/

double
emd::uni::eval_bond_order ( const double & k_x, const double & k_y, const double & k_z,
    const double & r_cut_min, const double & r_cut_max)

  {

    gsl_vector * tmp_r12    =  gsl_vector_alloc(3);
    gsl_vector * r12_para   =  gsl_vector_alloc(3);
    gsl_vector * r12_perp   =  gsl_vector_alloc(3);
    gsl_vector * d_ref_1    =  gsl_vector_alloc(3);
    gsl_vector * d_ref_2    =  gsl_vector_alloc(3);
    gsl_vector * k_normal   =  gsl_vector_alloc(3);
    gsl_vector * img_vec    =  gsl_vector_alloc(3);

    double r12_norm, r12_perp_norm, r12_perp_1, r12_perp_2, r12_dot_k, k_dot_d,
    r12_dot_d, ang, curr_w, sum_w, C6;

    gsl_complex curr_C6, tmp_C6;

    /* check to see if the neighbor list is initialized */
    bool flag_not_init   = !nl_init_flag;
    bool flag_bad_cutoff = ((re2::r_c + re2::r_s)<r_cut_max);

    if ( flag_not_init || flag_bad_cutoff)
      {

        /* set the correct cutoff distance */
        double r_c_temp = r_cut_max - re2::r_s; 
        if (r_c_temp<0) r_c_temp = 0;
        re2::r_c = r_c_temp;

        /* (re)init the neighbor list */
        if (flag_bad_cutoff)
          re2::free_memory(); /* clear the previous nl memory */

        use_cell_info = false;
        init_neighbor_list(NL_VERLET);
        re2::update_neighbor_list();

      }

    gsl_vector_set( k_normal, 0, k_x );
    gsl_vector_set( k_normal, 1, k_y );
    gsl_vector_set( k_normal, 2, k_z );

    double norm_k = gsl_blas_dnrm2 ( k_normal );
    gsl_vector_scale ( k_normal, 1.0/norm_k );

    /* find 2 fixed perp direction on the plane */
    if (abs(k_x)>abs(k_y))
      {
        gsl_vector_set(d_ref_1, 0, k_z);
        gsl_vector_set(d_ref_1, 1, 0);
        gsl_vector_set(d_ref_1, 2, -k_x);
      }
    else
      {
        gsl_vector_set(d_ref_1, 0, 0);
        gsl_vector_set(d_ref_1, 1, -k_z);
        gsl_vector_set(d_ref_1, 2, k_y);
      };

    gsl_vector_normalize(d_ref_1);

    /* d_ref_2 = k_hat x d_ref_1 */
    gsl_vector_set(d_ref_2, 0, gsl_vector_get(k_normal, 1) - gsl_vector_get(d_ref_1, 2));
    gsl_vector_set(d_ref_2, 1, gsl_vector_get(k_normal, 2) - gsl_vector_get(d_ref_1, 0));
    gsl_vector_set(d_ref_2, 2, gsl_vector_get(k_normal, 0) - gsl_vector_get(d_ref_1, 1));

    C6       = 0;

    for (unsigned int m1=0 ; m1 < rigid_bodies_count ; m1++)
      {

        GSL_SET_COMPLEX(&curr_C6, 0, 0);
        sum_w    = 0;

        for (re2::neighbor_it it = re2::n_list_re2[m1].begin();
        it!=re2::n_list_re2[m1].end(); it++)

          {

            unsigned int m2 = (*it).idx;

            /* locating the respective image of the neighboring BODY */
            gsl_vector_set(img_vec, 0, (*it).img_x*x_width);
            gsl_vector_set(img_vec, 1, (*it).img_y*y_width);
            gsl_vector_set(img_vec, 2, (*it).img_z*z_width);

            gsl_vector_memcpy (tmp_r12, rigid_bodies[m2].x);
            gsl_vector_sub    (tmp_r12, rigid_bodies[m1].x);
            gsl_vector_add    (tmp_r12, img_vec);          /* adding image guiding vector */

            r12_norm = gsl_blas_dnrm2 (tmp_r12);

            if (r12_norm>r_cut_max) continue;

            gsl_blas_ddot (k_normal, tmp_r12 , &r12_dot_k);
            gsl_vector_memcpy (r12_para ,k_normal);
            gsl_vector_scale (r12_para, r12_dot_k );   /* component of r12 parallel to k */ 
            gsl_vector_memcpy (r12_perp ,tmp_r12 ); 
            gsl_vector_sub (r12_perp ,r12_para );     /* component of r12 perpendicular to k */
            r12_perp_norm = gsl_blas_dnrm2 ( r12_perp ); 

            gsl_blas_ddot (r12_perp, d_ref_1, &r12_perp_1);
            gsl_blas_ddot (r12_perp, d_ref_2, &r12_perp_2);

            /* find the angle using d_ref_1 as the branch cut */
            ang = atan2(abs(r12_perp_2), abs(r12_perp_1));
            if ( (r12_perp_1<0) && (r12_perp_2>0) )
              ang = M_PI - ang;
            else if ( (r12_perp_1<0) && (r12_perp_2<0) )
              ang = M_PI + ang;
            else if ( (r12_perp_1>0) && (r12_perp_2<0) )
              ang = 2*M_PI - ang;

            /* find the weight */
            curr_w  = get_neighbor_weight_fun(r12_norm, r_cut_min, r_cut_max);

            tmp_C6  = gsl_complex_mul_real(gsl_complex_polar(1.0, 6.0*ang), curr_w); 
            curr_C6 = gsl_complex_add(curr_C6, tmp_C6);
            sum_w  += curr_w;

          }

        curr_C6 = gsl_complex_div_real(curr_C6, sum_w);
        C6      += gsl_complex_abs(curr_C6);

      }

    C6 = C6/rigid_bodies_count;

    gsl_vector_free(tmp_r12);
    gsl_vector_free(r12_para);
    gsl_vector_free(r12_perp);
    gsl_vector_free(d_ref_1);
    gsl_vector_free(d_ref_2);
    gsl_vector_free(k_normal);
    gsl_vector_free(img_vec);

    return C6;

  }

/***************************************************************************/
double
emd::uni::get_neighbor_weight_fun( const double & r, const double & r_cut_min,
    const double & r_cut_max )
  {

    if (r>r_cut_max)
      return 0;
    else if (r<r_cut_min)
      return 1;
    else
      return (r_cut_max - r)/(r_cut_max - r_cut_min);

  }

/***************************************************************************/

double
emd::uni::get_inc_sphere_volume( const double & r, const double & h )
  {
    return M_PI * ( h*(r*r-h*h) + 2*r*r*r*(1+h/r) ) / 3;
  }

/***************************************************************************/

double
emd::uni::get_inc_cylinder_volume( const double & r, const double & h )
  {
    return 4 * M_PI * ( r*r*r - gsl_pow_3(sqrt(r*r-h*h)) ) / 3;
  }
/***************************************************************************/


double emd::uni::eval_number_of_regions(const unsigned int Class_id,
    const double max_neighbor_distance_x, const double max_neighbor_distance_y,
    const double max_neighbor_distance_z)
  {

    unsigned int idx_right;
    unsigned int idx_front;
    unsigned int idx_up;

    unsigned int idx_x;
    unsigned int idx_y;
    unsigned int idx_z;
    unsigned int cell_idx;

    int * particle_list;

    unsigned int n_x = (int)ceil(x_width/max_neighbor_distance_x);
    double cell_size_x = x_width/n_x;

    unsigned int n_y = (int)ceil(y_width/max_neighbor_distance_y);
    double cell_size_y = y_width/n_y;

    unsigned int n_z = (int)ceil(z_width/max_neighbor_distance_z);
    double cell_size_z = z_width/n_z;

    unsigned int total_cell_number = n_x * n_y * n_z;
    particle_list = new int[total_cell_number];

    for (unsigned int i=0; i<total_cell_number; i++)
      particle_list[i] = 0;

    //Initialization of cells by assigning 1 to each cell containing particles...

    for (unsigned int m=0; m<uni::rigid_bodies_count; m++)
      {

        if (emd::uni::rigid_bodies[m].class_id != Class_id) continue;

        idx_x = int((-x_lower_bound + gsl_vector_get(emd::uni::rigid_bodies[m].x,0))/cell_size_x);
        idx_y = int((-y_lower_bound + gsl_vector_get(emd::uni::rigid_bodies[m].x,1))/cell_size_y);
        idx_z = int((-z_lower_bound + gsl_vector_get(emd::uni::rigid_bodies[m].x,2))/cell_size_z);

        //        cout<<endl<<"hello..."<<endl;exit(0);

        cell_idx = idx_x + idx_y * n_x + idx_z * n_x*n_y;

        particle_list[cell_idx]++;


        if(cell_idx>=total_cell_number){
          cout<<"Error in emd::uni::eval_number_of_regions() ...."<<endl;
          exit(0);
        }

      }
    Graph G(total_cell_number);

    std::vector<size_type> rank(num_vertices(G));
    std::vector<Vertex> parent(num_vertices(G));

    disjoint_sets<Rank, Parent> ds(&rank[0], &parent[0]);

    initialize_incremental_components(G, ds);
    incremental_components(G, ds);

    graph_traits<Graph>::edge_descriptor e;

    bool flag;
    cell_idx=0;

    for (unsigned int idx_z = 0; idx_z < n_z; idx_z++)
      {
        for (unsigned int idx_y = 0; idx_y < n_x; idx_y++)
          {
            for (unsigned int idx_x = 0; idx_x < n_x; idx_x++)
              {
                if(particle_list[cell_idx]){

                  idx_right = (idx_x == n_x - 1) ? 0 : cell_idx + 1;
                  idx_front = (idx_y == n_y - 1) ? 0 : cell_idx + n_x;
                  idx_up    = (idx_z == n_z - 1) ? 0 : cell_idx + n_x*n_y;

                  if(particle_list[idx_right])
                    {
                      boost::tie(e, flag) = add_edge(cell_idx, idx_right, G);
                      ds.union_set(cell_idx, idx_right);
                    }

                  if(particle_list[idx_front])
                    {
                      boost::tie(e, flag) = add_edge(cell_idx, idx_front, G);
                      ds.union_set(cell_idx, idx_front);
                    }

                  if(particle_list[idx_up])
                    {
                      boost::tie(e, flag) = add_edge(cell_idx, idx_up, G);
                      ds.union_set(cell_idx, idx_up);
                    }

                }
                cell_idx++;
              }
          }
      }


    Components components(&parent[0], &parent[0] + parent.size());
    unsigned int number_of_regions = 0;
    for (Components::size_type c = 0; c < components.size(); ++c) {
      Components::value_type::iterator
      j = components[c].begin(),
      jend = components[c].end();
      int N=0;

      int num_of_particles = 0;
      for ( ; j != jend; ++j)
        {
          num_of_particles += particle_list[*j];
          N++;
        }


      if(N>1)
        {
          number_of_regions++;
          cout<<"Componenet[ "<<setw(4)<<c<<" ] :    "
          <<"     Number of cells = "<<setw(4)<<N
          <<"     Number of particles = "<<setw(4)<<num_of_particles<<endl;
        }

    }
    delete [] particle_list;

    cout<<endl<<"n_x = "<<n_x<<endl
    <<"n_y = "<<n_y<<endl
    <<"n_z = "<<n_z<<endl<<endl;
    return number_of_regions;
  }
/***************************************************************************/

void
emd::uni::total_density_fourier( const int Class_id,  gsl_vector * dir,
    const unsigned int divs,double min_wl,double &dk,
    gsl_vector * para_freq,gsl_vector * perp_freq,gsl_vector * scalar_freq)
  {
    double max_wl = emd::uni::x_width;
    if (emd::uni::y_width>max_wl) max_wl = emd::uni::y_width;
    if (emd::uni::z_width>max_wl) max_wl = emd::uni::z_width;

    double min_k_norm = 2*M_PI/max_wl;
    double min_k_norm2 = min_k_norm * min_k_norm;
    double max_k_norm = 2*M_PI/min_wl;
    double max_k_norm2 = max_k_norm *max_k_norm ;

    gsl_vector * n   = gsl_vector_alloc(3);
    gsl_vector_memcpy(n, dir);
    gsl_vector_scale(n, 1.0/gsl_blas_dnrm2(dir));
    gsl_complex comp;


    gsl_vector * scalar_count = gsl_vector_calloc(divs);
    gsl_vector * para_count = gsl_vector_calloc(divs);
    gsl_vector * perp_count = gsl_vector_calloc(divs);

    dk =  max_k_norm  / divs;
    gsl_vector_set_zero(para_freq);
    gsl_vector_set_zero(perp_freq);
    gsl_vector_set_zero(scalar_freq);

    for (double k_x=-max_k_norm; k_x<=max_k_norm; k_x += dk )
      {
        for (double k_y=-max_k_norm; k_y<=max_k_norm; k_y += dk )
          {
            for (double k_z=-max_k_norm; k_z<=max_k_norm; k_z += dk )
              {
                double k_norm2  = k_x*k_x + k_y*k_y + k_z*k_z;
                if( k_norm2>max_k_norm2 || k_norm2<min_k_norm2 ) continue;
                emd::uni::get_fourier_density_comps(Class_id, k_x , k_y , k_z , comp);

                double comp_abs = gsl_complex_abs(comp);
                double k_norm = sqrt(k_norm2);

                double s_para_freq = fabs(gsl_vector_get(n,0)*k_x + gsl_vector_get(n,0)*k_y + gsl_vector_get(n,0)*k_z);

                double s_perp_freq = sqrt( gsl_pow_2(k_norm) - gsl_pow_2(s_para_freq) );

                int idx_scalar  = int( k_norm/dk );
                int idx_para    = int( s_para_freq/dk );
                int idx_perp    = int( s_perp_freq/dk );
                

                if(idx_scalar<divs)
                  {
                    gsl_vector_set( scalar_count, idx_scalar,
                        gsl_vector_get( scalar_count, idx_scalar) + 1);
                    gsl_vector_set( scalar_freq, idx_scalar,
                        gsl_vector_get( scalar_freq, idx_scalar) + comp_abs);

                  }

                if(idx_para<divs)
                  {
                    gsl_vector_set( para_count, idx_para,
                        gsl_vector_get( para_count, idx_para) + 1);
                    gsl_vector_set( para_freq, idx_para,
                        gsl_vector_get( para_freq, idx_para) + comp_abs);

                  }

                if(idx_perp<divs)
                  {
                    gsl_vector_set( perp_count, idx_perp,
                        gsl_vector_get( perp_count, idx_perp) + 1);
                    gsl_vector_set( perp_freq, idx_perp,
                        gsl_vector_get( perp_freq, idx_perp) + comp_abs);

                  }
              }
          }
      }
    for(unsigned int i=0;i<divs;i++){

      if(gsl_vector_get(perp_freq,i)!=0)
        {
          gsl_vector_set( perp_freq, i,
              gsl_vector_get( perp_freq, i) / gsl_vector_get(perp_count,i));
        }

      if(gsl_vector_get(scalar_freq,i)!=0)
        {
          gsl_vector_set( scalar_freq, i,
              gsl_vector_get( scalar_freq, i) / gsl_vector_get(scalar_count,i)); 
        }

      if(gsl_vector_get(para_freq,i)!=0)
        {
          gsl_vector_set( para_freq, i,
              gsl_vector_get( para_freq, i) / gsl_vector_get(para_count,i));
        }    
    }
    gsl_vector_free(scalar_count);
    gsl_vector_free(para_count);
    gsl_vector_free(perp_count);

  }
/***************************************************************************/
void
emd::uni::total_structure_factor( gsl_vector * dir,
    const unsigned int divs,double min_wl,double &dk,
    gsl_vector * para_Snn , gsl_vector * perp_Snn , gsl_vector * scalar_Snn,
    gsl_vector * para_Scc , gsl_vector * perp_Scc , gsl_vector * scalar_Scc)
  {
    double x1 = 0.5;
    double x2 = 0.5;
    double Snn,Scc;
    
    double max_wl = emd::uni::x_width;
    if (emd::uni::y_width>max_wl) max_wl = emd::uni::y_width;
    if (emd::uni::z_width>max_wl) max_wl = emd::uni::z_width;

    double min_k_norm = 2*M_PI/max_wl;
    double min_k_norm2 = min_k_norm * min_k_norm;
    double max_k_norm = 2*M_PI/min_wl;
    double max_k_norm2 = max_k_norm *max_k_norm ;

    gsl_vector * n   = gsl_vector_alloc(3);
    gsl_vector_memcpy(n, dir);
    gsl_vector_scale(n, 1.0/gsl_blas_dnrm2(dir));

    gsl_vector * scalar_count = gsl_vector_calloc(divs);
    gsl_vector * para_count = gsl_vector_calloc(divs);
    gsl_vector * perp_count = gsl_vector_calloc(divs);

    dk =  max_k_norm  / divs;
    gsl_vector_set_zero(para_Snn);
    gsl_vector_set_zero(perp_Snn);
    gsl_vector_set_zero(scalar_Snn);
    gsl_vector_set_zero(para_Scc);
    gsl_vector_set_zero(perp_Scc);
    gsl_vector_set_zero(scalar_Scc);

    for (double k_x=-max_k_norm; k_x<=max_k_norm; k_x += dk )
      {
        cout<<"Progress ... "<<setw(3)<<right<<int((k_x /max_k_norm + 1)*50) <<" %"<<endl;
        for (double k_y=-max_k_norm; k_y<=max_k_norm; k_y += dk )
          {
            for (double k_z=-max_k_norm; k_z<=max_k_norm; k_z += dk )
              {
                double k_norm2  = k_x*k_x + k_y*k_y + k_z*k_z;
                if( k_norm2>max_k_norm2 || k_norm2<min_k_norm2 ) continue;
                
                double S11 = emd::uni::eval_structure_factor(1,1,k_x,k_y,k_z);
                double S12 = emd::uni::eval_structure_factor(1,2,k_x,k_y,k_z);
                double S22 = emd::uni::eval_structure_factor(2,2,k_x,k_y,k_z);
                if(max_k_norm/dk*k_x)
                
                
//                cout <<showpos<<setprecision(5)
//                <<setw(20)<<scientific<<S11<<setw(20)<<S12<<setw(20)<<S22<<endl;
                Snn = S11 + 2*S12 + S22;
                Scc = x2*x2*S11 + x1*x1*S22 - 2*x1*x2*S12;   
//                cout <<showpos<<setprecision(5)
//                <<setw(20)<<scientific<<Snn<<setw(20)<<Scc<<endl;
                double k_norm = sqrt(k_norm2);

                double s_para_freq = fabs(gsl_vector_get(n,0)*k_x + gsl_vector_get(n,0)*k_y + gsl_vector_get(n,0)*k_z);

                double s_perp_freq = sqrt( gsl_pow_2(k_norm) - gsl_pow_2(s_para_freq) );

                int idx_scalar  = int( k_norm/dk );
                int idx_para    = int( s_para_freq/dk );
                int idx_perp    = int( s_perp_freq/dk );
                

                if(idx_scalar<divs)
                  {
                    gsl_vector_set( scalar_count, idx_scalar,
                        gsl_vector_get( scalar_count, idx_scalar) + 1);
                    gsl_vector_set( scalar_Snn, idx_scalar,
                        gsl_vector_get( scalar_Snn, idx_scalar)  + Snn);
                    gsl_vector_set( scalar_Scc, idx_scalar,
                        gsl_vector_get( scalar_Scc, idx_scalar) + Scc);

                  }
               
                if(idx_para<divs)
                  {
                    gsl_vector_set( para_count, idx_para,
                        gsl_vector_get( para_count, idx_para) + 1);
                    gsl_vector_set( para_Snn, idx_para,
                        gsl_vector_get( para_Snn, idx_para) + Snn);
                    gsl_vector_set( para_Scc, idx_para,
                        gsl_vector_get( para_Scc, idx_para) + Scc);

                  }

                if(idx_perp<divs)
                  {
                    gsl_vector_set( perp_count, idx_perp,
                        gsl_vector_get( perp_count, idx_perp) + 1);
                    gsl_vector_set( perp_Snn, idx_perp,
                        gsl_vector_get( perp_Snn, idx_perp) + Snn);
                    gsl_vector_set( perp_Scc, idx_perp,
                        gsl_vector_get( perp_Scc, idx_perp) + Scc);

                  }
              }
          }
      }
    for(unsigned int i=0;i<divs;i++){
//      cout <<"idx_scalar = "<<showpos<<setprecision(5)
//      <<setw(20)<<scientific<<i
//      <<setw(20)<<"scalar_count = "<<gsl_vector_get( scalar_count, i)
//      <<setw(20)<<"scalar_Snn = "<<gsl_vector_get( scalar_Snn, i)<<endl;

      if(gsl_vector_get(perp_count,i)!=0)
        {
          gsl_vector_set( perp_Snn, i,
              gsl_vector_get( perp_Snn, i) / gsl_vector_get(perp_count,i));
        }

      if(gsl_vector_get(perp_count,i)!=0)
        {
          gsl_vector_set( perp_Scc, i,
              gsl_vector_get( perp_Scc, i) / gsl_vector_get(perp_count,i));
        }

      if(gsl_vector_get(scalar_count,i)!=0)
        {
          gsl_vector_set( scalar_Snn, i,
              gsl_vector_get( scalar_Snn, i) / gsl_vector_get(scalar_count,i)); 
        }

      if(gsl_vector_get(scalar_count,i)!=0)
        {
          gsl_vector_set( scalar_Scc, i,
              gsl_vector_get( scalar_Scc, i) / gsl_vector_get(scalar_count,i)); 
        }

      if(gsl_vector_get(para_count,i)!=0)
        {
          gsl_vector_set( para_Snn, i,
              gsl_vector_get( para_Snn, i) / gsl_vector_get(para_count,i));
        }    

      if(gsl_vector_get(para_count,i)!=0)
        {
          gsl_vector_set( para_Scc, i,
              gsl_vector_get( para_Scc, i) / gsl_vector_get(para_count,i));
        }    

    }
    gsl_vector_free(scalar_count);
    gsl_vector_free(para_count);
    gsl_vector_free(perp_count);

  }
/***************************************************************************/
