/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef MOLUNIVERSE_COMP_H
#define MOLUNIVERSE_COMP_H

#include <iostream>
#include <iomanip>
#include <vector>
#include <cmath>

#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_eigen.h>

#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/graph/graph_utility.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/pending/disjoint_sets.hpp>
#include <boost/graph/incremental_components.hpp>

#include "quaternion/gsl_quaternion.h"

#include "consts.h"
#include "gslext.h"
#include "rigid_body.h"
#include "surface.h"
#include "paramdb.h"
#include "stringext.h"
#include "moluniverse.h"
#include "int_re2_uni.h"
#include "cell_grid.h"

using namespace boost;
namespace emd
  {

    namespace uni
      {
        /**************************************************************************/
        // Definitions for determination of the number of regions

        typedef adjacency_list <vecS, vecS, undirectedS> Graph;
        typedef graph_traits<Graph>::vertex_descriptor Vertex;
        typedef graph_traits<Graph>::vertices_size_type size_type;
        typedef size_type* Rank;
        typedef Vertex* Parent;
        typedef component_index<unsigned int> Components;

        /**************************************************************************/
        int eval_Q_tensor( std::vector<unsigned int> & indices, const int & axis,
            gsl_matrix * Q_tensor);

        int eval_Q_tensor( std::vector<unsigned int> & indices, const int & axis,
            gsl_matrix * Q_tensor, gsl_vector * Q_eigen_val, gsl_matrix * Q_eigen_vec);

        int eval_radial_dist( const int Class_id, const double & max_rad, const int & divs, gsl_vector * dir,
            gsl_vector * para_dist, gsl_vector * perp_dist,
            gsl_vector * scalar_dist);

        void total_density_fourier( const int Class_id, gsl_vector * dir,
            const unsigned int freq_divs,double min_wl,double &dk, gsl_vector * para_freq,
            gsl_vector * perp_freq, gsl_vector * scalar_freq);

        void total_structure_factor( gsl_vector * dir,
            const unsigned int divs,double min_wl,double &dk,
            gsl_vector * para_Snn , gsl_vector * perp_Snn , gsl_vector * scalar_Snn,
            gsl_vector * para_Scc , gsl_vector * perp_Scc , gsl_vector * scalar_Scc);

        /*******************************
         * detect planes
         *
         * @min_wl         minimum wavelength
         * @k_x            wavevector (x)
         * @k_y            wavevector (y)
         * @k_z            wavevector (z)
         * @phi            phase shift
         *
         */

        int detect_planes(const int Class_id, const double & min_wl, double & k_x, double & k_y, double & k_z,
            double & phi, gsl_vector * director, const double & penalty_coeff);

        int eval_in_plane_radial_dist( const int Class_id, const int & order, const double & max_rad, const int & divs,
            const double & k_x, const double & k_y, const double & k_z, const double & phi,
            gsl_vector * in_plane_dist);

        void get_fourier_density_comps(const int Class_id, const double & k_x, const double & k_y,
            const double & k_z, gsl_complex & comp);

        double eval_structure_factor(const int Class_id_1,const int Class_id_2,const double & k_x, const double & k_y, const double & k_z);
        void discrete_fourier_transform(gsl_vector * input,const int k,gsl_complex & comp);

        double eval_bond_order (const double & k_x, const double & k_y, const double & k_z,
            const double & r_cut_min, const double & r_cut_max);

        double get_neighbor_weight_fun(const double & r, const double & r_cut_min,
            const double & r_cut_max);

        double get_inc_sphere_volume( const double & rad, const double & height );

        double get_inc_cylinder_volume( const double & rad, const double & height );

        double eval_number_of_regions(const unsigned int Class_id,
            const double max_neighbor_distance_x,const double max_neighbor_distance_y,
            const double max_neighbor_distance_z);


      };

  };

#endif /* MOLUNIVERSE_COMP_H */
