/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef PARAMDB_H
#define PARAMDB_H

#include <cstdlib>

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <vector>

#include <gsl/gsl_math.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>

#include "consts.h"
#include "uniparser.h"
#include "stringext.h"

/***************************************************************************/

/* debug */
//#define PARAMDB_DEBUG

/* debug param caches */
//#define PARAMCACHE_RIGIDBODY_DEBUG
//#define PARAMCACHE_RE2_INT_DEBUG
//#define PARAMCACHE_GB_INT_DEBUG
//#define PARAMCACHE_EES_INT_DEBUG

/* conversion factors */

/* contiguous memory reservation */
#define EMD_PARAM_ATOMCLASS_RESERVE           50
#define EMD_PARAM_RE2_HOMO_INT_RESERVE        50
#define EMD_PARAM_RE2_HETERO_INT_RESERVE      50
#define EMD_PARAM_GB_HOMO_INT_RESERVE         50
#define EMD_PARAM_GB_HETERO_INT_RESERVE       50

//#define EMD_PARAMDB_SHOW_WARNINGS

/***************************************************************************/

namespace emd
{

namespace paramdb
{

  struct param_rigid_body
  {
    int class_id;                             /* class ID */
    std::string name;                         /* name */
    double mass;                              /* mass */
    double I_body_x, I_body_y, I_body_z;      /* diag. inertial tensor */
  };

  /***************************************************************************/

  struct param_int_surf
  {
    int class_id;                             /* class ID */
  };

  /***************************************************************************/

  struct param_RE2_homo_int
  {
    int class_id;
    int h_type;
    double hamaker, sigma_c;
    double sig1, sig2, sig3;
    double e1, e2, e3;
  };

  /***************************************************************************/

  struct param_RE2_hetero_int
  {
    int class_id_1, class_id_2;
    int h_type;
    double hamaker, sigma_c;
    bool spec; /* specific parameters */
    double sig11, sig12, sig13;
    double e11, e12, e13;
    double sig21, sig22, sig23;
    double e21, e22, e23;
  };

  /***************************************************************************/

  struct param_GB_homo_int
  {
    int class_id;
    int mu, nu;
    double eps, sig;
    double sig1, sig2, sig3;
    double e1, e2, e3;
  };

  /***************************************************************************/

  struct param_GB_hetero_int
  {
    int class_id_1, class_id_2;
    int mu, nu;
    double eps, sig;
    bool spec; /* specific parameters */
    double sig11, sig12, sig13;
    double e11, e12, e13;
    double sig21, sig22, sig23;
    double e21, e22, e23;
  };

  /***************************************************************************/

  struct param_EES_int
  {
    int class_id_surf, class_id_rigid_body;
    double A, sigma_c;
    double sig1, sig2, sig3;
    double e11, e22, e33;
    double e12, e23, e31;
  };

  /***************************************************************************/

  class param_cache_rigid_body
  {
  public:

    int           class_id;     /* class id */
    std::string   name;         /* name */
    double        mass;         /* mass */
    double        I1, I2, I3;   /* diag. inertia elements */
    gsl_matrix *  I_body;       /* diag. inertia tensor */
    gsl_matrix *  I_body_inv;   /* inv. diag. inertia tensor */

    param_cache_rigid_body ();
    ~param_cache_rigid_body ();

  };

  /***************************************************************************/

  class param_cache_int_surf
  {
  public:

    int           class_id;     /* class id */

  };

  /***************************************************************************/

  class param_cache_RE2_int
  {
  public:

    int              class_id_1;          /* class id 1 */
    int              class_id_2;          /* class id 2 */
    int              h_type;              /* least distance formula to use */

    double           hamaker;             /* hamaker constant */
    double           sigma_c;             /* atomic sigma */
    double           dS1;                 /* det[S_1] */
    double           dS2;                 /* det[S_2] */
    double           max_thickness;

    gsl_matrix *     S1;                  /* S_1 */
    gsl_matrix *     S1p2;                /* S_1^2 */
    gsl_matrix *     S1pm2;               /* S_1^-2 */
    gsl_matrix *     E1;                  /* E_1 */
    gsl_matrix *     E1p2;                /* E_1^2 */

    gsl_matrix *     S2;                  /* S_2 */
    gsl_matrix *     S2p2;                /* S_2^2 */
    gsl_matrix *     S2pm2;               /* S_2^-2 */
    gsl_matrix *     E2;                  /* E_2 */
    gsl_matrix *     E2p2;                /* E_2^2 */

    param_cache_RE2_int ();
    ~param_cache_RE2_int ();

  };

  /***************************************************************************/

  class param_cache_GB_int
  {
  public:

    int              class_id_1;          /* class id 1 */
    int              class_id_2;          /* class id 2 */

    double           eps;                 /* energy constant */
    double           sig;                 /* length constant */
    double           dS1;                 /* ~ det[S_1] */
    double           dS2;                 /* ~ det[S_2] */
    double           max_thickness;

    int              mu, nu;

    gsl_matrix *     S1;                  /* S_1 */
    gsl_matrix *     S1p2;                /* S_1^2 */
    gsl_matrix *     E1;                  /* E_1 */
    gsl_matrix *     E1p2;                /* E_1^2 */

    gsl_matrix *     S2;                  /* S_2 */
    gsl_matrix *     S2p2;                /* S_2^2 */
    gsl_matrix *     E2;                  /* E_2 */
    gsl_matrix *     E2p2;                /* E_2^2 */

    param_cache_GB_int ();
    ~param_cache_GB_int ();

  };

  /***************************************************************************/

  class param_cache_EES_int
  {
  public:

    int           class_id_surf;
    int           class_id_rigid_body;

    gsl_matrix *  S;
    gsl_matrix *  E;

    double        A_fact;            /* A*sigma_c^6 */
    double        R_fact;            /* A*sigma_c^12 */
    double        trE;               /* trace(E) */
    double        max_thickness;

    param_cache_EES_int ();
    ~param_cache_EES_int ();

  };

  /***************************************************************************/

  typedef int (*act) (const std::string & param_buff);

  static const int           N_KEYWORDS     =  7;            /* number of keywords to look for */
  static const int           H_FORMULA_GB   =  0;
  static const int           H_FORMULA_SPH  =  1;

  static std::string         KEYWORDS[N_KEYWORDS];           /* keywords */
  static act                 ACTIONS[N_KEYWORDS];            /* actions */

  static bool init = false;

  extern std::string        param_file_name;

  /* temporary parameter vectors */
  static std::vector<param_rigid_body>             vec_param_rigid_body;
  static std::vector<param_int_surf>               vec_param_int_surf;
  static std::vector<param_RE2_homo_int>           vec_param_RE2_homo_int;
  static std::vector<param_RE2_hetero_int>         vec_param_RE2_hetero_int;
  static std::vector<param_GB_homo_int>            vec_param_GB_homo_int;
  static std::vector<param_GB_hetero_int>          vec_param_GB_hetero_int;
  static std::vector<param_EES_int>                vec_param_EES_int;

  /* subroutines to parse each of the parameters */
  int parse_param_rigid_body (const std::string & param_string);
  int parse_param_int_surf (const std::string & param_string);
  int parse_param_RE2_homo_int (const std::string & param_string);
  int parse_param_RE2_hetero_int (const std::string & param_string);
  int parse_param_GB_homo_int (const std::string & param_string);
  int parse_param_GB_hetero_int (const std::string & param_string);
  int parse_param_EES_int (const std::string & param_string);

  /* parameter loading post-processing */
  int param_post_proc ();
  int param_RE2_hetero_int_post_proc ();
  int param_GB_hetero_int_post_proc ();

  int init_param_cache_rigid_body ();
  int init_param_cache_int_surf ();
  int init_param_cache_RE2_int ();
  int init_param_cache_GB_int ();
  int init_param_cache_EES_int ();

  /***************************************************************************/

  /* parameter caches */
  static int                                       rigid_body_param_max_idx;
  static int                                       int_surf_param_max_idx;

  extern param_cache_rigid_body *                  rigid_body_param;
  extern param_cache_int_surf *                    int_surf_param;
  extern param_cache_RE2_int **                    re2_int_param;
  extern param_cache_GB_int **                     gb_int_param;
  extern param_cache_EES_int **                    ees_int_param;

  /* opens and parses the provided param file */
  int load_param_file (const std::string & param_file_name, const bool & clear = true);

  bool is_init();

  /* class id to array index */
  int get_rigid_body_index (int class_id);
  int get_int_surf_index (int class_id);

  /* free memory */
  int free_memory ();

};

};

#endif /* PARAMDB_H */
