/* gsl_quaternion.h
 * 
 * Copyright (C) 2005 Mehrtash Babadi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "gsl_quaternion.h"
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_check_range.h>

#ifndef __GSL_QUATERNION_MATH_H__
#define __GSL_QUATERNION_MATH_H__

#undef __BEGIN_DECLS
#undef __END_DECLS
#ifdef __cplusplus
# define __BEGIN_DECLS extern "C" {
# define __END_DECLS }
#else
# define __BEGIN_DECLS /* empty */
# define __END_DECLS /* empty */
#endif

__BEGIN_DECLS

/* Quaternions */

gsl_quaternion gsl_quaternion_rect (double x, double y1, double y2, double y3);  /* r= real+i*imag1 + j*imag2 + k*imag3 */

#ifdef HAVE_INLINE
extern inline gsl_quaternion
gsl_quaternion_rect (double x, double y1, double y2, double y3)
{                               /* return z = x + i y1 + j y2 + k y3*/
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, x, y1, y2, y3);
  return q;
}
#endif

#define GSL_QUATERNION_ONE (gsl_quaternion_rect(1.0,0.0,0.0,0.0))
#define GSL_QUATERNION_ZERO (gsl_quaternion_rect(0.0,0.0,0.0,0.0))
#define GSL_QUATERNION_NEGONE (gsl_quaternion_rect(-1.0,0.0,0.0,0.0))

/* Properties of quaternions */

double gsl_quaternion_abs (gsl_quaternion q);   /* return |q|   */
double gsl_quaternion_abs2 (gsl_quaternion q);  /* return |q|^2 */

/* Quaternion arithmetic operators */

gsl_quaternion gsl_quaternion_add (gsl_quaternion a, gsl_quaternion b);  /* r=a+b */
gsl_quaternion gsl_quaternion_sub (gsl_quaternion a, gsl_quaternion b);  /* r=a-b */
gsl_quaternion gsl_quaternion_mul (gsl_quaternion a, gsl_quaternion b);  /* r=a*b */
gsl_quaternion gsl_quaternion_div (gsl_quaternion a, gsl_quaternion b);  /* r=a/b */

gsl_quaternion gsl_quaternion_add_real (gsl_quaternion a, double x);  /* r=a+x */
gsl_quaternion gsl_quaternion_sub_real (gsl_quaternion a, double x);  /* r=a-x */
gsl_quaternion gsl_quaternion_mul_real (gsl_quaternion a, double x);  /* r=a*x */
gsl_quaternion gsl_quaternion_div_real (gsl_quaternion a, double x);  /* r=a/x */

gsl_quaternion gsl_quaternion_add_imag (gsl_quaternion a, int idx, double y); /* r=a+(idx_imag)*y */
gsl_quaternion gsl_quaternion_sub_imag (gsl_quaternion a, int idx, double y); /* r=a-(idx_imag)*y */
gsl_quaternion gsl_quaternion_mul_imag (gsl_quaternion a, int idx, double y); /* r=a*(idx_imag)*y */
gsl_quaternion gsl_quaternion_div_imag (gsl_quaternion a, int idx, double y); /* r=a/(idx_imag)*y */

gsl_quaternion gsl_quaternion_conjugate (gsl_quaternion q);  /* r=conj(q) */
gsl_quaternion gsl_quaternion_inverse (gsl_quaternion a);    /* r=1/a */
gsl_quaternion gsl_quaternion_negative (gsl_quaternion a);    /* r=-a */

gsl_quaternion gsl_quaternion_normal (gsl_quaternion a);    /* r=a/|a| */
void gsl_quaternion_normalize (gsl_quaternion * a);    /* a<-a/|a| */

/* quaternion useful transformations */
gsl_quaternion gsl_quaternion_d0 (gsl_quaternion a);   /* identity */
gsl_quaternion gsl_quaternion_d1 (gsl_quaternion a);   /* (-q1, q0, q3, -q2) */
gsl_quaternion gsl_quaternion_d2 (gsl_quaternion a);   /* (-q2, -q3, q0, q1) */
gsl_quaternion gsl_quaternion_d3 (gsl_quaternion a);   /* (-q3, q2, -q1, q0) */

/* dot product */
double gsl_quaternion_dot_prod (gsl_quaternion a, gsl_quaternion b);

/* Unit quaternions <-> SO(3) matrices conversion routines */

/* converts a unit quaternion to a rotation 3-matrix */
void gsl_quaternion_unit_quat_to_mat(const gsl_quaternion * uq, gsl_matrix * mat);

/* converts a rotation 3-matrix to a unit quaternion */
gsl_quaternion gsl_quaternion_mat_to_unit_quat(const gsl_matrix * mat);

/* converts Euler angles to a unit quaternion */
gsl_quaternion gsl_quaternion_euler_to_unit_quat(double theta, double phi, double psi);

/* converts Euler angles to a unit quaternion */
void gsl_quaternion_unit_quat_to_euler(gsl_quaternion q, double * theta, double * phi, double * psi);

__END_DECLS

#endif /* __GSL_QUATERNION_MATH_H__ */
