/* gsl_quaternion_math.h
 * 
 * Copyright (C) 2005, Mehrtash Babadi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/* todo:
 *
 * - inline functions
 */

#include <math.h>
#include <stdio.h>

#include <gsl_quaternion.h>
#include <gsl_quaternion_math.h>
#include <gsl/gsl_math.h>

/* proper definition of gsl_quaternion_rect in case of inavailability of inline functions */

#ifndef HIDE_INLINE_STATIC
gsl_quaternion
gsl_quaternion_rect (double x, double y1, double y2, double y3)
{                               /* return z = x + i y1 + j y2 + k y3*/
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, x, y1, y2, y3);
  return q;
}
#endif

/* Properties of quaternions */

double
gsl_quaternion_abs (gsl_quaternion q)
{                               /* return |q| */
  double x = GSL_QUATERNION_REAL (q);
  double y1 = GSL_QUATERNION_IMAG1 (q);
  double y2 = GSL_QUATERNION_IMAG2 (q);
  double y3 = GSL_QUATERNION_IMAG3 (q);

  return sqrt(x*x + y1*y1 + y2*y2 + y3*y3);
}

double
gsl_quaternion_abs2 (gsl_quaternion q)
{                               /* return |q|^2 */
  double x = GSL_QUATERNION_REAL (q);
  double y1 = GSL_QUATERNION_IMAG1 (q);
  double y2 = GSL_QUATERNION_IMAG2 (q);
  double y3 = GSL_QUATERNION_IMAG3 (q);

  return (x*x + y1*y1 + y2*y2 + y3*y3);
}

/* Quaternion arithmetic operators */

/* r=a+b */
gsl_quaternion
gsl_quaternion_add (gsl_quaternion a, gsl_quaternion b)
{
  a.dat[0] += b.dat[0];
  a.dat[1] += b.dat[1];
  a.dat[2] += b.dat[2];
  a.dat[3] += b.dat[3];

  return a;
}

/* r=a-b */
gsl_quaternion
gsl_quaternion_sub (gsl_quaternion a, gsl_quaternion b)
{
  a.dat[0] -= b.dat[0];
  a.dat[1] -= b.dat[1];
  a.dat[2] -= b.dat[2];
  a.dat[3] -= b.dat[3];

  return a;
}

/* r=a*b */
gsl_quaternion
gsl_quaternion_mul (gsl_quaternion a, gsl_quaternion b)
{
  gsl_quaternion q;

  q.dat[0] = a.dat[0]*b.dat[0] - a.dat[1]*b.dat[1] - a.dat[2]*b.dat[2] - a.dat[3]*b.dat[3];
  q.dat[1] = a.dat[0]*b.dat[1] + b.dat[0]*a.dat[1] + a.dat[2]*b.dat[3] - a.dat[3]*b.dat[2];
  q.dat[2] = a.dat[0]*b.dat[2] + b.dat[0]*a.dat[2] + a.dat[3]*b.dat[1] - a.dat[1]*b.dat[3];
  q.dat[3] = a.dat[0]*b.dat[3] + b.dat[0]*a.dat[3] + a.dat[1]*b.dat[2] - a.dat[2]*b.dat[1];

  return q;
}

/* r=a/b */
gsl_quaternion
gsl_quaternion_div (gsl_quaternion a, gsl_quaternion b)
{
  return gsl_quaternion_mul (a, gsl_quaternion_inverse(b));
}

/* r=a+x */
gsl_quaternion gsl_quaternion_add_real (gsl_quaternion a, double x)
{
  a.dat[0] += x;

  return a;
}

/* r=a-x */
gsl_quaternion gsl_quaternion_sub_real (gsl_quaternion a, double x)
{
  a.dat[0] -= x;

  return a;
}

/* r=a*x */
gsl_quaternion gsl_quaternion_mul_real (gsl_quaternion a, double x)
{
  a.dat[0] *= x;
  a.dat[1] *= x;
  a.dat[2] *= x;
  a.dat[3] *= x;

  return a;
}

/* r=a/x */
gsl_quaternion gsl_quaternion_div_real (gsl_quaternion a, double x)
{
  a.dat[0] /= x;
  a.dat[1] /= x;
  a.dat[2] /= x;
  a.dat[3] /= x;

  return a;
}

/* r=a+(idx_imag)*y */
gsl_quaternion gsl_quaternion_add_imag (gsl_quaternion a, int idx, double y)
{
#ifndef GSL_RANGE_CHECK_OFF
  if (idx<0 || idx>3)
  {
    GSL_ERROR_VAL ( "imaginary index ranges from 0 to 2", GSL_EBADLEN, a);
  }
#endif
  a.dat[idx+1] += y;
  return a;
}

/* r=a-(idx_imag)*y */
gsl_quaternion gsl_quaternion_sub_imag (gsl_quaternion a, int idx, double y)
{
#ifndef GSL_RANGE_CHECK_OFF
  if (idx<0 || idx>3)
  {
    GSL_ERROR_VAL ( "imaginary index ranges from 0 to 2", GSL_EBADLEN, a);
  }
#endif
  a.dat[idx+1] -= y;
  return a;
}

/* r=a*[(idx_imag)y] */
gsl_quaternion gsl_quaternion_mul_imag (gsl_quaternion a, int idx, double y)
{
#ifndef GSL_RANGE_CHECK_OFF
  if (idx<0 || idx>3)
  {
    GSL_ERROR_VAL ( "imaginary index ranges from 0 to 2", GSL_EBADLEN, a);
  }
#endif
  gsl_quaternion rq = GSL_QUATERNION_ZERO;
  rq.dat[idx+1] = y;

  return gsl_quaternion_mul(a, rq);
}

/* r=a/[(idx_imag)y] */
gsl_quaternion gsl_quaternion_div_imag (gsl_quaternion a, int idx, double y)
{
#ifndef GSL_RANGE_CHECK_OFF
  if (idx<0 || idx>3)
  {
    GSL_ERROR_VAL ( "imaginary index ranges from 0 to 2", GSL_EBADLEN, a);
  }
#endif
  gsl_quaternion rq = GSL_QUATERNION_ZERO;
  rq.dat[idx+1] = y;

  return gsl_quaternion_div(a, rq);
}

/* r=conj(q) */
gsl_quaternion gsl_quaternion_conjugate (gsl_quaternion q)
{
  gsl_quaternion qconj;
  qconj.dat[0] = q.dat[0];
  qconj.dat[1] = -q.dat[1];
  qconj.dat[2] = -q.dat[2];
  qconj.dat[3] = -q.dat[3];

  return q;
}

/* r=1/a */
gsl_quaternion gsl_quaternion_inverse (gsl_quaternion a)
{
  gsl_quaternion q;
  double abs2 = gsl_quaternion_abs2(a);

  q.dat[0] = a.dat[0]/abs2;
  q.dat[1] = -a.dat[1]/abs2;
  q.dat[2] = -a.dat[2]/abs2;
  q.dat[3] = -a.dat[3]/abs2;

  return q;
}

/* r=-a */
gsl_quaternion gsl_quaternion_negative (gsl_quaternion a)
{
  gsl_quaternion q;
  q.dat[0] = -a.dat[0];
  q.dat[1] = -a.dat[1];
  q.dat[2] = -a.dat[2];
  q.dat[3] = -a.dat[3];

  return q;
}

/* r=a/|a| */
gsl_quaternion gsl_quaternion_normal (gsl_quaternion a)
{
  gsl_quaternion q;
  double abs = gsl_quaternion_abs(a);

  q.dat[0] = a.dat[0]/abs;
  q.dat[1] = a.dat[1]/abs;
  q.dat[2] = a.dat[2]/abs;
  q.dat[3] = a.dat[3]/abs;

  return q;
}

/* a<-a/|a| */
void gsl_quaternion_normalize (gsl_quaternion * a)
{
  double abs = gsl_quaternion_abs(*a);
  a->dat[0] /= abs;
  a->dat[1] /= abs;
  a->dat[2] /= abs;
  a->dat[3] /= abs;
}

/* identity */
gsl_quaternion gsl_quaternion_d0 (gsl_quaternion a)
{
  return a;
}

/* (-q1, q0, q3, -q2) */
gsl_quaternion gsl_quaternion_d1 (gsl_quaternion a)
{

  gsl_quaternion b;

  b.dat[0] = -a.dat[1];
  b.dat[1] = a.dat[0];
  b.dat[2] = a.dat[3];
  b.dat[3] = -a.dat[2];

  return b;

}

/* (-q2, -q3, q0, q1) */
gsl_quaternion gsl_quaternion_d2 (gsl_quaternion a)
{

  gsl_quaternion b;

  b.dat[0] = -a.dat[2];
  b.dat[1] = -a.dat[3];
  b.dat[2] = a.dat[0];
  b.dat[3] = a.dat[1];

  return b;

}

/* (-q3, q2, -q1, q0) */
gsl_quaternion gsl_quaternion_d3 (gsl_quaternion a)
{

  gsl_quaternion b;

  b.dat[0] = -a.dat[3];
  b.dat[1] = a.dat[2];
  b.dat[2] = -a.dat[1];
  b.dat[3] = a.dat[0];

  return b;

}

/* dot product */
double gsl_quaternion_dot_prod (gsl_quaternion a, gsl_quaternion b)
{

  return (a.dat[0]*b.dat[0] + a.dat[1]*b.dat[1] +
          a.dat[2]*b.dat[2] + a.dat[3]*b.dat[3]);

}

/* Unit quaternions <-> SO(3) matrices conversion routines */

/* converts a unit quaternion to a rotation 3-matrix */
void gsl_quaternion_unit_quat_to_mat(const gsl_quaternion * uq, gsl_matrix * mat)
{
#ifndef GSL_RANGE_CHECK_OFF
  if (mat->size1!=3 || mat->size2!=3)
  {
    GSL_ERROR_VOID ( "the supplied matrix must be 3x3", GSL_EBADLEN );
  }
#endif

  // diagonal elements
  gsl_matrix_set(mat, 0, 0,
    uq->dat[0]*uq->dat[0] + uq->dat[1]*uq->dat[1] - uq->dat[2]*uq->dat[2] - uq->dat[3]*uq->dat[3]);
  gsl_matrix_set(mat, 1, 1,
    uq->dat[0]*uq->dat[0] - uq->dat[1]*uq->dat[1] + uq->dat[2]*uq->dat[2] - uq->dat[3]*uq->dat[3]);
  gsl_matrix_set(mat, 2, 2,
    uq->dat[0]*uq->dat[0] - uq->dat[1]*uq->dat[1] - uq->dat[2]*uq->dat[2] + uq->dat[3]*uq->dat[3]);

  // off-diagonal
  gsl_matrix_set(mat, 0, 1, 2.0*uq->dat[1]*uq->dat[2] - 2.0*uq->dat[0]*uq->dat[3]);
  gsl_matrix_set(mat, 0, 2, 2.0*uq->dat[1]*uq->dat[3] + 2.0*uq->dat[0]*uq->dat[2]);
  gsl_matrix_set(mat, 1, 0, 2.0*uq->dat[1]*uq->dat[2] + 2.0*uq->dat[0]*uq->dat[3]);
  gsl_matrix_set(mat, 1, 2, 2.0*uq->dat[2]*uq->dat[3] - 2.0*uq->dat[0]*uq->dat[1]);
  gsl_matrix_set(mat, 2, 0, 2.0*uq->dat[1]*uq->dat[3] - 2.0*uq->dat[0]*uq->dat[2]);
  gsl_matrix_set(mat, 2, 1, 2.0*uq->dat[2]*uq->dat[3] + 2.0*uq->dat[0]*uq->dat[1]);

}

/* converts a rotation 3-matrix to a unit quaternion */
gsl_quaternion gsl_quaternion_mat_to_unit_quat(const gsl_matrix * mat)
{
#ifndef GSL_RANGE_CHECK_OFF
  if (mat->size1!=3 || mat->size2!=3)
  {
    GSL_ERROR_VAL ( "the supplied matrix must be 3x3", GSL_EBADLEN, GSL_QUATERNION_ZERO);
  }
#endif

  gsl_quaternion q;
  double tr, s;
  int i;

  // trace of the matrix
  tr = gsl_matrix_get(mat,0,0) + gsl_matrix_get(mat,1,1) + gsl_matrix_get(mat,2,2);

  if (tr>=0)
  {

    s = sqrt(1 + tr);

    GSL_QUATERNION_SET_REAL(&q, 0.5*s);

    s = 0.5/s;

    GSL_QUATERNION_SET_IMAG1 (&q, (gsl_matrix_get(mat,2,1) - gsl_matrix_get(mat,1,2))*s);
    GSL_QUATERNION_SET_IMAG2 (&q, (gsl_matrix_get(mat,0,2) - gsl_matrix_get(mat,2,0))*s);
    GSL_QUATERNION_SET_IMAG3 (&q, (gsl_matrix_get(mat,1,0) - gsl_matrix_get(mat,0,1))*s);

  }
  else
  {

    i = 0;
    if (gsl_matrix_get(mat,1,1) > gsl_matrix_get(mat,0,0)) i = 1;
    if (gsl_matrix_get(mat,2,2) > gsl_matrix_get(mat,i,i)) i = 2;

    switch (i) {

      case 0:
      {

        s = sqrt(1 + gsl_matrix_get(mat,0,0) -
          gsl_matrix_get(mat,1,1) - gsl_matrix_get(mat,2,2));

        GSL_QUATERNION_SET_IMAG1 (&q, 0.5*s);

        s = 0.5/s;

        GSL_QUATERNION_SET_IMAG2 (&q, (gsl_matrix_get(mat,0,1) + gsl_matrix_get(mat,1,0))*s);
        GSL_QUATERNION_SET_IMAG3 (&q, (gsl_matrix_get(mat,2,0) + gsl_matrix_get(mat,0,2))*s);
        GSL_QUATERNION_SET_REAL (&q, (gsl_matrix_get(mat,2,1) - gsl_matrix_get(mat,1,2))*s);

        break;
      }

      case 1:
      {
        s = sqrt(1 + gsl_matrix_get(mat,1,1) -
          gsl_matrix_get(mat,2,2) - gsl_matrix_get(mat,0,0));

        GSL_QUATERNION_SET_IMAG2 (&q, 0.5*s);

        s = 0.5/s;

        GSL_QUATERNION_SET_IMAG3 (&q, (gsl_matrix_get(mat,1,2) + gsl_matrix_get(mat,2,1))*s);
        GSL_QUATERNION_SET_IMAG1 (&q, (gsl_matrix_get(mat,1,0) + gsl_matrix_get(mat,0,1))*s);
        GSL_QUATERNION_SET_REAL (&q, (gsl_matrix_get(mat,0,2) - gsl_matrix_get(mat,2,0))*s);

        break;
      }

      case 2:
      {
        s = sqrt(1 + gsl_matrix_get(mat,2,2) -
          gsl_matrix_get(mat,0,0) - gsl_matrix_get(mat,1,1));

        GSL_QUATERNION_SET_IMAG3 (&q, 0.5*s);

        s = 0.5/s;

        GSL_QUATERNION_SET_IMAG1 (&q, (gsl_matrix_get(mat,2,0) + gsl_matrix_get(mat,0,2))*s);
        GSL_QUATERNION_SET_IMAG2 (&q, (gsl_matrix_get(mat,2,1) + gsl_matrix_get(mat,1,2))*s);
        GSL_QUATERNION_SET_REAL (&q, (gsl_matrix_get(mat,1,0) - gsl_matrix_get(mat,0,1))*s);

        break;
      }

    }

  }

  return q;

}

/* converts Euler angles to a unit quaternion */
gsl_quaternion gsl_quaternion_euler_to_unit_quat(double theta, double phi, double psi)
{

  gsl_quaternion q_rot;

  q_rot.dat[0] = cos(phi/2)*cos(theta/2)*cos(psi/2) + sin(phi/2)*sin(theta/2)*sin(psi/2);
  q_rot.dat[1] = sin(phi/2)*cos(theta/2)*cos(psi/2) - cos(phi/2)*sin(theta/2)*sin(psi/2);
  q_rot.dat[2] = cos(phi/2)*sin(theta/2)*cos(psi/2) + sin(phi/2)*cos(theta/2)*sin(psi/2);
  q_rot.dat[3] = cos(phi/2)*cos(theta/2)*sin(psi/2) - sin(phi/2)*sin(theta/2)*cos(psi/2);

  return q_rot;

}

/* converts a unit quaternion to euler angles */
void gsl_quaternion_unit_quat_to_euler(gsl_quaternion q, double * theta, double * phi, double * psi)
{

  *phi   = atan2( 2*(q.dat[0]*q.dat[1]+q.dat[2]*q.dat[3]), 1.0 - 2*(q.dat[1]*q.dat[1]+q.dat[2]*q.dat[2]) );

  *theta = asin( 2*(q.dat[0]*q.dat[2]-q.dat[3]*q.dat[1]) );

  *psi   = atan2( 2*(q.dat[0]*q.dat[3]+q.dat[1]*q.dat[2]), 1.0 - 2*(q.dat[2]*q.dat[2]+q.dat[3]*q.dat[3]) );

}
