/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <rigid_body.h>

using namespace std;

/*** static workspace init *************************************************/

gsl_vector *     emd::rigid_body::tmp_v  =  gsl_vector_alloc(3);

/***************************************************************************/

/* constructor */
emd::rigid_body::rigid_body(const int & class_id, const int & idx, const int & stat)
{

  /* type */
  this->class_id    =  class_id;
  this->idx         =  idx;
  this->status      =  stat;
  this->cell_state  =  EMD_RIGIDBODY_CELL_STATE_INIT;
  this->cell_idx    =  -1;

  potential = 0;
  kinetic   = 0;

  /* basic memory allocation */
  x = gsl_vector_alloc(3);
  R = gsl_matrix_alloc(3, 3);

  /* extended allocation on request */
  ext_alloc       =  false;
  dipole_init     =  false;

#ifdef RIGID_BODY_DEBUG
  std::cout << "* emd::rigid_body: constructed at " << this << std::endl;
#endif

}

/***************************************************************************/

/* copy constructor */
emd::rigid_body::rigid_body(const rigid_body & r_rigid_body)
{

  x = gsl_vector_alloc(3);
  R = gsl_matrix_alloc(3, 3);

  ext_alloc    =  false;
  dipole_init  =  false;

  class_id     = r_rigid_body.class_id;
  idx          = r_rigid_body.idx;
  int_mask     = r_rigid_body.int_mask;
  status       = r_rigid_body.status;
  cell_state   = r_rigid_body.cell_state;
  cell_idx     = r_rigid_body.cell_idx;

  potential    = r_rigid_body.potential;
  kinetic      = r_rigid_body.kinetic;

  param        = r_rigid_body.param;

  gsl_vector_memcpy(x, r_rigid_body.x);
  gsl_matrix_memcpy(R, r_rigid_body.R);

  Q            = r_rigid_body.Q;
  P            = r_rigid_body.P;
  l            = r_rigid_body.l;
  omega        = r_rigid_body.omega;

  if (r_rigid_body.is_ext_alloc())
  {

    malloc_ext();

    gsl_vector_memcpy(v, r_rigid_body.v);
    gsl_vector_memcpy(force, r_rigid_body.force);
    gsl_vector_memcpy(torque, r_rigid_body.torque);

  }

  if (r_rigid_body.has_dipole())
  {
    this->set_dipole(gsl_vector_get(r_rigid_body.dipole_cm, 0),
                     gsl_vector_get(r_rigid_body.dipole_cm, 1),
                     gsl_vector_get(r_rigid_body.dipole_cm, 2));
  }

#ifdef RIGID_BODY_DEBUG
  std::cout << "* emd::rigid_body: copy constructed at " << this << std::endl;
#endif

}

/***************************************************************************/

/* dynamical variables memory allocation */
void
emd::rigid_body::malloc_ext()
{

  v           = gsl_vector_calloc(3);
  force       = gsl_vector_calloc(3);
  torque      = gsl_vector_calloc(3);

  ext_alloc   = true;

  construct_phase();

#ifdef RIGID_BODY_DEBUG
  cout << "* emd::rigid_body:malloc_ext: extended memory allocated for (" << this << ")." << endl;
#endif
}

/***************************************************************************/

int
emd::rigid_body::construct_phase()
{

  /* construct everything from v, omega, x and R */

  Q = gsl_quaternion_mat_to_unit_quat(R);

  if (ext_alloc)
  {

    /* l = I_body omega */
    l.dat[0] = 0;
    l.dat[1] = (param->I1) * omega.dat[1];
    l.dat[2] = (param->I2) * omega.dat[2];
    l.dat[3] = (param->I3) * omega.dat[3];

    /* P = 2 Q l */
    P = gsl_quaternion_mul_real( gsl_quaternion_mul(Q, l), 2 );

  }

  return EXIT_SUCCESS;
}

/***************************************************************************/

void
emd::rigid_body::set_dipole(const double & p_x, const double & p_y, const double & p_z)
{

  if (!dipole_init)
  {
    dipole_cm = gsl_vector_alloc(3);
  }

  gsl_vector_set(dipole_cm, 0, p_x);
  gsl_vector_set(dipole_cm, 1, p_y);
  gsl_vector_set(dipole_cm, 2, p_z);

  dipole_init = true;

}

/***************************************************************************/

/* destructor */
emd::rigid_body::~rigid_body()
{

  /* state variables */
  gsl_vector_free(x);
  gsl_matrix_free(R);

  if (ext_alloc)
  {

    gsl_vector_free(v);

    gsl_vector_free(force);
    gsl_vector_free(torque);

  }

  if (dipole_init)
  {
    gsl_vector_free(dipole_cm);
  }

#ifdef RIGID_BODY_DEBUG
  std::cout << "* emd::rigid_body: destructed " << this << std::endl;
#endif

}

/***************************************************************************/

/* assignment */

const
emd::rigid_body & emd::rigid_body::operator=(const emd::rigid_body & r_rigid_body)
{

  ext_alloc    =  false;
  dipole_init  =  false;

  class_id     = r_rigid_body.class_id;
  idx          = r_rigid_body.idx;
  int_mask     = r_rigid_body.int_mask;
  status       = r_rigid_body.status;
  cell_state   = r_rigid_body.cell_state;
  cell_idx     = r_rigid_body.cell_idx;

  potential    = r_rigid_body.potential;
  kinetic      = r_rigid_body.kinetic;

  param        = r_rigid_body.param;

  gsl_vector_memcpy(x, r_rigid_body.x);
  gsl_matrix_memcpy(R, r_rigid_body.R);

  Q            = r_rigid_body.Q;
  P            = r_rigid_body.P;
  l            = r_rigid_body.l;
  omega        = r_rigid_body.omega;

  if (r_rigid_body.is_ext_alloc())
  {

    if (!this->is_ext_alloc()) malloc_ext();

    gsl_vector_memcpy(v, r_rigid_body.v);
    gsl_vector_memcpy(force, r_rigid_body.force);
    gsl_vector_memcpy(torque, r_rigid_body.torque);

  }

  if (r_rigid_body.has_dipole())
  {

    this->set_dipole(gsl_vector_get(r_rigid_body.dipole_cm, 0),
                     gsl_vector_get(r_rigid_body.dipole_cm, 1),
                     gsl_vector_get(r_rigid_body.dipole_cm, 2));
  }


#ifdef RIGID_BODY_DEBUG
  std::cout << "* emd::rigid_body: copied " << this << std::endl;
#endif

  return *this;

}
