/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/***************************************************************************
 * todo:                                                                   *
 *                                                                         *
 ***************************************************************************/

/***************************************************************************
 * changelog:                                                              *
 *                                                                         *
 ***************************************************************************/


#ifndef RIGID_BODY_H
#define RIGID_BODY_H

#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_linalg.h>

#include <quaternion/gsl_quaternion.h>
#include <quaternion/gsl_quaternion_math.h>

#include "consts.h"
#include "gslext.h"
#include "paramdb.h"

//#define RIGID_BODY_DEBUG

namespace emd
{

  class rigid_body
  {
  private:

    bool ext_alloc;
    bool dipole_init;

  public:

    /*************************************************************************************/

    static const int  EMD_RIGIDBODY_STATUS_INACTIVE      = 0;
    static const int  EMD_RIGIDBODY_STATUS_ACTIVE        = 1;

    static const int  EMD_RIGIDBODY_CELL_STATE_NONE      = 0;
    static const int  EMD_RIGIDBODY_CELL_STATE_INIT      = 1;
    static const int  EMD_RIGIDBODY_CELL_STATE_UPDATE    = 2;

    /*************************************************************************************/

    /* variables */
    int               class_id;        /* sub type */
    int               idx;             /* index in paramdb caches */
    int               int_mask;        /* interaction bit mask */
    int               status;          /* state of activity */
    int               cell_state;      /* cell state */
    int               cell_idx;        /* cell index */

    double            potential;
    double            kinetic;

    /* a pointer to parameters */
    paramdb::param_cache_rigid_body *   param;

    /* state variables */
    gsl_vector *      x;                /* center of mass position    x(t)  */
    gsl_vector *      v;                /* center of mass velocity    v(t)  */

    gsl_quaternion    Q;                /* rotation quaternion        q(t)  */
    gsl_quaternion    P;                /* conjugation momentum of q  dq(t) */

    /* derived quantities (auxilliary variables) */
    gsl_matrix *      R;                /* rotation matrix (to be constructed from q)   R(t) */
    gsl_quaternion    l;                /* total angular momentum  (im part of a quat)  L(t) */
    gsl_quaternion    omega;            /* angular velocity (im part of a quat)         \omega(t) */

    gsl_vector *      dipole_cm;

    /* computed quantities */
    gsl_vector *      force;            /* total force                       F(t) */
    gsl_vector *      torque;           /* total torque with respect to CM   N(t) */

    /*** static workspace ********************************************/

    static gsl_vector *      tmp_v;

    /*****************************************************************/

    /* constructor */
    rigid_body (const int & class_id = 0, const int & idx = 0, const int & stat = 1);

    /* copy constructor */
    rigid_body (const rigid_body & r_rigid_body);

    /* lazy initialization */
    void malloc_ext();

    /* evaluate phase space variables using the input file variables */
    int construct_phase();

    /* add dipole */
    void set_dipole(const double & p_x, const double & p_y, const double & p_z);

    bool is_ext_alloc() const { return ext_alloc; };

    bool has_dipole() const { return dipole_init; };

    /* destructor */
    ~rigid_body();

    /* assignment */
    const rigid_body & operator=(const rigid_body & r_rigid_body);

  };

  /*******************************************************************/

  struct rigid_body_ref
  {

    unsigned int idx;         /* index */

    int n_x;                  /* image index (x) */
    int n_y;                  /* image index (y) */
    int n_z;                  /* image index (z) */

  };

};

#endif
