/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/************************************************
 * a simple and universal input file parser     *
 ************************************************/

/*************************************************************
 * todo:                                                     *
 *                                                           *
 * - ignore the characters after '#' in the middle of a line *
 *************************************************************/

#ifndef UNIPARSER_H
#define UNIPARSER_H

#include <cstdlib>

#include <iostream>
#include <fstream>
#include <string>
#include <vector>

#include <stringext.h>

#define UNIPARSER_DEBUG

namespace emd
{

  /*
   *  an input file parser
   *
   *  @inf              input file stream
   *  @n_keywords       number of keywords to look for
   *  @str_keywords     array of std::string keywords
   *  @key_action       sub parser (array of pointer to function)
   *  @comm             comment character
   *  @comm_active      consider comments or not
   *
   */
  static int
  file_parser(
    std::ifstream & inf, const int & n_keywords,
    std::string* str_keywords, int (**key_action)(const std::string & param_buff),
    const char& comm, const bool& comm_active)
  {

    if (!inf)
    {
      std::cerr << "* emd::file_parser: Error opening the input file!" << std::endl;
      return EXIT_FAILURE;
    }

    /* parsing the file */

    std::string curr_line;
    std::string param_buffer, keystr;

    int stat, key_id = -1;
    int ln = 0; /* line num */

    while (true)
    {

      getline(inf, curr_line);
      ln++;

      if (inf.eof()) /* eof */
      {
        /* parse the buffer */
        if (key_id!=-1) /* was parsing a keyword, eof reached */
        {
          /* remove the keyword from the buffer */
          param_buffer.replace(tolower(param_buffer).find(str_keywords[key_id]),
            str_keywords[key_id].size(), "");

          stat = (*key_action[key_id])(param_buffer);
          if (stat==EXIT_FAILURE)
          {
            std::cerr << "* emd::file_parser: error parsing the input file! (around line: " <<
              ln << ")" << std::endl;
            return EXIT_FAILURE;
          }
        }
        break;
      }

      if (!curr_line.length()) /* empty line */
        continue;

      /* lets investigate if it is a comment line */
      size_t begin, end;
      if (std::string_next_token(curr_line, 0, begin, end)) continue;

      if (comm_active)
      {
        if (curr_line[begin]==comm) continue;

        /* lets remove the trailing characters after the comment */
        int pos = curr_line.find(comm);
        if (pos!=std::string::npos)
          curr_line = curr_line.substr(0, pos);
      }

      /* it is not a comment line or an empty line
        it must be
         1) the begining of a parameter set
         2) or the rest of the previous parameter set
      */

      keystr = tolower(curr_line.substr(begin, end-begin+1));

      /* look for keywords */
      for (int i=0; i<n_keywords; i++)
        if ( keystr == str_keywords[i] )
        {
          if (key_id!=-1) /* was parsing a keyword, a new one is reached */
          {
            /* remove the keyword from the buffer */
            param_buffer.replace(tolower(param_buffer).find(str_keywords[key_id]),
              str_keywords[key_id].size(), "");

            stat = (*key_action[key_id])(param_buffer); /* parse the buffer */
            if (stat)
            {
              std::cerr << "* emd::file_parser: error parsing the input file (around line: " << ln 
                << ")" << std::endl;
              return EXIT_FAILURE;
            }
          }

          /* was not parsing a keyword, a new one is reached */
          key_id = i;
          param_buffer = "";

          break;
        }

      if (key_id==-1) /* was not parsing a keyword, no keyword was noticed either! */
      {
        std::cerr << "* emd::file_parser: error parsing the input file! (ln: " <<
          ln << ")" << std::endl;
        return EXIT_FAILURE;
      }
      else
      {
        /* append */
        param_buffer.append(curr_line);
      }

    }

    return EXIT_SUCCESS;

  }

};

#endif /* UNIPARSER_H */
