/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "universeloader.h"
//{{{restart: include "dynamics.h"
#include "dynamics.h"
//}}}
using namespace std;
//{{restart :initialization of t_start
double emd::uni::loader::t_start           = 0 ;
//}}}
/*
 * loads a universe from a file
 *
 * @input_file_name            a saved universe file
 * @clear                      clear the current vectors
 *
 */
int
emd::uni::loader::load_file (const std::string & input_file_name, const bool& clear)
{

  if (clear)
    emd::uni::free_memory();

  parsed_rigid_bodies = 0;
  parsed_flat_int_surfs = 0;
  parsed_sph_int_surfs = 0;

  /* opening the specified file as ro */
  ifstream inf(input_file_name.c_str(), ios::in);

  /* parsing */
  int stat =  emd::file_parser(inf, sizeof(KEYWORDS)/sizeof(string*), KEYWORDS, ACTIONS, '#', true);
  if (stat==EXIT_FAILURE) return EXIT_FAILURE;

  if ( parsed_rigid_bodies < emd::uni::rigid_bodies_count )
    cout << "* emd::uni::loader::load_file: warning! the number of defined rigid bodys "
         << "is less than the allocated memory!" << endl;

  if ( parsed_flat_int_surfs < emd::uni::flat_int_surfs_count )
    cout << "* emd::uni::loader::load_file: warning! the number of defined flat int. surfaces "
         << "is less than the allocated memory!" << endl;

  if ( parsed_sph_int_surfs < emd::uni::sph_int_surfs_count )
    cout << "* emd::uni::loader::load_file: warning! the number of defined spherical int. surfaces "
         << "is less than the allocated memory!" << endl;

  /* post processing */
  emd::uni::init_universe_post();

  return stat;

}
//********************************************************************'
void emd::uni::loader::set_ext_param_file(std::string inp_param_file){
  ext_param_file = inp_param_file;
  ext_param = true;
}


/*****************************************************************************************/

/* parses the "param_file" keyword */
int
emd::uni::loader::parse_param_file (const std::string & param_buffer)
{

  vector<string> tokenized_buffer;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  /* a certain number of parameters are required! */
  if (tokenized_buffer.size()<1)
    return EXIT_FAILURE;
  //change
  if(emd::uni::loader::ext_param) 
    tokenized_buffer[0] = emd::uni::loader::ext_param_file;
     
  int stat = emd::uni::load_param_file( tokenized_buffer[0], true );

  if (stat==EXIT_FAILURE)
    return EXIT_FAILURE;

  return EXIT_SUCCESS;
}

/*****************************************************************************************/

/* parses the "rigid_body_num" keyword */
int
emd::uni::loader::parse_rigid_body_num (const std::string & param_buffer)
{

  vector<string> tokenized_buffer;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  /* a certain number of parameters are required! */
  if (tokenized_buffer.size()<1)
    return EXIT_FAILURE;

  int num = atoi(tokenized_buffer[0].c_str());

  int stat = emd::uni::malloc_rigid_body(num);

  if (stat==EXIT_FAILURE)
    return EXIT_FAILURE;

  return EXIT_SUCCESS;
}

/*****************************************************************************************/

/* parses the "flat_int_surf_num" keyword */
int
emd::uni::loader::parse_flat_int_surf_num (const std::string & param_buffer)
{

  vector<string> tokenized_buffer;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  /* a certain number of parameters are required! */
  if (tokenized_buffer.size()<1)
    return EXIT_FAILURE;

  int num = atoi(tokenized_buffer[0].c_str());

  int stat = emd::uni::malloc_flat_int_surf(num);

  if (stat==EXIT_FAILURE)
    return EXIT_FAILURE;

  return EXIT_SUCCESS;
}

/*****************************************************************************************/

/* parses the "sph_int_surf_num" keyword */
int
emd::uni::loader::parse_sph_int_surf_num (const std::string & param_buffer)
{

  vector<string> tokenized_buffer;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  /* a certain number of parameters are required! */
  if (tokenized_buffer.size()<1)
    return EXIT_FAILURE;

  int num = atoi(tokenized_buffer[0].c_str());

  int stat = emd::uni::malloc_spherical_int_surf(num);

  if (stat==EXIT_FAILURE)
    return EXIT_FAILURE;

  return EXIT_SUCCESS;
}

/*****************************************************************************************/

/* parses the "rigid_body" keyword */
int
emd::uni::loader::parse_rigid_body (const std::string & param_buffer)
{

  if ( !emd::paramdb::is_init() )
  {
    cerr << "* emd::uni::loader::load_file: parameters are not available!" << endl;
    return EXIT_FAILURE;
  }

  if (parsed_rigid_bodies>emd::uni::rigid_bodies_count)
  {
    cerr << "* emd::uni::loader::parse_rigid_body: allocate more memory for rigid bodys!" << endl;
    return EXIT_FAILURE;
  }

  vector<string> tokenized_buffer;
  rigid_body& new_rigid_body = emd::uni::rigid_bodies[parsed_rigid_bodies];

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  /* a certain number of parameters are required! */
  if (tokenized_buffer.size()<9)
    return EXIT_FAILURE;

  /* types and status */
  new_rigid_body.class_id  =   atoi(tokenized_buffer[0].c_str());
  new_rigid_body.idx       =   emd::paramdb::get_rigid_body_index(new_rigid_body.class_id);
  new_rigid_body.status    =   atoi(tokenized_buffer[1].c_str());
  new_rigid_body.int_mask  =   atoi(tokenized_buffer[2].c_str());
  new_rigid_body.param     =   &emd::paramdb::rigid_body_param[ new_rigid_body.idx ];

  /* CM location */
  gsl_vector_set( new_rigid_body.x, 0, atof(tokenized_buffer[3].c_str()) );
  gsl_vector_set( new_rigid_body.x, 1, atof(tokenized_buffer[4].c_str()) );
  gsl_vector_set( new_rigid_body.x, 2, atof(tokenized_buffer[5].c_str()) );

  /* euler angles */
  double theta = TO_RAD * atof(tokenized_buffer[6].c_str());
  double phi   = TO_RAD * atof(tokenized_buffer[7].c_str());
  double psi   = TO_RAD * atof(tokenized_buffer[8].c_str());

  gsl_matrix_euler_rot( theta, phi, psi, new_rigid_body.R );

  /* extended initial condition (dynamical variables) */
  if (tokenized_buffer.size()>=15)
  {

    new_rigid_body.malloc_ext();

    /* CM velocity */
    gsl_vector_set( new_rigid_body.v, 0, atof(tokenized_buffer[9].c_str()) );
    gsl_vector_set( new_rigid_body.v, 1, atof(tokenized_buffer[10].c_str()) );
    gsl_vector_set( new_rigid_body.v, 2, atof(tokenized_buffer[11].c_str()) );

    /* angular velocity */
    new_rigid_body.omega.dat[0] = 0;
    new_rigid_body.omega.dat[1] = atof(tokenized_buffer[12].c_str());
    new_rigid_body.omega.dat[2] = atof(tokenized_buffer[13].c_str());
    new_rigid_body.omega.dat[3] = atof(tokenized_buffer[14].c_str());

  }

  new_rigid_body.construct_phase();

  tokenized_buffer.clear();

  parsed_rigid_bodies++;


#ifdef UNIVERSELOADER_DEBUG
  cout << "* emd::uni::loader::parse_rigid_body: (defining)" << endl
       << "  class_id:  " << new_rigid_body.class_id << endl
       << "  idx:       " << new_rigid_body.idx << endl
       << "  status:    " << new_rigid_body.status << endl;
  cout << setprecision(8) << showpos << showpoint << left;
  cout << "  x_c:       " << setw(8) << gsl_vector_get(new_rigid_body.x, 0) << endl
       << "  y_c:       " << setw(8) << gsl_vector_get(new_rigid_body.x, 1) << endl
       << "  z_c:       " << setw(8) << gsl_vector_get(new_rigid_body.x, 2) << endl
//       << "  theta:     " << setw(8) << theta << " (rad)" << endl
//       << "  phi:       " << setw(8) << phi << " (rad)" << endl
//       << "  psi:       " << setw(8) << psi << " (rad)" << endl;
       << "  RE[Q]:     " << setw(8) << new_rigid_body.Q.dat[0] << endl
       << "  IM1[Q]:    " << setw(8) << new_rigid_body.Q.dat[1] << endl
       << "  IM2[Q]:    " << setw(8) << new_rigid_body.Q.dat[2] << endl
       << "  IM3[Q]:    " << setw(8) << new_rigid_body.Q.dat[3] << endl;

  if ( new_rigid_body.is_ext_alloc() )
    cout << "  v_x:       " << setw(8) << gsl_vector_get(new_rigid_body.v, 0) << endl
         << "  v_y:       " << setw(8) << gsl_vector_get(new_rigid_body.v, 1) << endl
         << "  v_z:       " << setw(8) << gsl_vector_get(new_rigid_body.v, 2) << endl
         << "  omega_1:   " << setw(8) << new_rigid_body.omega.dat[1] << endl
         << "  omega_2:   " << setw(8) << new_rigid_body.omega.dat[2] << endl
         << "  omega_3:   " << setw(8) << new_rigid_body.omega.dat[3] << endl;

  cout << noshowpos << noshowpoint;
#endif

  return EXIT_SUCCESS;
}

/*****************************************************************************************/

/* parses the "rigid_body"_dipole keyword */
int
emd::uni::loader::parse_rigid_body_dipole (const std::string & param_buffer)
{

  vector<string> tokenized_buffer;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  /* a certain number of parameters are required! */
  if (tokenized_buffer.size()<3)
    return EXIT_FAILURE;

  double p_x = atof(tokenized_buffer[0].c_str());
  double p_y = atof(tokenized_buffer[1].c_str());
  double p_z = atof(tokenized_buffer[2].c_str());

  emd::uni::rigid_bodies[parsed_rigid_bodies-1].set_dipole(p_x, p_y, p_z);

#ifdef UNIVERSELOADER_DEBUG
  cout << "* emd::uni::loader::parse_rigid_body_dipole:" << endl
       << "  adding a dipole to rigid_body " << (parsed_rigid_bodies-1) << ": " << endl;
  cout << setprecision(8) << showpos << showpoint << left
       << "  p_x: " << setw(8) << p_x << endl
       << "  p_y: " << setw(8) << p_y << endl
       << "  p_z: " << setw(8) << p_z << endl;

  cout << noshowpos << noshowpoint;
#endif

  return EXIT_SUCCESS;
}

/*****************************************************************************************/

/* parses the "flat_int_surf" keyword */
int
emd::uni::loader::parse_flat_int_surf (const std::string & param_buffer)
{

  if ( !emd::paramdb::is_init() )
  {
    cerr << "* emd::uni::loader::load_file: parameters are not available!" << endl;
    return EXIT_FAILURE;
  }

  if ( parsed_flat_int_surfs > emd::uni::flat_int_surfs_count )
  {
    cerr << "* emd::uni::loader::parse_flat_int_surf: allocate more memory for flat int. surfaces!" << endl;
    return EXIT_FAILURE;
  }

  vector<string> tokenized_buffer;
  flat_int_surface& new_flat_int_surface = emd::uni::flat_int_surfs[parsed_flat_int_surfs];

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  /* a certain number of parameters are required! */
  if (tokenized_buffer.size()<5)
    return EXIT_FAILURE;

  new_flat_int_surface.class_id  =   atoi(tokenized_buffer[0].c_str());  /* class_id */
  new_flat_int_surface.idx       =   emd::paramdb::get_int_surf_index(new_flat_int_surface.class_id);
  double r                       =   atof(tokenized_buffer[1].c_str());  /* distance from origin */

  /* nhat */
  gsl_vector_set( new_flat_int_surface.n, 0, atof(tokenized_buffer[2].c_str()) );
  gsl_vector_set( new_flat_int_surface.n, 1, atof(tokenized_buffer[3].c_str()) );
  gsl_vector_set( new_flat_int_surface.n, 2, atof(tokenized_buffer[4].c_str()) );

  /* normalize nhat */
  gsl_vector_normalize( new_flat_int_surface.n );

  /* r0 */
  gsl_vector_memcpy( new_flat_int_surface.r0, new_flat_int_surface.n);
  gsl_vector_scale( new_flat_int_surface.r0, -r);

  tokenized_buffer.clear();

  parsed_flat_int_surfs++;


#ifdef UNIVERSELOADER_DEBUG
  cout << "* emd::uni::loader::parse_flat_int_surf: (defining)" << endl
       << "  class_id:  " << new_flat_int_surface.class_id << endl
       << "  idx:       " << new_flat_int_surface.idx << endl;
  cout << setprecision(8) << showpos << showpoint << left;
  cout << "  x_0:       " << setw(8) << gsl_vector_get(new_flat_int_surface.r0, 0) << endl
       << "  y_0:       " << setw(8) << gsl_vector_get(new_flat_int_surface.r0, 1) << endl
       << "  z_0:       " << setw(8) << gsl_vector_get(new_flat_int_surface.r0, 2) << endl
       << "  n_x:       " << setw(8) << gsl_vector_get(new_flat_int_surface.n, 0) << endl
       << "  n_y:       " << setw(8) << gsl_vector_get(new_flat_int_surface.n, 1) << endl
       << "  n_z:       " << setw(8) << gsl_vector_get(new_flat_int_surface.n, 2) << endl;

  cout << noshowpos << noshowpoint;
#endif

  return EXIT_SUCCESS;
}

/*****************************************************************************************/

/* parses the "sph_int_surf" keyword */
int
emd::uni::loader::parse_sph_int_surf (const std::string & param_buffer)
{

  if ( !emd::paramdb::is_init() )
  {
    cerr << "* emd::uni::loader::load_file: parameters are not available!" << endl;
    return EXIT_FAILURE;
  }

  if ( parsed_flat_int_surfs > emd::uni::flat_int_surfs_count )
  {
    cerr << "* emd::uni::loader::parse_sph_int_surf: allocate more memory for sph int. surfaces!" << endl;
    return EXIT_FAILURE;
  }

  vector<string> tokenized_buffer;
  spherical_int_surface& new_spherical_int_surface = emd::uni::sph_int_surfs[parsed_sph_int_surfs];

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  /* a certain number of parameters are required! */
  if (tokenized_buffer.size()<5)
    return EXIT_FAILURE;

  new_spherical_int_surface.class_id  =   atoi(tokenized_buffer[0].c_str());  /* class_id */
  new_spherical_int_surface.idx       =   emd::paramdb::get_int_surf_index(new_spherical_int_surface.class_id);
  new_spherical_int_surface.rad       =   atof(tokenized_buffer[1].c_str());  /* radius */

  /* r0 */
  gsl_vector_set( new_spherical_int_surface.r0, 0, atof(tokenized_buffer[2].c_str()) );
  gsl_vector_set( new_spherical_int_surface.r0, 1, atof(tokenized_buffer[3].c_str()) );
  gsl_vector_set( new_spherical_int_surface.r0, 2, atof(tokenized_buffer[4].c_str()) );

  tokenized_buffer.clear();

  parsed_sph_int_surfs++;

#ifdef UNIVERSELOADER_DEBUG
  cout << "* emd::uni::loader::parse_sph_int_surf: (defining)" << endl
       << "  class_id:  " << new_spherical_int_surface.class_id << endl
       << "  idx:       " << new_spherical_int_surface.idx << endl;
  cout << setprecision(8) << showpos << showpoint << left;
  cout << "  radius:    " << setw(8) << new_spherical_int_surface.rad << endl
       << "  x_0:       " << setw(8) << gsl_vector_get(new_spherical_int_surface.r0, 0) << endl
       << "  y_0:       " << setw(8) << gsl_vector_get(new_spherical_int_surface.r0, 1) << endl
       << "  z_0:       " << setw(8) << gsl_vector_get(new_spherical_int_surface.r0, 2) << endl;

  cout << noshowpos << noshowpoint;
#endif

  return EXIT_SUCCESS;
}

/*****************************************************************************************/

/* parses the "x_periodic" keyword */
int
emd::uni::loader::parse_x_periodic(const std::string & param_buffer)
{

  vector<string> tokenized_buffer;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  /* a certain number of parameters are required! */
  if (tokenized_buffer.size()<2)
    return EXIT_FAILURE;

  /* types and status */
  double x_lower =   atof(tokenized_buffer[0].c_str());
  double x_upper =   atof(tokenized_buffer[1].c_str());

  if (x_upper>x_lower)
  {
    emd::uni::x_periodic    = true;
    emd::uni::x_lower_bound = x_lower;
    emd::uni::x_upper_bound = x_upper;
    emd::uni::x_width       = x_upper - x_lower;
    emd::uni::x_center      = (x_lower + x_upper)/2.0;
  }
  else
  {
    cout << "* emd::uni::loader::parse_x_periodic: the lower bound must be lower than the upper bound!" << endl;
    return EXIT_FAILURE;
  }

#ifdef UNIVERSELOADER_DEBUG
  cout << "* emd::uni::loader::parse_x_eriodic:" << endl;
  cout << setprecision(8) << showpos << showpoint << left;
  cout << "  x_lower_bound :  " << setw(8) << emd::uni::x_lower_bound << endl
       << "  x_upper_bound :  " << setw(8) << emd::uni::x_upper_bound << endl;
  cout << noshowpos << noshowpoint;
#endif

  return EXIT_SUCCESS;
}

/*****************************************************************************************/

/* parses the "y_periodic" keyword */
int
emd::uni::loader::parse_y_periodic(const std::string & param_buffer)
{

  vector<string> tokenized_buffer;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  /* a certain number of parameters are required! */
  if (tokenized_buffer.size()<2)
    return EXIT_FAILURE;

  /* types and status */
  double y_lower =   atof(tokenized_buffer[0].c_str());
  double y_upper =   atof(tokenized_buffer[1].c_str());

  if (y_upper>y_lower)
  {
    emd::uni::y_periodic    = true;
    emd::uni::y_lower_bound = y_lower;
    emd::uni::y_upper_bound = y_upper;
    emd::uni::y_width       = y_upper - y_lower;
    emd::uni::y_center      = (y_lower + y_upper)/2.0;
  }
  else
  {
    cout << "* emd::uni::loader::parse_y_periodic: the lower bound must be lower than the upper bound!" << endl;
    return EXIT_FAILURE;
  }

#ifdef UNIVERSELOADER_DEBUG
  cout << "* emd::uni::loader::parse_y_periodic:" << endl;
  cout << setprecision(8) << showpos << showpoint << left;
  cout << "  y_lower_bound :  " << setw(8) << emd::uni::y_lower_bound << endl
       << "  y_upper_bound :  " << setw(8) << emd::uni::y_upper_bound << endl;
  cout << noshowpos << noshowpoint;
#endif

  return EXIT_SUCCESS;
}

/*****************************************************************************************/

/* parses the "z_periodic" keyword */
int
emd::uni::loader::parse_z_periodic(const std::string & param_buffer)
{

  vector<string> tokenized_buffer;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  /* a certain number of parameters are required! */
  if (tokenized_buffer.size()<2)
    return EXIT_FAILURE;

  /* types and status */
  double z_lower =   atof(tokenized_buffer[0].c_str());
  double z_upper =   atof(tokenized_buffer[1].c_str());

  if (z_upper>z_lower)
  {
    emd::uni::z_periodic    = true;
    emd::uni::z_lower_bound = z_lower;
    emd::uni::z_upper_bound = z_upper;
    emd::uni::z_width       = z_upper - z_lower;
    emd::uni::z_center      = (z_lower + z_upper)/2.0;
  }
  else
  {
    cout << "* emd::uni::loader::parse_z_periodic: the lower bound must be lower than the upper bound!" << endl;
    return EXIT_FAILURE;
  }

#ifdef UNIVERSELOADER_DEBUG
  cout << "* emd::uni::loader::parse_z_periodic:" << endl;
  cout << setprecision(8) << showpos << showpoint << left;
  cout << "  z_lower_bound :  " << setw(8) << emd::uni::z_lower_bound << endl
       << "  z_upper_bound :  " << setw(8) << emd::uni::z_upper_bound << endl;
  cout << noshowpos << noshowpoint;
#endif

  return EXIT_SUCCESS;
}

/*****************************************************************************************/

/* parses the "interior_point" keyword */
int
emd::uni::loader::parse_interior_point(const std::string & param_buffer)
{

  vector<string> tokenized_buffer;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  /* a certain number of parameters are required! */
  if (tokenized_buffer.size()<3)
    return EXIT_FAILURE;

  /* types and status */
  double int_point_x =   atof(tokenized_buffer[0].c_str());
  double int_point_y =   atof(tokenized_buffer[1].c_str());
  double int_point_z =   atof(tokenized_buffer[2].c_str());

  gsl_vector_set(emd::uni::interior_point, 0, int_point_x);
  gsl_vector_set(emd::uni::interior_point, 1, int_point_y);
  gsl_vector_set(emd::uni::interior_point, 2, int_point_z);

#ifdef UNIVERSELOADER_DEBUG
  cout << "* emd::uni::loader::parse_interior_point:" << endl;
  cout << setprecision(8) << showpos << showpoint << left;
  cout << "  x :  " << setw(8) << gsl_vector_get(emd::uni::interior_point, 0) << endl
       << "  y :  " << setw(8) << gsl_vector_get(emd::uni::interior_point, 1) << endl
       << "  z :  " << setw(8) << gsl_vector_get(emd::uni::interior_point, 2) << endl;
  cout << noshowpos << noshowpoint;
#endif

  return EXIT_SUCCESS;
}

/*****************************************************************************************/

/* parses the "global_electrical_field" keyword */
int
emd::uni::loader::parse_global_electrical_field(const std::string & param_buffer)
{

  vector<string> tokenized_buffer;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  /* a certain number of parameters are required! */
  if (tokenized_buffer.size()<3)
    return EXIT_FAILURE;

  /* types and status */
  double E_x =   atof(tokenized_buffer[0].c_str());
  double E_y =   atof(tokenized_buffer[1].c_str());
  double E_z =   atof(tokenized_buffer[2].c_str());

  gsl_vector_set(emd::uni::E_field, 0, E_x);
  gsl_vector_set(emd::uni::E_field, 1, E_y);
  gsl_vector_set(emd::uni::E_field, 2, E_z);

#ifdef UNIVERSELOADER_DEBUG
  cout << "* emd::uni::loader::parse_global_electrical_field:" << endl;
  cout << setprecision(8) << showpos << showpoint << left;
  cout << "  E_x :  " << setw(8) << gsl_vector_get(emd::uni::E_field, 0) << endl
       << "  E_y :  " << setw(8) << gsl_vector_get(emd::uni::E_field, 1) << endl
       << "  E_z :  " << setw(8) << gsl_vector_get(emd::uni::E_field, 2) << endl;
  cout << noshowpos << noshowpoint;
#endif

  return EXIT_SUCCESS;
}
/*****************************************************************************************/
//{{{restart : definition of parse_time 
/* parses the "time" keyword */
int
emd::uni::loader::parse_time (const std::string & param_buffer)
{

  vector<string> tokenized_buffer;

  /* tokenize the buffer */
  string_tokenize(param_buffer, tokenized_buffer);

  /* a certain number of parameters are required! */
  if (tokenized_buffer.size()<1)
    return EXIT_FAILURE;

  //emd::uni::loader::t_start = atof(tokenized_buffer[0].c_str());
  emd::uni::loader::t_start = atof(tokenized_buffer[0].c_str());

  return EXIT_SUCCESS;
}

/*****************************************************************************************/
//}}}
