/***************************************************************************
 *   emd - Ellipsoids Molecular Dynamics                                   *
 *                                                                         *
 *   Copyright (C) 2006 by Mehrtash Babadi                                 *
 *   mehrtash@dena.sharif.edu                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "universesaver.h"

using namespace std;
using namespace emd::uni;

/***************************************************************************/

int
emd::uni::saver::save_scaled_universe_to_file(const double & scale_factor,
  const string & out_file_name)
{

  for (int m=0; m<rigid_bodies_count; m++)
    gsl_vector_scale(rigid_bodies[m].x, scale_factor);

  for (int m=0; m<flat_int_surfs_count; m++)
    gsl_vector_scale(flat_int_surfs[m].r0, scale_factor);

  for (int m=0; m<sph_int_surfs_count; m++)
    gsl_vector_scale(sph_int_surfs[m].r0, scale_factor);

  gsl_vector_scale(interior_point, scale_factor);

  x_lower_bound *= scale_factor;
  x_upper_bound *= scale_factor;
  x_width       *= scale_factor;
  x_center      *= scale_factor;

  y_lower_bound *= scale_factor;
  y_upper_bound *= scale_factor;
  y_width       *= scale_factor;
  y_center      *= scale_factor;

  z_lower_bound *= scale_factor;
  z_upper_bound *= scale_factor;
  z_width       *= scale_factor;
  z_center      *= scale_factor;

  save_universe_to_file (out_file_name);

  for (int m=0; m<rigid_bodies_count; m++)
    gsl_vector_scale(rigid_bodies[m].x, 1/scale_factor);

  for (int m=0; m<flat_int_surfs_count; m++)
    gsl_vector_scale(flat_int_surfs[m].r0, 1/scale_factor);

  for (int m=0; m<sph_int_surfs_count; m++)
    gsl_vector_scale(sph_int_surfs[m].r0, 1/scale_factor);

  gsl_vector_scale(interior_point, 1/scale_factor);

  x_lower_bound /= scale_factor;
  x_upper_bound /= scale_factor;
  x_width       /= scale_factor;
  x_center      /= scale_factor;

  y_lower_bound /= scale_factor;
  y_upper_bound /= scale_factor;
  y_width       /= scale_factor;
  y_center      /= scale_factor;

  z_lower_bound /= scale_factor;
  z_upper_bound /= scale_factor;
  z_width       /= scale_factor;
  z_center      /= scale_factor;

}

/***************************************************************************/

int
emd::uni::saver::save_universe_to_file(const string & out_file_name)
{

  vector<rigid_body_ref> indices;

  for (unsigned int m=0; m<rigid_bodies_count; m++)
  {

    rigid_body_ref ref;

    ref.idx = m;

    ref.n_x = 0;
    ref.n_y = 0;
    ref.n_z = 0;

    indices.push_back(ref);

  }

  save_universe_to_file(indices, out_file_name);

}

/****************************************************************************
 * NOTE:                                                                    *
 *                                                                          *
 * It is a temporary fix; saving a subset of molecules might cause indexing *
 * problems for bonded interactions.                                        *
  ***************************************************************************/

int
emd::uni::saver::save_universe_to_file(const vector<rigid_body_ref> & indices,
  const string & out_file_name)
{

  /* open the file */
  ofstream of(out_file_name.c_str());
  if (!of) return EXIT_FAILURE;

  /***************************/

  of << "#############################" << endl
     << "## saved universe snapshot ##" << endl
     << "#############################" << endl << endl;

  /* param file name */
  of << "param_file           " << emd::paramdb::param_file_name << endl;

  //{{{restart: adding "time" field to the output universe file
  of << "time                 " << emd::uni::dyn::t<<endl<<endl;
   //}}}
  /* write out the number of rigid bodys */
  if (rigid_bodies_count>0)
    of << "rigid_body_num       " << indices.size() << endl << endl; /* NOTE */
  if (flat_int_surfs_count>0)
    of << "flat_int_surf_num    " << emd::uni::flat_int_surfs_count << endl << endl;
  if (sph_int_surfs_count>0)
    of << "sph_int_surf_num     " << emd::uni::sph_int_surfs_count << endl << endl;

  of <<  setprecision(3) << left << fixed << noshowpos;

  if (x_periodic)
    of << "x_periodic               "
       << setw(DEF_WIDTH) << x_lower_bound
       << setw(DEF_WIDTH) << x_upper_bound << endl;

  if (y_periodic)
    of << "y_periodic               "
       << setw(DEF_WIDTH) << y_lower_bound
       << setw(DEF_WIDTH) << y_upper_bound << endl;

  if (z_periodic)
    of << "z_periodic               "
       << setw(DEF_WIDTH) << z_lower_bound
       << setw(DEF_WIDTH) << z_upper_bound << endl;

  of << "interior_point             " << gsl_vector_get(emd::uni::interior_point, 0) << "   "
                                      << gsl_vector_get(emd::uni::interior_point, 1) << "   "
                                      << gsl_vector_get(emd::uni::interior_point, 2) << endl;

  of << "global_electrical_field    " << gsl_vector_get(emd::uni::E_field, 0) << "   "
                                      << gsl_vector_get(emd::uni::E_field, 1) << "   "
                                      << gsl_vector_get(emd::uni::E_field, 2) << endl;

  of << endl;

  /* flat int surfs */

  for (int n=0; n<flat_int_surfs_count; n++)
  {

    flat_int_surface & c_s = emd::uni::flat_int_surfs[n];

    of << "# " << n << endl;
    of << "flat_int_surf    ";
    of <<  setprecision(DEF_PREC) << left << scientific << noshowpos;

    /* id */
    of << setw(DEF_WIDTH) << c_s.class_id << endl;

    of << showpos;

    double dist;
    gsl_blas_ddot(c_s.r0, c_s.n, &dist);

    /* x */
    of << "                 "
       << setw(DEF_WIDTH) << (-dist) << endl
       << "                 "
       << setw(DEF_WIDTH) << gsl_vector_get(c_s.n, 0)
       << setw(DEF_WIDTH) << gsl_vector_get(c_s.n, 1)
       << setw(DEF_WIDTH) << gsl_vector_get(c_s.n, 2) << endl;

    of << endl;

  }

  /* spherical int surfs */

  for (int n=0; n<sph_int_surfs_count; n++)
  {

    spherical_int_surface & c_s = emd::uni::sph_int_surfs[n];

    of << "# " << n << endl;
    of << "sph_int_surf     ";
    of <<  setprecision(DEF_PREC) << left << scientific << noshowpos;

    /* id */
    of << setw(DEF_WIDTH) << c_s.class_id << endl;

    of << showpos;

    /* x */
    of << "                 "
       << setw(DEF_WIDTH) << c_s.rad << endl
       << "                 "
       << setw(DEF_WIDTH) << gsl_vector_get(c_s.r0, 0)
       << setw(DEF_WIDTH) << gsl_vector_get(c_s.r0, 1)
       << setw(DEF_WIDTH) << gsl_vector_get(c_s.r0, 2) << endl;

    of << endl;

  }

  /* rigid bodys */
  double            c_theta, c_phi, c_psi;

  for (int idx=0; idx<indices.size(); idx++)
  {

    int n = indices[idx].idx;

    rigid_body &  c_m = emd::uni::rigid_bodies[n];

    of << "# " << idx << endl;
    of << "rigid_body       ";
    of <<  setprecision(DEF_PREC) << left << scientific << noshowpos;

    /* id */
    of << setw(DEF_WIDTH) << c_m.class_id
       << setw(DEF_WIDTH) << c_m.status
       << setw(DEF_WIDTH) << c_m.int_mask
       << endl;


    of << showpos;

    /* x */
    of << "                 "
       << setw(DEF_WIDTH) << gsl_vector_get(c_m.x, 0)
       << setw(DEF_WIDTH) << gsl_vector_get(c_m.x, 1)
       << setw(DEF_WIDTH) << gsl_vector_get(c_m.x, 2) << endl;

    /* euler angles */
    gsl_quaternion_unit_quat_to_euler(c_m.Q, &c_theta, &c_phi, &c_psi);
    of << "                 "
       << setw(DEF_WIDTH) << (TO_DEG * c_theta)
       << setw(DEF_WIDTH) << (TO_DEG * c_phi)
       << setw(DEF_WIDTH) << (TO_DEG * c_psi) << endl;

    /* dynamical variables */
    if ( c_m.is_ext_alloc() )
    {

      /* v_cm */
      of << "                 "
         << setw(DEF_WIDTH) << gsl_vector_get(c_m.v, 0)
         << setw(DEF_WIDTH) << gsl_vector_get(c_m.v, 1)
         << setw(DEF_WIDTH) << gsl_vector_get(c_m.v, 2) << endl;

      /* omega */
      of << "                 "
         << setw(DEF_WIDTH) << c_m.omega.dat[1]
         << setw(DEF_WIDTH) << c_m.omega.dat[2]
         << setw(DEF_WIDTH) << c_m.omega.dat[3] << endl;

      of << endl;

    }

    if ( c_m.has_dipole() )
    {
      of << "rigid_body_dipole  "
         << setw(DEF_WIDTH) << gsl_vector_get(c_m.dipole_cm, 0)
         << setw(DEF_WIDTH) << gsl_vector_get(c_m.dipole_cm, 1)
         << setw(DEF_WIDTH) << gsl_vector_get(c_m.dipole_cm, 2) << endl;
    }

    of << endl;

  }

  return EXIT_SUCCESS;

}

/***************************************************************************/

int
emd::uni::saver::save_universe_to_povray_file(const string & out_file_name)
{

  vector<rigid_body_ref> indices;

  for (unsigned int m=0; m<rigid_bodies_count; m++)
  {

    rigid_body_ref ref;

    ref.idx = m;

    ref.n_x = 0;
    ref.n_y = 0;
    ref.n_z = 0;

    indices.push_back(ref);

  }

  save_universe_to_povray_file(indices, out_file_name);

}

/***************************************************************************/

int
emd::uni::saver::save_universe_to_povray_file(
  const vector<rigid_body_ref> & indices, const string & out_file_name)
{

  /* open the file */
  ofstream of(out_file_name.c_str());
  if (!of) return EXIT_FAILURE;

  /***************************/

  /* header */

  of << endl;
  of << "#include \"colors.inc\"" << endl
     << "#include \"default_view.inc\"" << endl
     << "#include \"rigid_bodies.inc\"" << endl << endl;

  of << "union" << endl << "{" << endl;

  /* rigid bodys */

  for (vector<rigid_body_ref>::const_iterator it = indices.begin();
    it != indices.end(); it++)
  {

    unsigned int idx  = (*it).idx;
    int n_x           = (*it).n_x;
    int n_y           = (*it).n_y;
    int n_z           = (*it).n_z;

    rigid_body &  c_m = emd::uni::rigid_bodies[idx];

    of << "  object" << endl
       << "  {"      << endl
       << "    class_id_" << c_m.class_id << endl 
       << "    matrix < " << setprecision(4) << right
       << setw(6) << gsl_matrix_get(c_m.R, 0, 0) << ", "
       << setw(6) << gsl_matrix_get(c_m.R, 1, 0) << ", "
       << setw(6) << gsl_matrix_get(c_m.R, 2, 0) << ", " << endl
       << "             "
       << setw(6) << gsl_matrix_get(c_m.R, 0, 1) << ", "
       << setw(6) << gsl_matrix_get(c_m.R, 1, 1) << ", "
       << setw(6) << gsl_matrix_get(c_m.R, 2, 1) << ", " << endl
       << "             "
       << setw(6) << gsl_matrix_get(c_m.R, 0, 2) << ", "
       << setw(6) << gsl_matrix_get(c_m.R, 1, 2) << ", "
       << setw(6) << gsl_matrix_get(c_m.R, 2, 2) << ", " << endl
       << "             "
       << setw(6) << (gsl_vector_get(c_m.x, 0) + n_x*x_width) << ", "
       << setw(6) << (gsl_vector_get(c_m.x, 1) + n_y*y_width) << ", "
       << setw(6) << (gsl_vector_get(c_m.x, 2) + n_z*z_width) << " >" << endl;

    of << "  }" << endl << endl;

  }

  of << "}" << endl;

  string frame_material_str =
       string("    finish\n") +
       string("    {\n") +
       string("      PlaneFinish\n") +
       string("    }\n") +
       string("    pigment\n") +
       string("    {\n") +
       string("      Red transmit 0.9\n") +
       string("    }");

  string surf_material_str =
       string("  finish\n") +
       string("  {\n") +
       string("    PlaneFinish\n") +
       string("  }\n") +
       string("  pigment\n") +
       string("  {\n") +
       string("    Blue transmit 0.9\n") +
       string("  }");

  string sphere_material_str =
       string("  finish\n") +
       string("  {\n") +
       string("    SphereFinish\n") +
       string("  }\n") +
       string("  pigment\n") +
       string("  {\n") +
       string("    Blue transmit 0.75\n") +
       string("  }");

  /* periodic bounds and surfaces */

  /* spheres */

  for (int idx=0; idx<sph_int_surfs_count; idx++)
  {

    spherical_int_surface &  c_s = emd::uni::sph_int_surfs[idx];

    of << "object" << endl
       << "{"      << endl
       << "  sphere" << endl
       << "  {" << endl
       << "    <" << gsl_vector_get(c_s.r0, 0) << ", "
                  << gsl_vector_get(c_s.r0, 1) << ", "
                  << gsl_vector_get(c_s.r0, 2) << ">, "
                  << c_s.rad << endl
       << "  }" << endl
       << sphere_material_str << endl
       << "}" << endl << endl;

  }

  /* planes */
  vector<emd::geom::plane_3d>  planes;
  emd::geom::plane_3d          new_plane;

  if (x_periodic)
  {
    new_plane.nx = 1;  new_plane.ny = 0;  new_plane.nz = 0;
    new_plane.d  = x_lower_bound;
    planes.push_back(new_plane);
    new_plane.nx = 1;  new_plane.ny = 0;  new_plane.nz = 0;
    new_plane.d  = x_upper_bound;
    new_plane.taste = 0;
    planes.push_back(new_plane);
  }

  if (y_periodic)
  {
    new_plane.nx = 0;  new_plane.ny = 1;  new_plane.nz = 0;
    new_plane.d  = y_lower_bound;
    planes.push_back(new_plane);
    new_plane.nx = 0;  new_plane.ny = 1;  new_plane.nz = 0;
    new_plane.d  = y_upper_bound;
    new_plane.taste = 0;
    planes.push_back(new_plane);
  }

  if (z_periodic)
  {
    new_plane.nx = 0;  new_plane.ny = 0;  new_plane.nz = 1;
    new_plane.d  = z_lower_bound;
    planes.push_back(new_plane);
    new_plane.nx = 0;  new_plane.ny = 0;  new_plane.nz = 1;
    new_plane.d  = z_upper_bound;
    new_plane.taste = 0;
    planes.push_back(new_plane);
  }

  /* flat int surfaces */
  for (int i=0; i<flat_int_surfs_count; i++)
  {
    new_plane.nx = -gsl_vector_get( flat_int_surfs[i].n, 0 );
    new_plane.ny = -gsl_vector_get( flat_int_surfs[i].n, 1 );
    new_plane.nz = -gsl_vector_get( flat_int_surfs[i].n, 2 );
    gsl_blas_ddot( flat_int_surfs[i].n, flat_int_surfs[i].r0, &(new_plane.d) );
    new_plane.d *= -1;
    new_plane.taste = 1;
    planes.push_back(new_plane);
  }

  geom::point_3d interior_point_3d = { gsl_vector_get(interior_point, 0),
                                       gsl_vector_get(interior_point, 1),
                                       gsl_vector_get(interior_point, 2) };

  vector<emd::geom::vertex_point_3d>  points;
  vector<emd::geom::point_3d> *       faces;
  vector<int>                         tastes;
  int                                 n_faces;

  /* evaluate the volume bounded by planes and periodic bounds */
  int stat = emd::geom::eval_interior_volume( planes, interior_point_3d, &faces, tastes,
                                              n_faces, volume );

  if (stat==EXIT_SUCCESS)
  {

    of << "union" << endl << "{" << endl;

    for (int i=0; i<n_faces; i++)
    {

      if (faces[i].size() <= 2) continue;

      switch (tastes[i])
      {

        case 1:
        {

          of << "  polygon {\n"
             << "    " << faces[i].size() << "," << endl;

          for (int j=0; j<faces[i].size()-1; j++)
          {
            of << "    <" << faces[i][j].x << ", "
                          << faces[i][j].y << ", " 
                          << faces[i][j].z << ">,\n";
          }
          int j = faces[i].size()-1;
          of << "    <" << faces[i][j].x << ", "
                        << faces[i][j].y << ", " 
                        << faces[i][j].z << ">\n";
          of << surf_material_str << endl;
          of << "  }\n";

          break;

        }

        case 0:
        {

          of << "  union {\n";
          for (int j=0; j<faces[i].size()-1; j++)
          {
            of << "    cylinder {\n"
               << "      <" << faces[i][j].x << ", " << faces[i][j].y << ", " << faces[i][j].z << ">,\n"
               << "      <" << faces[i][j+1].x << ", " << faces[i][j+1].y << ", " << faces[i][j+1].z << ">,\n"
               << "      " << 0.05
               << "    }\n";
          }

          of << frame_material_str << endl;
          of << "  }\n";

          break;

        }

      }

    }

    of << "}\n";

  }

  return EXIT_SUCCESS;

}

/***************************************************************************/

int
emd::uni::saver::save_nhc_vars_to_file(const string & out_file_name)
{

  bool save_T = false,
       save_P = false;

  switch (dyn::integrator)
  {

    case dyn::NVE_MILLER_SYMPLECTIC:
    {

      break;

    }

    case dyn::NVT_MILLER_SYMPLECTIC_NHC:
    {
      save_T = true;
      break;
    }

    case dyn::NPT_MILLER_SYMPLECTIC_NHC:
    {
      save_T = true;
      save_P = true;
      break;
    }

  }

  if ( !(save_T || save_P) )
    return EXIT_SUCCESS;

  /* open the file */
  ofstream of(out_file_name.c_str());
  if (!of) return EXIT_FAILURE;

  /***************************/

  of << "############################" << endl
     << "## NHC variables snapshot ##" << endl
     << "############################" << endl << endl;

  if (save_T)
  {

    of << "NHC_T_ksi_t      ";
    for (int i=0; i<dyn::NHC_T_length; i++)
      of << dyn::NHC_T_ksi_t[i] << endl
         << "                 ";

    of << endl;

    of << "NHC_T_ksi_r      ";
    for (int i=0; i<dyn::NHC_T_length; i++)
      of << dyn::NHC_T_ksi_r[i] << endl
         << "                 ";

    of << endl;

    of << "NHC_T_s_t        ";
    for (int i=0; i<dyn::NHC_T_length; i++)
      of << dyn::NHC_T_s_t[i] << endl
         << "                 ";

    of << endl;

    of << "NHC_T_s_r        ";
    for (int i=0; i<dyn::NHC_T_length; i++)
      of << dyn::NHC_T_s_r[i] << endl
         << "                 ";

    of << endl;

  }

  if (save_P)
  {

    of << "NHC_P_eps        " << dyn::NHC_P_eps << endl << endl;

    of << "NHC_P_ksi_eps    " << dyn::NHC_P_ksi_eps << endl << endl;

    of << "NHC_P_ksi_eta    ";
    for (int i=0; i<dyn::NHC_P_length; i++)
      of << dyn::NHC_P_ksi_eta[i] << endl
         << "                 ";

    of << endl;

    of << "NHC_P_eta        ";
    for (int i=0; i<dyn::NHC_P_length; i++)
      of << dyn::NHC_P_eta[i] << endl
         << "                 ";

  }

  return EXIT_SUCCESS;

}

/***************************************************************************/
